/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef SOCKET_USNET_VFS_H
#define SOCKET_USNET_VFS_H

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "fs/vfs.h"
#include "posix/sys/socket.h"
#include "fs/socket/socket.h"

#include "sched/scheduler.h"
#include "process/processcontrol.h"

///////////////////////////////////////////////////////////////////////////////
// Debug definitions
///////////////////////////////////////////////////////////////////////////////
#define NETWORK_DEBUG_USNET_ENTRY 0
#define NETWORK_DEBUG_USNET_OTHER 0

#if NETWORK_DEBUG_USNET_ENTRY
	#define DEBUG_ENTRY(str) str
#else
	#define DEBUG_ENTRY(str)
#endif

#if NETWORK_DEBUG_USNET_OTHER
	#define DEBUG_USN_OTHER(str) str
#else
	#define DEBUG_USN_OTHER(str)
#endif

///////////////////////////////////////////////////////////////////////////////
// Definitions
///////////////////////////////////////////////////////////////////////////////
#define USNET_INIT_PATH "usnet-init"
#define USNET_INIT_MAGIC 968745
#define USNET_INIT_VER 1

#define USNET_CALLTYP_BIND 1
#define USNET_CALLTYP_RECVFROM 2
#define USNET_CALLTYP_SENDTO 3
#define USNET_CALLTYP_SOCKET 4
#define USNET_CALLTYP_CLOSE 5
#define USNET_CALLTYP_LISTEN 6
#define USNET_CALLTYP_ACCEPT 7
#define USNET_CALLTYP_CONNECT 8
#define USNET_CALLTYP_SOCKOPT 9
#define USNET_CALLTYP_SOCKMISC 10
#define USNET_CALLTYP_NETDB 11

///////////////////////////////////////////////////////////////////////////////
// Non-standard structures
///////////////////////////////////////////////////////////////////////////////
struct ioctl_init_cmd
{
	int version;
	int domain;
};

struct usnet_queue_item_transfer
{
	// Unique identification of this socket.
	// New file descriptor in case of accept
	// returned as regular data.
	int pid;
	int fd;

	// File descriptor flags or return value
	union {
		int flags;
		int returnvalue;
	};

	// Identification of this transfer item
	int seq;
	char calltype;

	// Address often supplied with socket calls
	int addrlen;
	int addrdat[8];
};

struct usnet_queue_item
{
	// The client process
	short pid;
	short fd;
	struct ThreadControl *trc;

	// Data
	union {
		struct recvfrom_args recvargs;
		struct sendto_args sendargs;
		int socket[3];
	};

	// Queue item sequence number
	int seq;

	// True when not handled by user stack
	char fresh;

	// Call type, 1 for write 0 for read
	char calltype;

	// Next queue item
	struct usnet_queue_item *next;
};

struct usnet_internal
{
	struct netstack usnet_stack;
	struct usnet_queue_item *queue;
	int seq;
	int isdead;
	struct List blocked;
	struct List blocked_owner;
};

///////////////////////////////////////////////////////////////////////////////
// Exports
///////////////////////////////////////////////////////////////////////////////
extern struct netstack *usnet_registered_stacks;
extern struct netstack net_usnet;

///////////////////////////////////////////////////////////////////////////////
// Functions
///////////////////////////////////////////////////////////////////////////////
struct usnet_queue_item *usnet_queue_pull (struct usnet_internal *usnet);
void usnet_queue_push (struct usnet_internal *usnet, struct usnet_queue_item *it);
struct usnet_queue_item *usnet_queue_find (struct usnet_internal *usnet, int seq);
void usnet_queue_remove (struct usnet_internal *usnet, struct usnet_queue_item *it);
int usnet_cancel (struct ThreadControl *ctc, unsigned int args);
int usnet_init_ioctl (struct FileNode *node, unsigned int code, void *data);

///////////////////////////////////////////////////////////////////////////////
// Network stack user entry points
///////////////////////////////////////////////////////////////////////////////
int usnet_user_bind (struct FileNode *node, const struct sockaddr *addr, socklen_t addrlen);
int usnet_user_recvfrom (struct FileNode *node, struct recvfrom_args *args);
int usnet_user_sendto (struct FileNode *node, struct sendto_args *args);
int usnet_user_socket (struct FileNode *node, int domain, int type, int protocol);
int usnet_user_listen (struct FileNode *node, void *args);
int usnet_user_accept (struct FileNode *node, struct FileNode *newnode, void *args);
int usnet_user_connect (struct FileNode *node, void *args);
int usnet_user_sockopt (struct FileNode *node, void *args);
int usnet_user_sockmisc (struct FileNode *node, void *args);
int usnet_user_netdb (struct FileNode *node, void *args);
int usnet_user_close (struct ThreadControl *ctc, struct FileNode *node);

///////////////////////////////////////////////////////////////////////////////
// Network stack interface entry points
///////////////////////////////////////////////////////////////////////////////
int usnet_bind (struct FileNode *node, const struct sockaddr *addr, socklen_t addrlen);
int usnet_recvfrom (struct FileNode *node, struct recvfrom_args *args);
int usnet_sendto (struct FileNode *node, struct sendto_args *args);
int usnet_socket (struct FileNode *node, int domain, int type, int protocol);
int usnet_close (struct ThreadControl *ctc, struct FileNode *node);

#endif
