/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef SOCKET_VFS_H
#define SOCKET_VFS_H

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "fs/vfs.h"
#include "posix/sys/socket.h"

///////////////////////////////////////////////////////////////////////////////
// Non-standard structures
///////////////////////////////////////////////////////////////////////////////
struct recvfrom_args
{
	void *buf;
	unsigned int len;
	int flags;
	struct sockaddr *addr;
	socklen_t *addrlen;
};

struct sendto_args
{
	const void *buf;
	unsigned int len;
	int flags;
	const struct sockaddr *addr;
	socklen_t addrlen;
};

///////////////////////////////////////////////////////////////////////////////
// Network stack declaration structure
///////////////////////////////////////////////////////////////////////////////
struct netstack
{
	int (*bind) (struct FileNode *, const struct sockaddr *, socklen_t);
	int (*recvfrom) (struct FileNode *, struct recvfrom_args *);
	int (*sendto) (struct FileNode *, struct sendto_args *);
	int (*socket) (struct FileNode *, int, int, int);

	int (*listen) (struct FileNode *, void *);
	int (*accept) (struct FileNode *, struct FileNode *, void *);
	int (*connect) (struct FileNode *, void *);
	int (*sockopt) (struct FileNode *, void *);
	int (*sockmisc) (struct FileNode *, void *);
	int (*netdb) (struct FileNode *, void *);

	int (*close) (struct ThreadControl *ctc, struct FileNode *);

	int domain;

	// Only used when list is in RAM
	struct netstack *next;
};

///////////////////////////////////////////////////////////////////////////////
// Definitions for unsupported calls
///////////////////////////////////////////////////////////////////////////////
#define SOCKET_UNSUPPORTED_LISTEN \
	(int(*)(struct FileNode*, void*)) vfs_unsupported
#define SOCKET_UNSUPPORTED_ACCEPT \
	(int(*)(struct FileNode*, struct FileNode*, void*)) vfs_unsupported
#define SOCKET_UNSUPPORTED_CONNECT \
	(int(*)(struct FileNode*, void*)) vfs_unsupported
#define SOCKET_UNSUPPORTED_SOCKOPT \
	(int(*)(struct FileNode*, void*)) vfs_unsupported
#define SOCKET_UNSUPPORTED_SOCKMISC \
	(int(*)(struct FileNode*, void*)) vfs_unsupported
#define SOCKET_UNSUPPORTED_NETDB \
	(int(*)(struct FileNode*, void*)) vfs_unsupported

///////////////////////////////////////////////////////////////////////////////
// Macros
///////////////////////////////////////////////////////////////////////////////
#define NET_DESCRIPTOR(descriptorname,bind,recvfrom,sendto,socket, \
	listen,accept,connect,sockopt,sockmisc,netdb,close,domain) \
	struct netstack net_##descriptorname \
	__attribute((__section__(".net_list"))) \
	__attribute((__used__)) \
	= {bind, recvfrom, sendto, socket, \
		listen, accept, connect, sockopt, sockmisc, netdb, \
		close, domain, 0};

///////////////////////////////////////////////////////////////////////////////
// Exports
///////////////////////////////////////////////////////////////////////////////
extern const struct VFSDriver vfs_driver_socket;

///////////////////////////////////////////////////////////////////////////////
// Functions
///////////////////////////////////////////////////////////////////////////////
int vfs_socket_bind (struct FileNode *node, const struct sockaddr *addr, socklen_t addrlen);
int vfs_socket_recvfrom (struct FileNode *node, struct recvfrom_args *args);
int vfs_socket_sendto (struct FileNode *node, struct sendto_args *args);
int vfs_socket_socket (struct FileNode *node, int domain, int type, int protocol);

int vfs_socket_listen (struct FileNode *node, void *args);
int vfs_socket_accept (struct FileNode *node, struct FileNode *newnode, void *args);
int vfs_socket_connect (struct FileNode *node, void *args);
int vfs_socket_sockopt (struct FileNode *node, void *args);
int vfs_socket_sockmisc (struct FileNode *node, void *args);
int vfs_socket_netdb (struct FileNode *node, void *args);

int vfs_socket_domain_exists (int domain);
int vfs_socket_register_stack (struct netstack *stack);
int vfs_socket_unregister_stack (struct netstack *stack);
void vfs_socket_kill_references (struct netstack *stack);

#endif
