/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef VFS_H
#define VFS_H

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "posix/sys/stat.h"
#include "posix/sys/statvfs.h"
#include "posix/dirent.h"
#include "sched/threadcontrol.h"
#include "../config/live/kernelconfig.h"

///////////////////////////////////////////////////////////////////////////////
// Structures
///////////////////////////////////////////////////////////////////////////////

/*
 * File node structure. Every open file descriptor will point to an entry
 * like this. It keeps track of the responsible file system, device driver,
 * how many times it is referenced (dup2 etc) and some flags. At last it
 * provides a pointer that a device driver may use for private data.
 *
 * Size: 18 bytes
 */
struct FileNode
{
	const struct VFSDriver *fs;			// File system driver for this node
	union {
		const struct DeviceDriver *dev;	// Device driver responsible for this node
		struct netstack *net;			// Network stack responsible for this node
	};
	void *priv;							// Private driver data
	short flags;						// Open flags
	short usecounter;					// Number of references
	short fd;							// File descriptor in the used program
};

/*
 * File system driver definition, providing the pointers to the functions
 * implementing a particular file system. It is recommended to keep this as
 * a static struct in flash. Unused functions can be set to a function
 * that simply returns an error value.
 */
struct VFSDriver
{
	// Basic file operations, aligns with the DeviceDriver structure
	int (*open)(struct FileNode*, const char*, int, int);
	int (*close)(struct ThreadControl*, struct FileNode*);
	int (*read)(struct FileNode*, void*, int);
	int (*write)(struct FileNode*, void*, int);
	int (*ioctl_fcntl)(struct FileNode*, unsigned int, void*);

	// Extended file operations
	int (*lseek)(struct FileNode*, int, int);
	int (*fstat)(struct FileNode*, struct stat *st);

	// File system operations
	int (*stat)(const char*, struct stat *st);
	int (*link)(const char*, const char*);
	int (*unlink)(const char*);
	int (*rename)(const char*, const char*);
	int (*mkdir)(const char*, int);
	int (*opendir)(const char*);
	int (*closedir)();
	int (*readdir)();
	int (*statvfs)(const char *path, struct statvfs*);

	// Mount and unmount functions
	int (*mount)(struct FileNode*, char*);
	int (*unmount)(void);
};

///////////////////////////////////////////////////////////////////////////////
// External variables
///////////////////////////////////////////////////////////////////////////////
extern struct FileNode nodetable[NODETABLE_SIZE];

///////////////////////////////////////////////////////////////////////////////
// VFS General Functions
///////////////////////////////////////////////////////////////////////////////
/*
 * Mount a disk (or file) at mountpoint with driver vfs.
 */
int vfs_mount (const char *mountpoint, struct VFSDriver *vfs, struct FileNode *disk);

/*
 * Unmount a disk/file from mountpoint and close disk/file handle.
 */
int vfs_unmount (const char *mountpoint);

/*
 * Relative path to absolute path conversion. Thread-safe as long as all
 * argument strings are private to the caller. The output is string inb
 * relative to the "working" directory string ina.
 */
void vfs_reltoabs (char *out, const char *ina, const char *inb);

///////////////////////////////////////////////////////////////////////////////
// File System Operations
///////////////////////////////////////////////////////////////////////////////
int vfs_open (struct FileNode *fnode, const char *cwd, const char *filename, int flags, int mode);
int vfs_close (struct ThreadControl *ctc, struct FileNode *fnode);
int vfs_read (struct FileNode *fnode, char *data, int n);
int vfs_write (struct FileNode *fnode, char *data, int n);
int vfs_stat (const char *cwd, const char *filename, struct stat *st);
int vfs_unlink (const char *cwd, const char *filename);
int vfs_rename (const char *cwd, const char *oldname, const char *newname);
int vfs_mkdir (const char *cwd, const char *name, int mode);
int vfs_opendir (const char *cwd, const char *name);
int vfs_closedir (void);
int vfs_readdir (struct dirent *de);
int vfs_chdir (char *cwd, const char *dst);
int vfs_statvfs (const char *cwd, const char *path, struct statvfs *buf);

int vfs_unsupported (void);

#endif
