/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef IOCTL_H
#define IOCTL_H

/*
 * This file defines all ioctl command numbers. All ioctls are defined using a
 * macro just like on linux, in order to make defining unique ioctls easier.
 * Just look below for some examples. The macro assembles a 32-bit integer
 * using a magic byte, the call id and the size of the data argument.
 *
 * The magic byte should be selected for the driver. It can be anything but
 * should be as unique as possible for the driver. The id is a byte-sized
 * integer for use within the driver, you should start with zero. At last
 * the size of the data argument type is included. It helps making the ioctl
 * unique and helps catching errors caused by modifying the size of something
 * unintentionally.
 */

// Definitions for ioctls
#define _IO(magic,id,size)		((0 << 30) | (sizeof(size) << 16) | (magic << 8) | (id << 0))
#define _IOW(magic,id,size)		((1 << 30) | (sizeof(size) << 16) | (magic << 8) | (id << 0))
#define _IOR(magic,id,size)		((2 << 30) | (sizeof(size) << 16) | (magic << 8) | (id << 0))
#define _IORW(magic,id,size)	((3 << 30) | (sizeof(size) << 16) | (magic << 8) | (id << 0))

// Definitions for tty
#define TCGETS		_IOR('t', 0, struct termios)
#define TCSETS		_IOW('t', 1, struct termios)
#define TISATTY		_IO('t', 2, void)
#define TCSBRK		_IO('t', 3, void)
#define TCXONC		_IO('t', 4, void)
#define TCFLSH		_IO('t', 5, void)
#define TCGETPGRP	_IO('t', 6, void)
#define TCSETPGRP	_IO('t', 7, void)

#define TTYSETMODE	_IO('t', 8, int)
#define TTYGETRXBUSY _IO('t', 9, int)

#define TTYSETCBLINK _IO('t', 10, int)
#define TTYLCDDIS	_IO('t', 11, int)

#define TTYSETRXCIRC _IO('t', 12, void)
#define TTYGETWRPOS _IO('t', 13, void)

// Definitions for SPI
#define SPIBUSY		_IO('s', 0, void)
#define SPICS		_IOW('s', 1, unsigned int)
#define SPIFORCEUNLOCK _IO('s', 2, void)

// Definitions for SPI FLASH
#define SPIFSETSECT	_IOW('f', 0, unsigned int)
#define SPIFGETSECT	_IOR('f', 1, void)
#define SPIFJEDEC	_IOR('f', 2, void)
#define SPIFGETSECTSIZE _IOR('f', 3, short)
#define SPIFGETSECTCNT _IOR('f', 4, unsigned int)

// Definitions for OWL VHF
#define OWLPTYPE	_IOW('o', 0, unsigned char)
#define OWLSEXT		_IOW('o', 1, unsigned char)
#define OWLCLEAR	_IO('o', 2, void)

// Definitions for sys
#define SYSSHOWREL	_IOW('y', 0, char*)
#define SYSREBOOT	_IO('y', 1, char*)

// Definitions for rtc
#define RTCGETTIME	_IOR('r', 0, int*)
#define RTCSETTIME	_IOW('r', 1, int*)

// Definitions for i2s
#define I2SSTART	_IOW('i', 0, int)
#define I2SSTOP		_IOW('i', 1, void)
#define I2SGETPEAKS	_IOW('i', 2, int)

// General file control
#define F_SETOWN	_IOW('a', 0, int)
#define F_GETFL		_IOW('a', 1, int)
#define F_SETFL		_IOW('a', 2, int)

#endif

