/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef SCHEDULER_H
#define SCHEDULER_H

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "sched/threadcontrol.h"

///////////////////////////////////////////////////////////////////////////////
// Definitions
///////////////////////////////////////////////////////////////////////////////
#define SCHED_NPRIORITIES 3
#define SCHED_PRIORITY_LOW 2
#define SCHED_PRIORITY_MEDIUM 1
#define SCHED_PRIORITY_HIGH 0

#define SCHED_STATUS_RUN		0
#define SCHED_STATUS_BLOCKED	1
#define SCHED_STATUS_ZOMBIE		2
#define SCHED_STATUS_SLEEPING	4
#define SCHED_STATUS_STOPPED	8
#define SCHED_STATUS_DEFERRED	16
#define SCHED_STATUS_CRASHED	15

///////////////////////////////////////////////////////////////////////////////
// Exports
///////////////////////////////////////////////////////////////////////////////
extern struct ThreadControl *currentthc;

///////////////////////////////////////////////////////////////////////////////
// Functions
///////////////////////////////////////////////////////////////////////////////

/*
 * Initialize the scheduler and its lists. Only to be called once on kernel
 * startup.
 */
void SchedInitialize (void);

/*
 * Start the scheduler. This does not return.
 */
void SchedStart (void);

/*
 * Platform specific critical section start. May disable interrupts completely
 * or raise the base priority.
 */
int SchedEnterCritical (void);

/*
 * Exit a critical section.
 */
void SchedExitCritical (int old);

/*
 * Caller has to allocate ThreadControl and stack space. The new task is put
 * on the run list, ready to execute. Privilege level has to be set before
 * calling this.
 */
void SchedAddThread (struct ThreadControl *t,
	void (*entry)(void),
	unsigned int *stack,
	int stacksize,
	int priority);

/*
 * Remove a thread from all lists and set its state to zombie. This is
 * intended for permanent removal. For any other purpose, see SchedState.
 */
void SchedRemoveProcess (struct ThreadControl *thc);

/*
 * Change state, see C file for notes.
 */
void SchedState (
	struct ThreadControl *thc,
	unsigned char newstate,
	int (*callback)(struct ThreadControl*, unsigned int),
	unsigned int args,
	unsigned int time);

/*
 * When a thread is set to blocked state with a callback, it goes on the
 * work list until fetched by this function by a worker thread. Note that
 * if there are no work, this function blocks (sleeps) until there is.
 */
struct ThreadControl *SchedFetchWork (void);

/*
 * This MUST be called whenever the system tick counter is incremented,
 * and ONLY then.
 */
void SchedSystickHandler (void);

/*
 * When a thread returns from a signal handler, it executes a system call.
 * Call this from the system call handler to reset state.
 */
void SchedLeaveCurrentProcessSignal (void);

/*
 * Crash recovery, only for use by MemManage_Handler. Called
 * in the same way as PanicInternal, except stacktype is unused.
 * This function removes the current process from all active lists
 * and marks it as crashed, storing the stack pointer as well as
 * the hfsr, cfsr, bfar, mmar and faulttype. Note that the caller is
 * responsible for any memory that has to be freed.
 */
void SchedCrashRecover (unsigned int *stack, int stacktype, int faulttype);

/*
 * Yield the current process. May not happen until interrupts are enabled
 * or the current SVC returns.
 */
void SchedYield (void);

#endif
