/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef TIMER_H
#define TIMER_H

/*
 * How to Use:
 *  1. Allocate timer handle somewhere.
 *  2. Set all members of the handle to zero, memset recommended.
 *  3. Set the callback, interval (milliseconds) and flags:
 *       TIMER_FLAG_REPEATING: Callback is called every interval milliseconds.
 *       TIMER_FLAG_NOIRQ: Interrupts are disabled when Callback is called.
 *  4. Use TimerRegister to set the timer.
 *  5. Wait for timer to trigger.
 *  6. Use TimerUnregister to stop the timer. If TIMER_FLAG_REPEATING is not
 *     specified, it will be unregistered on trigger automatically. Otherwise,
 *     use it inside the callback. If called from outside the callback, it must
 *     be protected in a critical section.
 *
 * WARNING:
 *  The register and unregister functions are NOT thread safe. If you are
 *  calling them outside of interrupts and such, the calls have to be
 *  protected.
 */

///////////////////////////////////////////////////////////////////////////////
// Definitions
///////////////////////////////////////////////////////////////////////////////
#define TIMER_FLAG_REPEATING 1
#define TIMER_FLAG_NOIRQ 2

///////////////////////////////////////////////////////////////////////////////
// Structures
///////////////////////////////////////////////////////////////////////////////
typedef struct TimerHandle
{
	int reserved1; // Next
	int reserved2; // Prev
	void (*Callback)(struct TimerHandle*);
	int reserved3; // Container
	int reserved4; // Sort value
	unsigned int interval;
	int flags;
}TimerHandle;

///////////////////////////////////////////////////////////////////////////////
// Functions
///////////////////////////////////////////////////////////////////////////////

/*
 * Start the timer worker thread. Only to be called on kernel startup.
 */
void TimerInitialize (void);

/*
 * Register an already initialized timer handle (see above). This function is
 * not thread safe, so protect it in a critical section (interrupts disabled)
 * if required.
 */
void TimerRegister (TimerHandle *handle);

/*
 * Unregister a timer handle. Not thread safe, see TimerRegister.
 */
void TimerUnRegister (TimerHandle *handle);

/*
 * Returns true if the timer is currently registered.
 */
int TimerIsRegistered (TimerHandle *handle);

#endif
