/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

.section .text
.syntax unified
.cpu cortex-r4
.arm

/*
 * IMPORTANT
 *
 * If the crash happens when in thread mode, it is assumed that an unprivileged
 * user process has crashed. Since a process is never run again the crash code
 * does not care about the entire state of the process. Only the most important
 * register info is kept.
 */

/*
 * Get fault status registers
 */
.weak GetFaultRegs
.type GetFaultRegs, %function
GetFaultRegs:
	stmdb sp!, {r1}
	mrc p15, 0, r1, c5, c0, 0	@ Read the DFSR (Data Fault Status Register)
	str r1, [r0, #0]
	mrc p15, 0, r1, c5, c0, 1	@ Read the IFSR (Instruction Fault Status Register)
	str r1, [r0, #4]
	mrc p15, 0, r1, c5, c1, 0	@ Read the ADFSR (Auxiliary Data Fault Status Register)
	str r1, [r0, #8]
	mrc p15, 0, r1, c5, c1, 1	@ Read the AIFSR (Auxiliary Instruction Fault Status Register)
	str r1, [r0, #12]
	mrc p15, 0, r1, c6, c0, 0	@ Read the DFAR (Data Fault Address Register)
	str r1, [r0, #16]
	mrc p15, 0, r1, c6, c0, 2	@ Read the IFAR (Instruction Fault Address Register)
	str r1, [r0, #20]
	ldmia sp!, {r1}
	bx lr
.size GetFaultRegs, .-GetFaultRegs

/*
 * General fault handler
 */
.weak FaultHandler
.type FaultHandler, %function
FaultHandler:
	cmp r1, #1				@ Compare with PANIC_STACKTYPE_USER
	beq TryRecover
	b PanicInternal
TryRecover:
	bl SchedCrashRecover	@ Attempt crash recovery
	b SchedContextNext		@ Jump into scheduler and restore a different process
.size FaultHandler, .-FaultHandler

/*
 * Undefined instruction handler
 */
.section .text.UndefinedHandler
.global UndefinedHandler
.type UndefinedHandler, %function
UndefinedHandler:
	cps #31						@ Supervisor mode: Access to everything but with user stack
	stmdb sp, {r0-lr}			@ Stack all the user regs except PC and CPSR, R0-LR is now free to use
	sub r0, sp, #60				@ Cannot have writeback in the above, sub 15 words and store in R0
	cps #18						@ Go back to IRQ mode, now have access to old PC and CPSR
	mrs r1, spsr				@ Get the CPSR from the banked reg
	stmdb r0!, {r1,lr}			@ Stack CPSR and PC

	and r1, #31
	cmp r1, #16					@ Normally 16
	mov r1, #0					@ PANIC_STACKTYPE_KERNEL
	moveq r1, #1				@ PANIC_STACKTYPE_USER
	mov r2, #1					@ PANIC_FAULTTYPE_UNDEFINST
	b FaultHandler
.size UndefinedHandler, .-UndefinedHandler

/*
 * Prefetch abort handler (memory access error)
 */
.weak PrefetchAbort
.type PrefetchAbort, %function
PrefetchAbort:
	cps #31						@ Supervisor mode: Access to everything but with user stack
	stmdb sp, {r0-lr}			@ Stack all the user regs except PC and CPSR, R0-LR is now free to use
	sub r0, sp, #60				@ Cannot have writeback in the above, sub 15 words and store in R0
	cps #18						@ Go back to IRQ mode, now have access to old PC and CPSR
	mrs r1, spsr				@ Get the CPSR from the banked reg
	stmdb r0!, {r1,lr}			@ Stack CPSR and PC

	and r1, #31
	cmp r1, #16
	mov r1, #0					@ PANIC_STACKTYPE_KERNEL
	moveq r1, #1				@ PANIC_STACKTYPE_USER
	mov r2, #2					@ PANIC_FAULTTYPE_PREFABORT
	b FaultHandler
.size PrefetchAbort, .-PrefetchAbort

/*
 * Data abort handler (memory access error)
 */
.weak DataAbort
.type DataAbort, %function
DataAbort:

	@ Page handler code, should be in abort mode (cps #23)
	stmdb sp!, {r0-r6,lr}		@ Stack regs that will be overwritten
	sub sp, sp, #12

	mrs r1, spsr				@ Get the CPSR from the banked reg
	and r1, #31
	cmp r1, #16
	bne PagerFailed				@ Processor mode is not user, always fatal

	mrc p15, 0, r3, c5, c0, 0	@ Read the DFSR (Data Fault Status Register)
	movw r1, #1039				@ 0x40F
	and r3, r3, r1
	cmp r3, #13
	bne PagerFailed				@ Not a permission fault

	ldr	r4, =currentthc			@ r1 is currentthc pointer
	ldr r1, [r4]
	mov r2, sp					@ r2 is scratch stack space
	mrc p15, 0, r0, c6, c0, 0	@ Read the DFAR (Data Fault Address Register) into r0

	bl PagerWalk

	cmp r0, #0
	beq PagerFailed

	ldr r3, =pagermpuregion
	ldr r3, [r3]
	ldm r0, {r4-r6}
	mcr p15, 0, r3, c6, c2, 0	@ Set region
	mcr p15, 0, r4, c6, c1, 0	@ Set address
	mcr p15, 0, r6, c6, c1, 4	@ Set access rights
	mcr p15, 0, r5, c6, c1, 2	@ Set size and enable

	add sp, sp, #12
	ldmia sp!, {r0-r6,lr}		@ Unstack
	subs pc, lr, #8				@ Return

PagerFailed:
	add sp, sp, #12
	ldmia sp!, {r0-r6,lr}

	@ Previous code from here

	cps #31						@ Supervisor mode: Access to everything but with user stack
	stmdb sp, {r0-lr}			@ Stack all the user regs except PC and CPSR, R0-LR is now free to use
	sub r0, sp, #60				@ Cannot have writeback in the above, sub 15 words and store in R0
	cps #18						@ Go back to IRQ mode, now have access to old PC and CPSR
	mrs r1, spsr				@ Get the CPSR from the banked reg
	stmdb r0!, {r1,lr}			@ Stack CPSR and PC

	and r1, #31
	cmp r1, #16
	mov r1, #0					@ PANIC_STACKTYPE_KERNEL
	moveq r1, #1				@ PANIC_STACKTYPE_USER
	mov r2, #3					@ PANIC_FAULTTYPE_DATAABORT
	b FaultHandler
.size DataAbort, .-DataAbort
