/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "arch/stm32f407/rcc.h"
#include "mem/mem.h"
#include "print/print.h"
#include "drivers/gpio/gpio-stm32f407.h"
#include "../config/live/platform.h"

///////////////////////////////////////////////////////////////////////////////
// Register sets
///////////////////////////////////////////////////////////////////////////////
typedef struct _FMCSD
{
	volatile unsigned int SDCR1;
	volatile unsigned int SDCR2;
	volatile unsigned int SDTR1;
	volatile unsigned int SDTR2;
	volatile unsigned int SDCMR;
	volatile unsigned int SDRTR;
	volatile unsigned int SDSR;
} _FMCSD;

///////////////////////////////////////////////////////////////////////////////
// Definitions
///////////////////////////////////////////////////////////////////////////////
#define FMCSD ((_FMCSD*) 0xA0000140)

///////////////////////////////////////////////////////////////////////////////
// Functions
///////////////////////////////////////////////////////////////////////////////
void PlatformInitSDRAM (void)
{
	// Enable clocks: IO CDEFGH
	RCC->AHB1ENR |= 0xFC;
	RCC->AHB3ENR |= 1;

	// Pins
	GPIOInit (GPIOC, PLATFORM_FSMC_PINS_PORTC,
		GPIO_AF | GPIO_SPEED_VERYHIGH | GPIO_ALTERNATE(12));

	GPIOInit (GPIOD, PLATFORM_FSMC_PINS_PORTD,
		GPIO_AF | GPIO_SPEED_VERYHIGH | GPIO_ALTERNATE(12));

	GPIOInit (GPIOE, PLATFORM_FSMC_PINS_PORTE,
		GPIO_AF | GPIO_SPEED_VERYHIGH | GPIO_ALTERNATE(12));

	GPIOInit (GPIOF, PLATFORM_FSMC_PINS_PORTF,
		GPIO_AF | GPIO_SPEED_VERYHIGH | GPIO_ALTERNATE(12));

	GPIOInit (GPIOG, PLATFORM_FSMC_PINS_PORTG,
		GPIO_AF | GPIO_SPEED_VERYHIGH | GPIO_ALTERNATE(12));

	GPIOInit (GPIOH, PLATFORM_FSMC_PINS_PORTH,
		GPIO_AF | GPIO_SPEED_VERYHIGH | GPIO_ALTERNATE(12));

	// Program memory device features
	FMCSD->SDCR1 =
		(0 << 13) |	// RPIPE 2
		(1 << 12) |	// RBURST 1
		(2 << 10) |	// SDCLK HCLK/2 (168 MHz / 2 = 84 MHz)
		(0 << 9) |	// WP 0
		(2 << 7) |	// CAS 3
		(1 << 6) |	// NB 1 (4 banks)
		(1 << 4) |	// MWID 1 (16-bit)
		(1 << 2) |	// NR 1 (12-bit)
		(0 << 0);	// NC 0 (8-bit)
	FMCSD->SDCR2 = FMCSD->SDCR1;

	// Program memory device timing: TRP and TRC in SDTR1
	FMCSD->SDTR1 =
		(2 << 24) |		// Row to column delay, t_rcd (activate to read/write) (3)
		(2 << 20) |		// Row precharge delay, t_rp (3)
		(3 << 16) |		// Recovery delay (4)
		(10 << 12) |	// Row cycle delay (11)
		(15 << 8) |		// Self refresh time (16)
		(11 << 4) |		// Exit self-refresh delay (12)
		(1 << 0);		// Load mode register to active (2)
	FMCSD->SDTR2 = FMCSD->SDTR1;

	// Set mode bits to 001 and configure target bank bits in SDCMR, starts clock
	FMCSD->SDCMR = (0 << 9) | (0 << 5) | (0 << 4) | (1 << 3) | 1;

	// Wait for power up delay (should check datasheet)
	volatile unsigned int delay = 1680000;
	while (delay) delay--;

	// Set mode bits to 010 (precharge all) command
	FMCSD->SDCMR = (0 << 9) | (0 << 5) | (0 << 4) | (1 << 3) | 2;

	// Set mode bits to 011 and number of consecutive auto-refresh commands.
	FMCSD->SDCMR = (0 << 9) | (8 << 5) | (0 << 4) | (1 << 3) | 3;

	// Load mode register command
	unsigned int mode = (2 << 4);
	FMCSD->SDCMR = (mode << 9) | (8 << 5) | (0 << 4) | (1 << 3) | 4;

	// Program refresh rate
	FMCSD->SDRTR = 448 << 1;
}

#if PLATFORM_USE_EXTERNAL_MEMORY_SRAM
void PlatformInitSRAM (void)
{
	// Pins already enabled through LCD

	// Write enabled, 16-bit memory bus
	FSMC->bank[PLATFORM_MEM_FSMC].BCR = 0x00001010;

	// Timings
	FSMC->bank[PLATFORM_MEM_FSMC].BTR = PLATFORM_MEM_CLOCK_CONFIG;

	// Bank enable
	FSMC->bank[PLATFORM_MEM_FSMC].BCR |= 1;
}
#endif
