/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "panic/panic.h"
#include "print/print.h"
#include "sched/scheduler.h"
#include "process/processcontrol.h"

///////////////////////////////////////////////////////////////////////////////
// External Functions
///////////////////////////////////////////////////////////////////////////////
void GetFaultRegs (unsigned int *regs);

///////////////////////////////////////////////////////////////////////////////
// Functions
///////////////////////////////////////////////////////////////////////////////
#if PANIC_ENABLE_TTY
void PanicInternalTTY (unsigned int *stack, int stacktype, int faulttype, unsigned int *regs)
{
	// Print header
	Print ("Crash detected!\n");
	Print (stacktype == PANIC_STACKTYPE_USER ? "Userspace\n" : "Kernel space\n");

	// Fault type
	const char *type = "Unknown handler call\n";

	if (faulttype == PANIC_FAULTTYPE_KERNEL) type = "Kernel panic\n";
	if (faulttype == PANIC_FAULTTYPE_UNDEFINST) type = "Undefined instruction\n";
	if (faulttype == PANIC_FAULTTYPE_PREFABORT) type = "Prefetch abort\n";
	if (faulttype == PANIC_FAULTTYPE_DATAABORT) type = "Data abort\n";

	Print (type);

	// If fault type was a "manual" kernel panic
	if (faulttype == PANIC_FAULTTYPE_KERNEL)
	{
		Print ((const char*) stack);
	}
	else
	{
		// Print all regular registers
		Print ("psr 0x%x\n", stack[0]);
		Print ("pc  0x%x\n", stack[1]);
		Print ("r0  0x%x\n", stack[2]);
		Print ("r1  0x%x\n", stack[3]);
		Print ("r2  0x%x\n", stack[4]);
		Print ("r3  0x%x\n", stack[5]);
		Print ("r4  0x%x\n", stack[6]);
		Print ("r5  0x%x\n", stack[7]);
		Print ("r6  0x%x\n", stack[8]);
		Print ("r7  0x%x\n", stack[9]);
		Print ("r8  0x%x\n", stack[10]);
		Print ("r9  0x%x\n", stack[11]);
		Print ("r10 0x%x\n", stack[12]);
		Print ("r11 0x%x\n", stack[13]);
		Print ("r12 0x%x\n", stack[14]);
		Print ("sp  0x%x\n", stack[15]);
		Print ("lr  0x%x\n", stack[16]);

		// Print fault regs
		Print ("dfsr  0x%x\n", regs[0]); // Data Fault Status Register
		Print ("ifsr  0x%x\n", regs[1]); // Instruction Fault Status Register
		Print ("adfsr 0x%x\n", regs[2]); // Auxiliary Data Fault Status Register
		Print ("aifsr 0x%x\n", regs[3]); // Auxiliary Instruction Fault Status Register
		Print ("dfar  0x%x\n", regs[4]); // Data Fault Address Register
		Print ("ifar  0x%x\n", regs[5]); // Instruction Fault Address Register
	}
}
#endif

void PanicInternal (unsigned int *stack, int stacktype, int faulttype)
{
	// Disable interrupts
	SchedEnterCritical ();

	// Get fault regs
	unsigned int regs[6];
	GetFaultRegs (regs);

	// Print to lcd/tty
#if PANIC_ENABLE_TTY
	PanicInternalTTY (stack, stacktype, faulttype, regs);
#endif

	// Loop forever
	while (1);
}

void PanicPrintRegionSize (unsigned int reg)
{
	char *textstrings[] = {"Undef", "Undef", "Undef", "Undef",
		"32B", "64B", "128B", "256B", "512B",
		"1KB", "2KB", "4KB", "8KB", "16KB", "32KB", "64KB", "128KB", "256KB", "512KB",
		"1MB", "2MB", "4MB", "8MB", "16MB", "32MB", "64MB", "128MB", "256MB", "512MB",
		"1GB", "2GB", "4GB"};

	unsigned int size = (reg >> 1) & 0x01F;

	Print ("Region size ");
	Print (textstrings[size]);
	Print ("\n");
}

void PanicPrintReport (struct ThreadControl *ctc)
{
#if PANIC_ENABLE_TTY
	unsigned int regs[6] = {ctc->param1, ctc->param2, 0, 0, ctc->param3, ctc->sigmask};
	PanicInternalTTY (ctc->stacktop, PANIC_STACKTYPE_USER, ctc->sigpending, regs);
	Print ("Adjusted PC %x\n", ctc->stacktop[1] - ctc->pctrl->text);

	Print ("Region 1 AD %x\n", ctc->mpuregs[0]);
	Print ("Region 1 SE %x\n", ctc->mpuregs[1]);
	Print ("Region 1 AR %x\n", ctc->mpuregs[2]);

	Print ("Region 2 AD %x\n", ctc->mpuregs[3]);
	Print ("Region 2 SE %x\n", ctc->mpuregs[4]);
	Print ("Region 2 AR %x\n", ctc->mpuregs[5]);

	Print ("Region 3 AD %x\n", ctc->mpuregs[6]);
	Print ("Region 3 SE %x\n", ctc->mpuregs[7]);
	Print ("Region 3 AR %x\n", ctc->mpuregs[8]);

	Print ("Region 4 AD %x\n", ctc->mpuregs[9]);
	Print ("Region 4 SE %x\n", ctc->mpuregs[10]);
	Print ("Region 4 AR %x\n", ctc->mpuregs[11]);

	PanicPrintRegionSize (ctc->mpuregs[1]);
	PanicPrintRegionSize (ctc->mpuregs[4]);
	PanicPrintRegionSize (ctc->mpuregs[7]);
	PanicPrintRegionSize (ctc->mpuregs[10]);
#else
	(void)ctc;
#endif
}
