/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "sched/scheduler.h"
#include "arch/tms570ls1224/platform.h"
#include "mem/mem.h"
#include "print/print.h"
#include "mpu/pager.h"
#include "../config/live/platform.h"
#include "../test/inc/mem/mem.h"

///////////////////////////////////////////////////////////////////////////////
// Definitions
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// External variables
///////////////////////////////////////////////////////////////////////////////
extern unsigned int _dataend;
extern unsigned int _ramend;

///////////////////////////////////////////////////////////////////////////////
// Variables
///////////////////////////////////////////////////////////////////////////////
volatile unsigned int systemTick;

///////////////////////////////////////////////////////////////////////////////
// External functions
///////////////////////////////////////////////////////////////////////////////

// tickless.c
void TicklessClockSetup (void);

// vectortable.s
void UnknownHandler (void);

// scheduler-context.s
void SchedContext (void);

// tick-asm.s
void TickIrq (void);

// faulthandler.s
void GetFaultRegs (unsigned int *regs);

// tty-tms570-irqjump.s
void tty_asm_jump (void);

// can-tms570-irqjump.s
void can_asm_jump (void);
void can_asm_jump_ch1 (void);

///////////////////////////////////////////////////////////////////////////////
// Functions
///////////////////////////////////////////////////////////////////////////////
void PlatformInitTick (void)
{
	RTI->RTIGCTRL = 0;		// Disable all
	RTI->RTITBCTRL = 0;		// Default counter
	RTI->RTICOMPCTRL = 0;	// Default counter
	RTI->RTIUC0 = 0;		// Init counter
	RTI->RTIFRC0 = 0;		// Init counter
	RTI->RTICPUC0 = 1;		// Prescaler
	RTI->RTICOMP0 = 16000000 / 2 / 1000;
	RTI->RTIUDCP0 = 16000000 / 2 / 1000;
	RTI->RTIINTFLAG = 0x0007000F; // Clear IRQs
	RTI->RTICLEARINTENA = 0x00070F0F; // Disable IRQs
	RTI->RTISETINTENA = 1;	// Enable IRQ
	RTI->RTIGCTRL = 1;		// Enable counter
}

void PlatformIncrementTick (void)
{
	RTI->RTIINTFLAG = 1;	// Clear RTI interrupt
	systemTick++;
	SchedSystickHandler ();
}

void PlatformInitializePLL (void)
{
	// Disable both PLLs and wait for them to disable
	SYS->CSDIS = 0x42;
	while ((SYS->CSDIS & 0x42) != 0x42);

	// Clear PLL fault status register (under/over cycle slip, osc fail)
	SYS->GLBSTAT = 0x0301;

	// Configure PLL
	SYS->PLLCTL1 = (0 << 31)	// No reset on PLL slip
		| (0 << 29)				// No bypass on PLL slip
		| (31 << 24)			// PLL Output div (31 = div 32)
		| (0 << 23)				// Reset on oscillator fail disabled
		| (5 << 16)				// Reference clock divider (5 = div 6, f_intclk = f_oscin / 6)
		| (0x7700 << 0);		// PLL Multiplier (f_vcoclk = f_intclk * ((PLLMUL/256)+1), 120)

	SYS->PLLCTL2 = (0 << 31)	// No freq modulation
		| (0 << 22)				// Spreading rate
		| (0 << 12)				// Multiplier correction
		| (1 << 9)				// ODPLL Internal PLL Output divider (f_postodclk = f_vcoclk / 2)
		| (0 << 0);				// Spreading amount

	// Enable the PLL
	SYS->CSDIS = 2;
}

void PlatformInitializePins (void)
{
	IOMM->KICK[0] = 0x83E70B13;
	IOMM->KICK[1] = 0x95A4F1E0;

	IOMM->PINMMR[0] = (1 << 17) | (1 << 25);	// I2C

	IOMM->PINMMR[7] = 1 << 17;	// SCI
	IOMM->PINMMR[8] = 1 << 1;	// SCI

	IOMM->PINMMR[9] = 1 << 18;	// GIOB2

	IOMM->PINMMR[4] = 1 << 2;	// AENBL (EPWM2A)
	IOMM->PINMMR[27] = 1 << 2;	// BENBL (EPWM4A)
	IOMM->PINMMR[6] = 1 << 20;	// CENBL (EPWM7A)

	IOMM->PINMMR[37] = 1 << 16; // Enable clock to EPWM2
	IOMM->PINMMR[38] = (1 << 0) | (1 << 24); // EPWM4 and EPWM7
	IOMM->PINMMR[37] |= (1 << 1); // Enable the clock by sync

	IOMM->KICK[0] = 0;
	IOMM->KICK[1] = 0;
}

void PlatformInitializeClockSources (void)
{
	// Enable all clock domains
	SYS->CDDIS = 0;

	// Wait for clock sources to be ready
	while ((~SYS->CSDIS & 0x00FB) != (SYS->CSVSTAT & 0x00FB));

	// HCLK, VCLK and VCLK2 on wakeup, powerdown, and current: OSC OSC PLL1
	SYS->GHVSRC = (0 << 24) | (0 << 16) | (1 << 0);

	// RTI Clock source, default VCLK
	SYS->RCLKSRC = 0x01090109;

	// Peripheral Asynchronous Clock Source, default VCLK
	SYS->VCLKASRC = 0x00000909;

	// Set VCLK and VCLK2 ratios: div 2 and div 10
	SYS->CLKCNTL = (SYS->CLKCNTL & 0xF0FFFFFF) | (1 << 24);
	SYS->CLKCNTL = (SYS->CLKCNTL & 0xFFFFFFFF) | (9 << 16);

	// Set VBUS4 and VBUS3 ratios to the default
	SYS2->CLK2CNTRL = 0x00000101;

	// Set defaults in VCLKA4
	SYS2->VCLKACON1 = 0x01090109;

	// Increase PLL clock: div by 1
	SYS->PLLCTL1 = (SYS->PLLCTL1 & 0xE0FFFFFF) | (0 << 24);
}

void PlatformInitializeInterrupts (void)
{
	// Clear vim ram with default interrupt handler
	for (int i = 0; i < 128; i++)
		VIMRAM->ptr[i] = &UnknownHandler;

	// Now set some interrupts handlers
	VIMRAM->ptr[2+1] = &TickIrq;
	VIMRAM->ptr[21+1] = &SchedContext;
	VIMRAM->ptr[40+1] = &tty_asm_jump;
	VIMRAM->ptr[45+1] = &can_asm_jump;
	VIMRAM->ptr[0] = &can_asm_jump_ch1;

	// Set all intterupts to default priority
	VIM->FIRQPR[0] = 0x00000003;
	VIM->FIRQPR[1] = 0x00000000;
	VIM->FIRQPR[2] = 0x00000000;
	VIM->FIRQPR[3] = 0x00000000;

	// Disable all interrupts
	VIM->REQENACLR[0] = 0xFFFFFFFF;
	VIM->REQENACLR[1] = 0xFFFFFFFF;
	VIM->REQENACLR[2] = 0xFFFFFFFF;
	VIM->REQENACLR[3] = 0xFFFFFFFF;

	// Now enable the ones we need
	VIM->REQENASET[0] = 1 << 2; // RTI for tick
	VIM->REQENASET[0] = 1 << 21; // SSI for context switch
	VIM->REQENASET[1] = 1 << (40-32); // DMA Transfer complete
	VIM->REQENASET[1] = 1 << (45-32); // DCAN

	// Set both capture event sources to 0
	VIM->CAPEVT = 0;
}

void PlatformInitialize (void)
{
	// Configure main PLL
	PlatformInitializePLL ();

	// Disable peripherals
	SYS->CLKCNTL &= ~(1 << 8);

	// Power up peripherals
	PCR->PSPWRDWNCLR0 = 0xFFFFFFFF;
	PCR->PSPWRDWNCLR1 = 0xFFFFFFFF;
	PCR->PSPWRDWNCLR2 = 0xFFFFFFFF;
	PCR->PSPWRDWNCLR3 = 0xFFFFFFFF;

	// Enable peripherals
	SYS->CLKCNTL |= 1 << 8;

	// Setup pin muxing
	PlatformInitializePins ();

	// Flash control register
	FLASH->FRDCNTL = (3 << 8)	// RWAIT
		| (1 << 4)				// ASWSTEN
		| (1 << 0);				// ENPIPE

	// Set flash banks to active
	FLASH->FBFALLBACK = (3 << 14) | (3 << 2) | (3 << 0);

	// Set clock sources
	PlatformInitializeClockSources ();

	// Initialize IRQs
	PlatformInitializeInterrupts ();

	// Additional clock setup
	TicklessClockSetup ();

	// Initialize the pager
	PagerInitialize ();
}

void PlatformInitMemory (void)
{
	// Add internal memory
#if PLATFORM_USE_INTERNAL_MEMORY
	static struct MemBank bank;
	unsigned int bank0start = (unsigned int) &_dataend;
	unsigned int bank0end = (unsigned int) &_ramend;

	bank0start = (bank0start + 7) & ~3;
	unsigned int bank0size = bank0end - bank0start;

	MemBankInit (&bank, bank0start, bank0size);
	MemAddBank (&bank);

	Print ("Added %ib addr 0x%x\n", bank0size, bank0start);
#endif

	// Memory system test, compiles away if disabled
	PlatformTestMem ();
}

void PlatformStoreCrashRegs (struct ThreadControl *ctc)
{
	// TODO Find space for all 6 regs

	// Get registers
	unsigned int regs[6];
	GetFaultRegs (regs);

	// Re-use params and sigmask
	ctc->param1 = regs[0];
	ctc->param2 = regs[1];
	ctc->param3 = regs[4];
	ctc->sigmask = regs[5];
}

void PlatformMpuRegisterSetup (unsigned int *regs)
{
	unsigned int a[8] = {
		regs[0], regs[1], // (address), (permissions, flags, size)
		regs[2], regs[3],
		regs[4], regs[5],
		regs[6], regs[7]
	};

	regs[0] = a[0] & 0xFFFFFFE0;
	regs[3] = a[2] & 0xFFFFFFE0;
	regs[6] = a[4] & 0xFFFFFFE0;
	regs[9] = a[6] & 0xFFFFFFE0;

	regs[1] = a[1] & 0x0000FFFF;
	regs[4] = a[3] & 0x0000FFFF;
	regs[7] = a[5] & 0x0000FFFF;
	regs[10] = a[7] & 0x0000FFFF;

	regs[2] = (a[1] >> 16);
	regs[5] = (a[3] >> 16);
	regs[8] = (a[5] >> 16);
	regs[11] = (a[7] >> 16);
}

void PlatformPagerRegisterSetup (unsigned int *regs)
{
	regs[12] = regs[9];
	regs[13] = regs[10];
	regs[14] = regs[11];
	regs[15] = regs[9];
	regs[16] = regs[10];
	regs[17] = regs[11];
}
