/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

.syntax unified
.cpu cortex-r4
.arm

/*
Stack frame

Auto stacked in Cortex-M:
	CPSR
	PC/R15
	LR/R14
	R12
	R3
	R2
	R1
	R0
	*SP/R13 is not stacked, this is what the mrs r0, psp instruction is for.
	SP is loaded in available R0, stacked to then stored in the thc.

Manually stacked in Cortex-M:
	LR/R14???
	R11
	R10
	R9
	R8
	R7
	R6
	R5
	R4

Note: SPNA218.pdf from Ti has a good table illustrating the core regs (Table 2).

Cortex-R is banked. Instead of auto stacking, the PC is moved to the appropriate
LR for the new mode. The CPSR is also copied to the SPSR. The stack frame is then:
	R14/LR
	R13/SP
	R12
	R11
	R10
	R9
	R8
	R7
	R6
	R5
	R4
	R3
	R2
	R1
	R0
	PC
	CPSR
*/

.section .text.SchedContext
.global SchedContext
.type SchedContext, %function
SchedContext:
	cps #31						@ Supervisor mode: Access to everything but with user stack
	stmdb sp, {r0-lr}			@ Stack all the user regs except PC and CPSR, R0-LR is now free to use
	sub r0, sp, #60				@ Cannot have writeback in the above, sub 15 words and store in R0
	cps #18						@ Go back to IRQ mode, now have access to old PC and CPSR
	mrs r1, spsr				@ Get the CPSR from the banked reg
	stmdb r0!, {r1,lr}			@ Stack CPSR and PC

	ldr r1, =0xFFFFFFF4			@ Clear irq by reading SYS->SSIVEC
	ldr r2, [r1]

	ldr	r3, =currentthc			@ Load address of currentthc then currentthc itself
	ldr	r2, [r3]
	str r0, [r2]				@ Store the final stack pointer in the struct

.global SchedContextNext
SchedContextNext:				@ This entry is only used for crash recovery
	bl SchedSelectNext			@ Call scheduler selection function

	ldr	r3, =currentthc			@ r0 becomes the stacktop member of currentthc
	ldr r1, [r3]
	ldr r0, [r1]

	add r1, r1, #8				@ Point r1 to our data in ThreadControl
	ldm r1, {r4-r9}

	cmp r4, #0					@ If the next task does not use the MPU, avoid setting the regs
	beq skipmpu

	mov r3, #4
	mcr p15, 0, r3, c6, c2, 0	@ Set region
	mcr p15, 0, r4, c6, c1, 0	@ Set address
	mcr p15, 0, r6, c6, c1, 4	@ Set access rights
	mcr p15, 0, r5, c6, c1, 2	@ Set size and enable

	mov r3, #5
	mcr p15, 0, r3, c6, c2, 0	@ Set region
	mcr p15, 0, r7, c6, c1, 0	@ Set address
	mcr p15, 0, r9, c6, c1, 4	@ Set access rights
	mcr p15, 0, r8, c6, c1, 2	@ Set size and enable

	add r1, r1, #24
	ldm r1, {r4-r9}

	mov r3, #6
	mcr p15, 0, r3, c6, c2, 0	@ Set region
	mcr p15, 0, r4, c6, c1, 0	@ Set address
	mcr p15, 0, r6, c6, c1, 4	@ Set access rights
	mcr p15, 0, r5, c6, c1, 2	@ Set size and enable

	mov r3, #7
	mcr p15, 0, r3, c6, c2, 0	@ Set region
	mcr p15, 0, r7, c6, c1, 0	@ Set address
	mcr p15, 0, r9, c6, c1, 4	@ Set access rights
	mcr p15, 0, r8, c6, c1, 2	@ Set size and enable

	@ These are the two extra regions
	add r1, r1, #24
	ldm r1, {r4-r9}

	mov r3, #8
	mcr p15, 0, r3, c6, c2, 0	@ Set region
	mcr p15, 0, r4, c6, c1, 0	@ Set address
	mcr p15, 0, r6, c6, c1, 4	@ Set access rights
	mcr p15, 0, r5, c6, c1, 2	@ Set size and enable

	mov r3, #9
	mcr p15, 0, r3, c6, c2, 0	@ Set region
	mcr p15, 0, r7, c6, c1, 0	@ Set address
	mcr p15, 0, r9, c6, c1, 4	@ Set access rights
	mcr p15, 0, r8, c6, c1, 2	@ Set size and enable

skipmpu:

	// TODO FPU Registers

	ldmia r0!, {r1,lr}			@ Unstack CPSR and PC
	msr spsr_csxf, r1			@ Put CPSR back in IRQ SPSR. csxf restores all the flags, not just cf.
								@ See http://infocenter.arm.com/help/index.jsp?topic=/com.arm.doc.dui0489f/CIHFIDAJ.html
	cps #31						@ Switch to system mode
	ldmia r0, {r0-lr}			@ Unstack R0-LR
	cps #18						@ Switch back to irq mode
	dsb							@ Does this fix that memory access fault?
	subs pc, lr, #4				@ Return from IRQ
.size SchedContext, .-SchedContext
