/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

/*
 * For details on how the Real-Time Interrupt (RTI) module works, see section
 * 13 in the technical reference manual.
 *
 * During setup, systick is initialized like this:
 *  1. Counter 0 is disabled by clearing bit 1 in RTIGCTRL
 *  2. RTITBCTRL = 0
 *       Bit 1: RTIFRC0 will not be incremented on a failing external clock
 *       Bit 0: RTIUC0 clocks RTIFRC0
 *  3. RTICOMPCTRL = 0, all compare selects are set to counter 0
 *  4. RTIUC0 (RTI Up Counter 0 Register) is reset
 *  5. RTIFRC0 (RTI Free Running Counter 0 Register) is reset
 *  6. RTICPUC0 (RTI Compare Up Counter 0 Register) is set to 1
 *  7. RTICOMP0 (RTI Compare 0 Register) is set to
 *       (cpu clock / 2) / tick rate, essentially 80 MHz / 2 / 1000 Hz = 40000
 *  8. RTIUDCP0 (RTI Update Compare 0 Register) is set to
 *       (cpu clock / 2) / tick rate, essentially 80 MHz / 2 / 1000 Hz = 40000
 *  9. All interrupts are clearing by writing to RTIINTFLAG
 * 10. All interrupts are disabled by writing to RTICLEARINTENA
 * 11. Only compare interrupt 0 is enabled by writing to RTISETINTENA
 * 12. Counter 0 is enabled by writing 1 to RTIGCTRL
 *
 * With this configuration, the free running counter is updated at 40 MHz.
 * Divided by 40000 we get the 1 kHz timer tick.
 *
 * In order to implement tickless idle, we could briefly reconfigure the
 * timers by setting the compare value to some other multiple of 40k.
 * The value to program would then be 40k * ms. Programming the maximum
 * value possible (0xFFFFFFFF) makes the time 107 seconds.
 *
 * The tickless idle sleep function works by first disabling interrupts and
 * checking if it still wants to sleep. If not, interrupts are re-enabled
 * and it returns. If there are no tasks that want to be woken at a specific
 * time we may sleep indefinitely, so we just disable the systick and sleep.
 * When the CPU receives an interrupt in this state we just restart the
 * systick and enable interrupts again before returning.
 *
 * When we need to schedule a time to wake, things get more complicated. In
 * order to do this we reuse systick. First the new register value for the
 * timeout is calculated and the size is checked. The timer is then
 * re-initialized but with the new variable timeout and at compare module 1
 * instead of 0, just in case the interrupt handler is actually called. We
 * then enter sleep. When the timer overflows or any other interrupt is
 * pending, the processor wakes but does not execute any handlers as they are
 * disabled. Instead we disable the timer and calculate the elapsed time right
 * away. This value is then used to step the kernel, before we set the systick
 * back to normal operation, enable interrupts and return.
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "arch/tms570ls1224/platform.h"

///////////////////////////////////////////////////////////////////////////////
// Definitions
///////////////////////////////////////////////////////////////////////////////

// Constant used to calculate the timer register based on time in milliseconds.
//#define TIME_MULT_CONST 40000
#define TIME_MULT_CONST 8000

///////////////////////////////////////////////////////////////////////////////
// Externs
///////////////////////////////////////////////////////////////////////////////
extern void prvSetupTimerInterrupt (void);
extern void ClockHackSlow (void);
extern void ClockHackFast (void);

///////////////////////////////////////////////////////////////////////////////
// Functions
///////////////////////////////////////////////////////////////////////////////
/*
 * Various system settings has to be set for tickless idle to work properly.
 * These things are only done once on startup. Must be called from privileged
 * mode for write access to the registers.
 */
void TicklessClockSetup (void)
{
	// CAN Hardware clocked from OSC (16 MHz)
	SYS->VCLKASRC = 0;

	// Clock source for GCLK, HCLK, VCLK and VCLK2:
	// PLL1 after wake, during sleep and now. We set the clock properly in
	// assembly! Get the clocks into a known state
	SYS->GHVSRC = (1 << 24) | (1 << 16) | (1 << 0);
	SYS->CLKCNTL = (SYS->CLKCNTL & 0xFFF0FFFF) | (9 << 16);
	SYS->CLKCNTL = (SYS->CLKCNTL & 0xF0FFFFFF) | (1 << 24);

	// Flash bank power down setting: sleep mode with longest timeout
	FLASH->FBAC = 0x0000FF0F;
	FLASH->FBFALLBACK = 0x050503FF0;

	// Disable unused peripherals
	PCR->PSPWRDWNSET0 = 0x00FFF000; // PS7 - PS0: Only unassigned quadrants
	PCR->PSPWRDWNSET1 = 0xFFFFF0F0; // PS15 - PS8: Disable DCAN1, DCAN2, FlexRay, MIBADCx
	PCR->PSPWRDWNSET2 = 0xFFFFFFF0; // PS23 - PS16: Disable N2HETx, HTUx and FTU
}

// TODO Implement tickless idle
