/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

// mem32 FFF7DC00 4F

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "mem/string.h"
#include "posix/errno.h"
#include "fs/dev/dev.h"
#include "ioctl/ioctl.h"
#include "drivers/can/can-tms570.h"
#include "sched/scheduler.h"
#include "sched/scheduler-signal.h"
#include "posix/sys/signal.h"

///////////////////////////////////////////////////////////////////////////////
// Definitions
///////////////////////////////////////////////////////////////////////////////
#define N_CAN_MODULES 3
#define RX_MAILBOX_FIRST 2
#define RX_MAILBOX_LAST 16

///////////////////////////////////////////////////////////////////////////////
// Variables
///////////////////////////////////////////////////////////////////////////////

// These are used by the IRQ handlers to get back to the data
struct can *can_irq_instances[N_CAN_MODULES] = {0};

struct can_fifo
{
	struct can_frame frames[16]; // 16*16 bytes = 256
	short rdpos;
	short wrpos;
};

struct can_fifo cfifo;

///////////////////////////////////////////////////////////////////////////////
// Functions
///////////////////////////////////////////////////////////////////////////////
void dev_can_init_mailboxes (_CAN *t)
{
	// Wait timer
	int timer = 1000000;

	// Set command register for transmit with IF1
	while (timer && t->IF1STAT & 0x80)
		timer--;

	t->IF1CMD = 0x87;
	t->IF1MSK = 0xFFFFFFFF;

	// Initialize some mailboxes for a FIFO
	for (int i = RX_MAILBOX_FIRST; i <= RX_MAILBOX_LAST; i++)
	{
		// Wait for busy to go low
		while (timer && t->IF2STAT & 0x80)
			timer--;

		// Want to write Mask, Arb, Control
		t->IF2CMD = 0xF0;

		// Message object valid, 29-bit ID, receive
		t->IF2ARB = 0xC0000000;

		// Use mask, RxIE, EoB, DLC
		t->IF2MCTL = 0x00001408;

		// If this is the last box in the FIFO, set EoB
		if (i == RX_MAILBOX_LAST)
			t->IF2MCTL |= 0x00000080;

		// Clear mask
		t->IF2MSK = 0;

		// Initiate transfer
		t->IF2NO = i;
	}
}

int dev_can_init (struct can *t)
{
	// Static array for hardware instances
	static const _CAN *instances[] = {CAN1, CAN2, CAN3};

	// Set the correct instance
	t->instance = (_CAN*) instances[t->num];

	// Wait timer
	int timer = 1000000;

	// Write can control register, parity disabled, test mode, config enable
	// Test mode: 0x000000C1
	// Normal mode: 0x00000041
	// Bit 5: Disable automatic retransmission
	t->instance->CTL = 0x00000041 | (5 << 10) | (1 << 5);

	// Wait for init bit to be set
	while (timer && !(t->instance->CTL & 0x00000040))
		timer--;

	// Timeout
	t->instance->ABOTR = 0;

	// Initialize mailboxes
	dev_can_init_mailboxes (t->instance);
	t->rxindex = RX_MAILBOX_FIRST;

	// Set timing values
#if 0
	t->instance->BTR = (2 << 12)	// TSeg2
		| (3 << 8)					// TSeg1
		| (2 << 6)					// Sync Jump Width
		| 0x1F;						// Baud Rate Prescaler
#else
	// 1 Mbaud
	t->instance->BTR = (2 << 12)	// TSeg2
		| (3 << 8)					// TSeg1
		| (0 << 6)					// Sync Jump Width
		| 1;						// Baud Rate Prescaler
#endif

	// Clear error and status register
	t->instance->ES = 0xFFFFFFFF;

	// Test register: External loop-back if 0x00000100
	t->instance->TEST = 0x00000100;

	// Set interrupt levels to 0 for mailbox 1 - 64
	t->instance->INTMUX12 = 0;
	t->instance->INTMUX34 = 0;

	// Receive interrupt setup (IE0)
	t->instance->CTL |= 1 << 1;
	t->instance->CTL &= ~(1 << 16);

	// TX Pin: No pull up/down, push-pull mode
	t->instance->TIOC = 0x0002000E;

	// Rx Pin: Pull up
	t->instance->RIOC = 0x00040008;

	// Out of init and config mode
	t->instance->CTL &= ~0x00000041;

	// Wait for peripheral to go live
	while (timer && t->instance->CTL & 1)
		timer--;

	// Return the timer value
	return timer;
}

void dev_can_read_data (struct can *t, void *data)
{
	// Read data in the correct order
	volatile unsigned char *wraddr = (unsigned char*) &t->instance->IF2DATA;
	((unsigned char*)data)[11] = wraddr[0];
	((unsigned char*)data)[10] = wraddr[1];
	((unsigned char*)data)[9] = wraddr[2];
	((unsigned char*)data)[8] = wraddr[3];
	((unsigned char*)data)[15] = wraddr[4];
	((unsigned char*)data)[14] = wraddr[5];
	((unsigned char*)data)[13] = wraddr[6];
	((unsigned char*)data)[12] = wraddr[7];

	// Read can id and length
	struct can_frame *cf = (struct can_frame*) data;
	cf->can_id = (t->instance->IF2ARB & 0x1FFFFFFF) | CAN_EFF_FLAG;
	cf->can_dlc = t->instance->IF2MCTL & 0x0F;
}

void dev_can_read_fetch (struct can *t)
{
	// Wait timer
	int timer = 1000000;

	// Receive mailbox setup on IF2, wait for ready
	while (timer && t->instance->IF2STAT & 0x80)
		timer--;

	// Request read of Arb, Control, ClrIntPnt, TxRqst/NewDat, Data A and Data B
	t->instance->IF2CMD = 0x3F;

	// Set mailbox number, valid range 1 - 64, this triggers the transfer
	t->instance->IF2NO = t->rxindex;

	// Wait for data
	while (timer && t->instance->IF2STAT & 0x80)
		timer--;

	// Debug: Check MsgLst (message lost)
	//if (t->instance->IF2MCTL & (1 << 14))
	//	Print ("Message lost\n");
}

/*
 * Interrupt handler that dumps the entire hardware FIFO into
 * our software FIFO as per Figure 27-11 in the manual (page
 * 1496).
 *
 * Hardware FIFO cannot be used directly because it is not
 * really a FIFO, it always starts writing in the first
 * available physical position and not the logical one.
 *
 * As a consequence, once the first FIFO message has been read
 * it must read them all or we will risk messing up the order
 * of messages. The messages may then be read into our true FIFO.
 */
void dev_can_irqhandler (void)
{
	// Get the instance pointer
	struct can *t = can_irq_instances[2];

	// Restart from here (as per figure)
	int index;
dev_can_irqhandler_restart:

	// Get the index of the mailbox that caused the interrupt
	index = t->instance->INT & 0x0000FFFF;

	// Only care if related to some message box
	if (index >= 1 && index <= 0x80)
	{
		// Inner FIFO loop
		while (1)
		{
			// Attempt sending the signal to the process
			SchedPostSignal (t->trc, SIGURG);

			// Get message object
			t->rxindex = index;
			dev_can_read_fetch (t);

			// Check NewDat
			if (!(t->instance->IF2MCTL & (1 << 15)))
				goto dev_can_irqhandler_restart;

			// Read data
			dev_can_read_data (t, &cfifo.frames[cfifo.wrpos]);
			cfifo.wrpos++;

			if (cfifo.wrpos == 16)
				cfifo.wrpos = 0;

			// Check EoB
			if (t->instance->IF2MCTL & (1 << 7))
				goto dev_can_irqhandler_restart;

			index++;
		}
	}
}

void dev_can_irqhandler_ch1 (void)
{
	//Print ("dev_can_irqhandler_ch1 phantom!\n");
}

///////////////////////////////////////////////////////////////////////////////
// Private functions
///////////////////////////////////////////////////////////////////////////////
int dev_can_open (struct FileNode *node, const char *filename)
{
	// Get number
	unsigned int num = filename[3] - '0';

	// Check it
	if (num >= N_CAN_MODULES)
		return -ENXIO;

	// Already initialized?
	if (can_irq_instances[num])
		return -EBUSY;

	// Allocate (static for now)
	static struct can sp[N_CAN_MODULES];
	struct can *t = &sp[num];

	// Clear the structure
	MemSet (t, 0, sizeof (struct can));

	// Set the irq-handlers
	can_irq_instances[num] = t;

	// Set the file node
	node->priv = (void*) t;

	// Set the number
	t->num = num;

	// Set thread reference
	t->trc = currentthc;

	// Initialize it, checking for error
	if (!dev_can_init (t))
	{
		can_irq_instances[num] = 0;
		return -ETIME;
	}

	// Done
	return 0;
}

int dev_can_close (struct FileNode *node)
{
	// Extract pointer
	struct can *t = (struct can*) node->priv;

	// Disable IRQ
	t->instance->CTL &= ~(1 << 1);

	// Clear allocation
	can_irq_instances[t->num] = 0;

	// Done
	return 0;
}

int dev_can_read (struct FileNode *node, void *data, unsigned int nbytes)
{
	// Number of bytes must be the size of a can frame and aligned
	if (nbytes != 16 || (int)data & 0x00000003)
		return -EINVAL;

	// Extract pointer
	//struct can *t = (struct can*) node->priv;

	// Dequeue message item
	if (cfifo.rdpos != cfifo.wrpos)
	{
		struct can_frame *cf = (struct can_frame*) data;

		MemCpy (cf, &cfifo.frames[cfifo.rdpos], sizeof (struct can_frame));

		cfifo.rdpos++;

		if (cfifo.rdpos == 16)
			cfifo.rdpos = 0;
	}
	else
		return 0;

	// Done
	return nbytes;
}

int dev_can_write (struct FileNode *node, void *data, unsigned int nbytes)
{
	// Number of bytes must be the size of a can frame and aligned
	if (nbytes != 16 || (int)data & 0x00000003)
		return -EINVAL;

	// Extract pointer
	struct can *t = (struct can*) node->priv;

	// Do not continue if busy
	if (t->instance->TXRQ12 & 0x00000001)
		return -ENOBUFS;

	// Wait timer
	int timer = 1000000;

	// Transmit mailbox setup on IF1, wait for ready
	while (timer && t->instance->IF1STAT & 0x80)
		timer--;

	// Check timer
	if (!timer)
		return -ETIME;

	// Write, Transfer data and set TxRqst
	t->instance->IF1CMD = 0xF7;

	// Write data, can do this since we force alignment
	volatile unsigned char *wraddr = (unsigned char*) &t->instance->IF1DATA;
	wraddr[0] = ((unsigned char*)data)[11];
	wraddr[1] = ((unsigned char*)data)[10];
	wraddr[2] = ((unsigned char*)data)[9];
	wraddr[3] = ((unsigned char*)data)[8];
	wraddr[4] = ((unsigned char*)data)[15];
	wraddr[5] = ((unsigned char*)data)[14];
	wraddr[6] = ((unsigned char*)data)[13];
	wraddr[7] = ((unsigned char*)data)[12];

	// Cast socket struct
	struct can_frame *cf = (struct can_frame*) data;

	// Set message control register
	t->instance->IF1MCTL = 0x00000180 | (cf->can_dlc & 0x0F);

	// Set message identifier
	t->instance->IF1ARB = 0xE0000000 | (cf->can_id & 0x1FFFFFFF);

	// Set mailbox number, valid range 1 - 64, this triggers the transfer
	t->instance->IF1NO = 1;

	// Done
	return nbytes;
}

int dev_can_ioctl (struct FileNode *node, unsigned int code, void *data)
{
	// Extract pointer
	struct can *t = (struct can*) node->priv;
	(void)t;
	(void)data;

	// Switch on the ioctl control code
	switch (code)
	{
	default:
		// Unknown ioctl: Return "Inappropriate ioctl for device"
		return -ENOTTY;
	}

	// Default return value
	return 0;
}

/*
MODULE_DESCRIPTOR (can0,
	dev_can_open,
	dev_can_close,
	dev_can_read,
	dev_can_write,
	dev_can_ioctl,
	"can0");

MODULE_DESCRIPTOR (can1,
	dev_can_open,
	dev_can_close,
	dev_can_read,
	dev_can_write,
	dev_can_ioctl,
	"can1");
*/

MODULE_DESCRIPTOR (can2,
	dev_can_open,
	dev_can_close,
	dev_can_read,
	dev_can_write,
	dev_can_ioctl,
	"can2");
