/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "mem/string.h"
#include "posix/errno.h"
#include "fs/dev/dev.h"
#include "ioctl/ioctl.h"
#include "drivers/i2c/i2c-tms570.h"

///////////////////////////////////////////////////////////////////////////////
// Definitions
///////////////////////////////////////////////////////////////////////////////
#define N_I2C_MODULES 1

///////////////////////////////////////////////////////////////////////////////
// Variables
///////////////////////////////////////////////////////////////////////////////

// These are used by the IRQ handlers to get back to the data
struct i2c *i2c_irq_instances[N_I2C_MODULES] = {0};

///////////////////////////////////////////////////////////////////////////////
// Functions
///////////////////////////////////////////////////////////////////////////////
void dev_i2c_init (struct i2c *t)
{
	// Peripheral into reset
	I2C1->MDR = 0;

	// Disable all interrupts
	I2C1->IMR = 0;

	// Set prescaler, clock is 8 MHz to the I2C
	I2C1->PSC = 1;

	// Set low and high time, 4 for 400 KHz, 20 for 154 KHz
	I2C1->CKL = 20;
	I2C1->CKH = 20;

	// Pin function
	I2C1->PFNC = 0;

	// DMA Use
	I2C1->DMACR = 0;

	// Master mode and generate stop bit, transmitter
	I2C1->MDR = (1 << 11) | (1 << 10) | (1 << 9);

	// Enable
	I2C1->MDR |= (1 << 5);
}

///////////////////////////////////////////////////////////////////////////////
// Private functions
///////////////////////////////////////////////////////////////////////////////
int dev_i2c_open (struct FileNode *node, const char *filename)
{
	// Get number
	unsigned int num = 0;

	// Check it
	if (num >= N_I2C_MODULES)
		return -ENXIO;

	// Allocation
	struct i2c *t = i2c_irq_instances[num];

	// Already initialized?
	if (t)
		t->refcount++;
	else
	{
		// Allocate (static for now)
		static struct i2c sp[N_I2C_MODULES];
		t = &sp[num];

		// Set the irq-handlers
		i2c_irq_instances[num] = t;

		// Clear the structure
		MemSet (t, 0, sizeof (struct i2c));

		// Set the number
		t->num = num;
		t->refcount = 1;

		// Initialize it
		dev_i2c_init (t);
	}

	// Set the file node
	node->priv = (void*) t;

	// Done
	return 0;
}

int dev_i2c_close (struct FileNode *node)
{
	// Extract pointer
	struct i2c *t = (struct i2c*) node->priv;

	// Clear allocation
	t->refcount--;

	if (!t->refcount)
		i2c_irq_instances[t->num] = 0;

	// Done
	return 0;
}

int dev_i2c_read (struct FileNode *node, void *data, unsigned int nbytes)
{
	// Set slave address
	I2C1->SAR = ((unsigned char*)data)[0];

	// Set data count
	I2C1->CNT = nbytes - 1;

	// Set receive, master mode
	I2C1->MDR &= ~(1 << 9);
	I2C1->MDR |= 1 << 10;

	// Set stop
	I2C1->MDR |= 1 << 11;

	// Set start
	I2C1->MDR |= 1 << 13;

	// Wait timer
	int timer = 1000000;

	// Receive data
	for (int i = 1; i < nbytes; i++)
	{
		// Wait for peripheral ready
		timer = 1000000;

		while (timer && !(I2C1->STR & (1 << 3)))
			timer--;

		if (!timer)
			break;

		// Send data
		((unsigned char*)data)[i] = I2C1->DRR;
	}

	// Wait for completion
	while (timer && I2C1->STR & (1 << 12))
		timer--;

	// Wait for stop
	while (timer && !(I2C1->STR & (1 << 5)))
		timer--;

	// Clear stop flag
	I2C1->STR = 1 << 5;

	// Check for timeout
	if (!timer)
	{
		I2C1->MDR &= ~(1 << 5);
		I2C1->MDR |= 1 << 5;
		return -ETIME;
	}

	// Done
	return nbytes;
}

int dev_i2c_write (struct FileNode *node, void *data, unsigned int nbytes)
{
	// NOTE: To debug the I2C:
	// mem32 FFF7D400, 20

	// Set slave address
	I2C1->SAR = ((unsigned char*)data)[0];

	// Set data count
	I2C1->CNT = nbytes - 1;

	// Set direction out, master mode
	I2C1->MDR |= (1 << 9) | (1 << 10);

	// Set stop
	I2C1->MDR |= 1 << 11;

	// Set start
	I2C1->MDR |= 1 << 13;

	// Wait timer
	int timer = 1000000;

	// Send data
	for (int i = 1; i < nbytes; i++)
	{
		// Wait for peripheral ready
		timer = 1000000;

		while (timer && !(I2C1->STR & (1 << 4)))
			timer--;

		if (!timer)
			break;

		// Send data
		I2C1->DXR = ((unsigned char*)data)[i];
	}

	// Wait for completion
	while (timer && I2C1->STR & (1 << 12))
		timer--;

	// Wait for stop
	while (timer && !(I2C1->STR & (1 << 5)))
		timer--;

	// Clear stop flag
	I2C1->STR = 1 << 5;

	// Check for timeout
	if (!timer)
	{
		I2C1->MDR &= ~(1 << 5);
		I2C1->MDR |= 1 << 5;
		return -ETIME;
	}

	// Done
	return nbytes;
}

int dev_i2c_ioctl (struct FileNode *node, unsigned int code, void *data)
{
	// Extract pointer
	struct i2c *t = (struct i2c*) node->priv;
	(void)t;

	// Switch on the ioctl control code
	switch (code)
	{
	default:
		// Unknown ioctl: Return "Inappropriate ioctl for device"
		return -ENOTTY;
	}

	// Default return value
	return 0;
}

MODULE_DESCRIPTOR (i2c,
	dev_i2c_open,
	dev_i2c_close,
	dev_i2c_read,
	dev_i2c_write,
	dev_i2c_ioctl,
	"i2c");
