/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "mem/string.h"
#include "posix/errno.h"
#include "fs/dev/dev.h"
#include "ioctl/ioctl.h"
#include "drivers/dma/dma-tms570.h"
#include "arch/tms570ls1224/epwm.h"
#include "arch/tms570ls1224/gio.h"

///////////////////////////////////////////////////////////////////////////////
// Definitions
///////////////////////////////////////////////////////////////////////////////

// Register values to program
#define TB_COUNT_UP (0 << 0)	// Asymmetrical mode
#define TB_DISABLE (0 << 2)		// Phase loading disable
#define TB_SHADOW (0 << 3)
#define TB_SYNC_DISABLE (3 << 4)

#define SHDWAMODE_SHADOW (0 << 4)
#define SHDWBMODE_SHADOW (0 << 6)
#define LOADAMODE_CTR_ZERO (0 << 0)
#define LOADBMODE_CTR_ZERO (0 << 2)

#define PRD_AQ_CLEAR (1 << 2)
#define CAU_AQ_SET (2 << 4)

// Pin definitions
#define PIN_APHASE 0
#define PIN_BPHASE 1
#define PIN_CPHASE 2
#define PIN_DRVSLEEP 3

// GIO To use
#define GIO_N2HET1 ((_GIOAB*) 0xFFF7B84C)

// IOCTL Values
#define PWM_SETPERIODA		_IOW('p', 0, unsigned short)
#define PWM_SETPERIODB		_IOW('p', 1, unsigned short)
#define PWM_SETPERIODC		_IOW('p', 2, unsigned short)
#define PWM_SETCMPA			_IOW('p', 3, unsigned short)
#define PWM_SETCMPB			_IOW('p', 4, unsigned short)
#define PWM_SETCMPC			_IOW('p', 5, unsigned short)
#define PWM_SETPHASEA		_IOW('p', 6, unsigned short)
#define PWM_SETPHASEB		_IOW('p', 7, unsigned short)
#define PWM_SETPHASEC		_IOW('p', 8, unsigned short)
#define PWM_SETDRVSLEEP		_IOW('p', 9, unsigned short)

///////////////////////////////////////////////////////////////////////////////
// Structures
///////////////////////////////////////////////////////////////////////////////
struct pwm
{
	_EPWM *instance;
};

struct pwmControlPacket
{
	unsigned short period[3];
	unsigned short cmp[3];
	unsigned char pinvalues[4];
};

///////////////////////////////////////////////////////////////////////////////
// Private init functions
///////////////////////////////////////////////////////////////////////////////
void dev_pwm_initpin (_GIOAB *g, int pin)
{
	g->DIR |= 1 << pin;
	g->DSET = 1 << pin;
	g->PDR &= ~(1 << pin);
	g->PULDIS = 1 << pin;
}

void dev_pwm_setpin (int pin, int state)
{
	unsigned int value = 0;

	switch (pin)
	{
	case PIN_APHASE:
		value = (1 << 0);
		break;
	case PIN_BPHASE:
		value = (1 << 2);
		break;
	case PIN_CPHASE:
		value = (1 << 14);
		break;
	case PIN_DRVSLEEP:
		value = (1 << 5);
		break;
	default:
		break;
	}

	if (state)
		GIO_N2HET1->DSET = value;
	else
		GIO_N2HET1->DCLR = value;
}

void dev_pwm_init (struct pwm *s)
{
	// A
	EPWM2->TBPRD = 1200;	// Period = 1201 TBCLK counts
	EPWM2->TBPHS = 0;
	EPWM2->TBCTL = TB_COUNT_UP | TB_DISABLE | TB_SHADOW | TB_SYNC_DISABLE;

	EPWM2->CMPCTL = SHDWAMODE_SHADOW | SHDWBMODE_SHADOW | LOADAMODE_CTR_ZERO | LOADBMODE_CTR_ZERO;
	EPWM2->AQCTLA = PRD_AQ_CLEAR | CAU_AQ_SET;

	EPWM2->CMPA = 600;

	// B
	EPWM4->TBPRD = 1200;	// Period = 1201 TBCLK counts
	EPWM4->TBPHS = 0;
	EPWM4->TBCTL = TB_COUNT_UP | TB_DISABLE | TB_SHADOW | TB_SYNC_DISABLE;

	EPWM4->CMPCTL = SHDWAMODE_SHADOW | SHDWBMODE_SHADOW | LOADAMODE_CTR_ZERO | LOADBMODE_CTR_ZERO;
	EPWM4->AQCTLA = PRD_AQ_CLEAR | CAU_AQ_SET;

	EPWM4->CMPA = 200;

	// C
	EPWM7->TBPRD = 1200;	// Period = 1201 TBCLK counts
	EPWM7->TBPHS = 0;
	EPWM7->TBCTL = TB_COUNT_UP | TB_DISABLE | TB_SHADOW | TB_SYNC_DISABLE;

	EPWM7->CMPCTL = SHDWAMODE_SHADOW | SHDWBMODE_SHADOW | LOADAMODE_CTR_ZERO | LOADBMODE_CTR_ZERO;
	EPWM7->AQCTLA = PRD_AQ_CLEAR | CAU_AQ_SET;

	EPWM7->CMPA = 1000;

	// Pins
	dev_pwm_initpin (GIO_N2HET1, 0); // APHASE
	dev_pwm_initpin (GIO_N2HET1, 2); // BPHASE
	dev_pwm_initpin (GIO_N2HET1, 14); // CPHASE
	dev_pwm_initpin (GIO_N2HET1, 5); // DRVSLEEP

	dev_pwm_setpin (PIN_APHASE, 0);
	dev_pwm_setpin (PIN_BPHASE, 0);
	dev_pwm_setpin (PIN_CPHASE, 0);
	dev_pwm_setpin (PIN_DRVSLEEP, 1);
}

///////////////////////////////////////////////////////////////////////////////
// File node functions
///////////////////////////////////////////////////////////////////////////////
int dev_pwm_open (struct FileNode *node, const char *filename)
{
	// Get number
	unsigned int num = filename[3] - '0';

	// Check it
	if (num >= 1)
		return -ENXIO;

	// Initialize
	dev_pwm_init (0);

	// Done
	return 0;
}

int dev_pwm_close (struct FileNode *node)
{
	return 0;
}

int dev_pwm_read (struct FileNode *node, void *data, unsigned int nbytes)
{
	return -1;
}

int dev_pwm_write (struct FileNode *node, void *data, unsigned int nbytes)
{
	if (nbytes != sizeof (struct pwmControlPacket))
		return -EINVAL;

	struct pwmControlPacket *p = (struct pwmControlPacket*) data;

	EPWM2->TBPRD = p->period[0];
	EPWM2->CMPA = p->cmp[0];

	EPWM4->TBPRD = p->period[1];
	EPWM4->CMPA = p->cmp[1];

	EPWM7->TBPRD = p->period[2];
	EPWM7->CMPA = p->cmp[2];

	dev_pwm_setpin (PIN_APHASE, (int) p->pinvalues[0]);
	dev_pwm_setpin (PIN_BPHASE, (int) p->pinvalues[1]);
	dev_pwm_setpin (PIN_CPHASE, (int) p->pinvalues[2]);
	dev_pwm_setpin (PIN_DRVSLEEP, (int) p->pinvalues[3]);

	return nbytes;
}

int dev_pwm_ioctl (struct FileNode *node, unsigned int code, void *data)
{
	// Switch on the ioctl control code
	switch (code)
	{
	case PWM_SETPERIODA:
		EPWM2->TBPRD = (unsigned short) ((int) data);
		break;
	case PWM_SETPERIODB:
		EPWM4->TBPRD = (unsigned short) ((int) data);
		break;
	case PWM_SETPERIODC:
		EPWM7->TBPRD = (unsigned short) ((int) data);
		break;
	case PWM_SETCMPA:
		EPWM2->CMPA = (unsigned short) ((int) data);
		break;
	case PWM_SETCMPB:
		EPWM4->CMPA = (unsigned short) ((int) data);
		break;
	case PWM_SETCMPC:
		EPWM7->CMPA = (unsigned short) ((int) data);
		break;
	case PWM_SETPHASEA:
		dev_pwm_setpin (PIN_APHASE, (int) data);
		break;
	case PWM_SETPHASEB:
		dev_pwm_setpin (PIN_BPHASE, (int) data);
		break;
	case PWM_SETPHASEC:
		dev_pwm_setpin (PIN_CPHASE, (int) data);
		break;
	case PWM_SETDRVSLEEP:
		dev_pwm_setpin (PIN_DRVSLEEP, (int) data);
		break;
	default:
		// Unknown ioctl: Return "Inappropriate ioctl for device"
		return -ENOTTY;
	}

	return 0;
}

MODULE_DESCRIPTOR (pwm0,
	dev_pwm_open,
	dev_pwm_close,
	dev_pwm_read,
	dev_pwm_write,
	dev_pwm_ioctl,
	"pwm0");
