/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "fs/vfs.h"
#include "fs/dev/dev.h"
#include "posix/errno.h"
#include "mem/string.h"
#include "print/print.h"
#include "ioctl/ioctl.h"

///////////////////////////////////////////////////////////////////////////////
// Variables
///////////////////////////////////////////////////////////////////////////////
unsigned int ramdisk_data[1024*1/4];
unsigned int ramdisk_open = 0;
unsigned int ramdisk_pos;

///////////////////////////////////////////////////////////////////////////////
// Driver functions
///////////////////////////////////////////////////////////////////////////////
int dev_ramdisk_open (struct FileNode *node, const char *filename)
{
	if (ramdisk_open)
		return -EBUSY;

	MemSet (ramdisk_data, 0, sizeof (ramdisk_data));
	ramdisk_pos = 0;
	return 0;
}

int dev_ramdisk_close (struct FileNode *node)
{
	ramdisk_open = 0;
	return 0;
}

int dev_ramdisk_read (struct FileNode *node, void *data, unsigned int nbytes)
{
	unsigned char *p = (unsigned char*) ramdisk_data;

	if (ramdisk_pos > sizeof(ramdisk_data) || (ramdisk_pos + nbytes) > sizeof(ramdisk_data))
	{
		Print ("Indexing error on read\n");
		return 0;
	}

	MemCpy (data, p + ramdisk_pos, nbytes);
	return nbytes;
}

int dev_ramdisk_write (struct FileNode *node, void *data, unsigned int nbytes)
{
	unsigned char *p = (unsigned char*) ramdisk_data;

	if (ramdisk_pos > sizeof(ramdisk_data) || (ramdisk_pos + nbytes) > sizeof(ramdisk_data))
	{
		Print ("Indexing error on write\n");
		return nbytes;
	}

	MemCpy (p + ramdisk_pos, data, nbytes);
	return nbytes;
}

int dev_ramdisk_ioctl (struct FileNode *node, unsigned int code, void *data)
{
	switch (code)
	{
	case 0:
		return 0;
	case 1: // GET_SECTOR_COUNT 1
		*(unsigned int*)data = sizeof (ramdisk_data) / 512;
		return 0;
	case 2: // GET_SECTOR_SIZE 2
		*(unsigned short*)data = 512;
		return 0;
	case 3: // GET_BLOCK_SIZE 3
		*(unsigned int*)data = 1;
		return 0;
	case SPIFSETSECT:
		ramdisk_pos = (unsigned int) data * 512;
		return 0;
	default:
		// Unknown ioctl: Return "Inappropriate ioctl for device"
		return -ENOTTY;
	}

	// Default return value
	return 0;
}

///////////////////////////////////////////////////////////////////////////////
// Driver declarations
///////////////////////////////////////////////////////////////////////////////
MODULE_DESCRIPTOR (ramdisk,
	dev_ramdisk_open,
	dev_ramdisk_close,
	dev_ramdisk_read,
	dev_ramdisk_write,
	dev_ramdisk_ioctl,
	"ramdisk");
