/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

/*
 * Logical SPI1: Physical SPI5
 *   SCK on pin 100 (MIBSPI5CLK)
 *   MOSI on pin 99 (MIBSPI5SIMO)
 *   MISO on pin 98 (MIBSPI5SOMI)
 *   CS0 on pin 142 (GIOB2)			U4/U5, top-left SPINOR (default main boot)
 *   CS1 on pin 141 (N2HET1[20])	U6/U7, top-right SPINOR
 *   CS2 on pin 140 (N2HET1[18])	U8, bottom MBR85RS2MTA memory chip
 *   CS3 on pin 139 (N2HET1[16])	XS1, bottom SD-card socket
 *
 * Logical SPI2: Board design error
 *   SCK on pin 25 (N2HET1[0])		Board design error
 *   MOSI on pin 30 (N2HET1[2])		Board design error
 *   MISO on pin 31 (N2HET1[5])		Board design error
 *   CS0 on pin 127 (N2HET1[30])	U11/U12, bottom-right SPINOR
 *   CS1 on pin 126 (GIOB[0])		U9/U10, bottom-left SPINOR
 *   CS2 on pin 125 (N2HET1[14])	U13, top MBR85RS2MTA
 *   CS3 on pin 124 (N2HET1[12])	XS2, top SD-card socket
 *
 * Logical SPI3: Physical SPI3
 *   SCK on pin 53 (MIBSPI3CLK)
 *   MOSI on pin 52 (MIBSPI3SIMO)
 *   MISO on pin 51 (MIBSPI3SOMI)
 *   CS0 on pin 55 (MIBSPI3NCS[0])	U24, top sensor, CS_A/G
 *   CS1 on pin 54 (MIBSPI3NENA)	U24, top sensor, CS_M
 *   CS2 on pin not connected
 *   CS3 on pin not connected
 *
 * Logical SPI4: Physical SPI1
 *   SCK on pin 95 (MIBSPI1CLK)
 *   MOSI on pin 93 (MIBSPI1SIMO)
 *   MISO on pin 94 (MIBSPI1SOMI)
 *   CS0 on pin 92 (N2HET1[26])		U30, bottom sensor, CS_A/G
 *   CS1 on pin 105 (MIBSPI1NCS[0])	U30, bottom sensor, CS_M
 *   CS2 on pin 96 (MIBSPI1NENA)	P10, exposed pads, CS
 *   CS3 on pin not connected
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "mem/string.h"
#include "posix/errno.h"
#include "fs/dev/dev.h"
#include "ioctl/ioctl.h"
#include "drivers/dma/dma-tms570.h"
#include "drivers/spi/spi-tms570.h"
#include "arch/tms570ls1224/gio.h"

///////////////////////////////////////////////////////////////////////////////
// Definitions
///////////////////////////////////////////////////////////////////////////////
#define N_SPI_MODULES 4

///////////////////////////////////////////////////////////////////////////////
// Variables
///////////////////////////////////////////////////////////////////////////////

// These are used by the IRQ handlers to get back to the data
struct spi *spi_irq_instances[N_SPI_MODULES] = {0};

///////////////////////////////////////////////////////////////////////////////
// Chip select functions
///////////////////////////////////////////////////////////////////////////////
#define GIO_N2HET1 ((_GIOAB*) 0xFFF7B84C)

void dev_spi_initpin (_GIOAB *g, int pin)
{
	g->DIR |= 1 << pin;
	g->DSET = 1 << pin;
	g->PDR &= ~(1 << pin);
	g->PULDIS = 1 << pin;
}

void dev_spi_initgio_1 (void)
{
	GIO->GCR0 = 1;
	dev_spi_initpin (&GIO->b, 2); // CS0
	dev_spi_initpin (GIO_N2HET1, 20); // CS1
	dev_spi_initpin (GIO_N2HET1, 18); // CS2
	dev_spi_initpin (GIO_N2HET1, 16); // CS3
}

void dev_spi_initgio_2 (void)
{
	dev_spi_initpin (GIO_N2HET1, 30); // CS0
	dev_spi_initpin (&GIO->b, 0); // CS1
	//dev_spi_initpin (GIO_N2HET1, 14); // CS2
	dev_spi_initpin (GIO_N2HET1, 12); // CS3
}

void dev_spi_initgio_3 (void)
{
	dev_spi_initpin ((_GIOAB*) &SPI3->SPIPC1, 0); // CS0
	dev_spi_initpin ((_GIOAB*) &SPI3->SPIPC1, 8); // CS1
}

void dev_spi_initgio_4 (void)
{
	dev_spi_initpin (GIO_N2HET1, 26); // CS0
	dev_spi_initpin ((_GIOAB*) &SPI1->SPIPC1, 0); // CS1
	dev_spi_initpin ((_GIOAB*) &SPI1->SPIPC1, 8); // CS2
}

int dev_spi_handle_cs_1 (unsigned int num)
{
	switch (num)
	{
	case 0:
		// Disable all
		GIO->b.DSET = 1 << 2;
		GIO_N2HET1->DSET = (1 << 20) | (1 << 18) | (1 << 16);
		break;
	case 1:
		// Chip select 0
		GIO->b.DCLR = 1 << 2;
		break;
	case 2:
		// Chip select 1
		GIO_N2HET1->DCLR = 1 << 20;
		break;
	case 3:
		// Chip select 2
		GIO_N2HET1->DCLR = 1 << 18;
		break;
	case 4:
		// Chip select 3
		GIO_N2HET1->DCLR = 1 << 16;
		break;
	default:
		return -EINVAL;
	}

	return 0;
}

int dev_spi_handle_cs_2 (unsigned int num)
{
	switch (num)
	{
	case 0:
		// Disable all
		GIO->b.DSET = 1 << 0;
		GIO_N2HET1->DSET = (1 << 30) /*| (1 << 14)*/ | (1 << 12);
		break;
	case 1:
		// Chip select 0
		GIO_N2HET1->DCLR = 1 << 30;
		break;
	case 2:
		// Chip select 1
		GIO->b.DCLR = 1 << 0;
		break;
	/*case 3:
		// Chip select 2
		GIO_N2HET1->DCLR = 1 << 14;
		break;*/
	case 4:
		// Chip select 3
		GIO_N2HET1->DCLR = 1 << 12;
		break;
	default:
		return -EINVAL;
	}

	return 0;
}

int dev_spi_handle_cs_3 (unsigned int num)
{
	switch (num)
	{
	case 0:
		// Disable all
		SPI3->SPIPC4 = (1 << 0) | (1 << 8);
		break;
	case 1:
		// Chip select 0
		SPI3->SPIPC5 = 1 << 0;
		break;
	case 2:
		// Chip select 0
		SPI3->SPIPC5 = 1 << 8;
		break;
	default:
		return -EINVAL;
	}

	return 0;
}

int dev_spi_handle_cs_4 (unsigned int num)
{
	switch (num)
	{
	case 0:
		// Disable all
		GIO_N2HET1->DSET = 1 << 26;
		SPI1->SPIPC4 = (1 << 0) | (1 << 8);
		break;
	case 1:
		// Chip select 0
		GIO_N2HET1->DCLR = 1 << 26;
		break;
	case 2:
		// Chip select 1
		SPI1->SPIPC5 = 1 << 0;
		break;
	case 3:
		// Chip select 2
		SPI1->SPIPC5 = 1 << 8;
		break;
	default:
		return -EINVAL;
	}

	return 0;
}

///////////////////////////////////////////////////////////////////////////////
// Private init functions
///////////////////////////////////////////////////////////////////////////////
void dev_spi_init (struct spi *s)
{
	// Tables
	static const _SPI *instances[] = {SPI5, SPI2, SPI3, SPI1};
	static const unsigned char reqchrx[] = {30, 0, 14, 0};
	static const unsigned char reqchtx[] = {31, 1, 15, 1};
	static const unsigned char dmachrx[] = {12, 6, 8, 6};
	static const unsigned char dmachtx[] = {13, 7, 9, 7};

	// Set instance pointer
	s->instance = (_SPI*) instances[s->spinum];

	// Set DMA stuff
	s->reqchrx = reqchrx[s->spinum];
	s->reqchtx = reqchtx[s->spinum];
	s->dmachrx = dmachrx[s->spinum];
	s->dmachtx = dmachtx[s->spinum];

	// Reset the peripheral
	s->instance->SPIGCR0 = 0;
	s->instance->SPIGCR0 = 1;

	// Set pin direction (1 = output)
	// Note that these only apply in gio mode, we may actually skip setting it
	s->instance->SPIPC1 = 0x0000063F;

	// Value of pins in gio output mode
	s->instance->SPIPC3 = 1;

	// Open-drain output mode
	s->instance->SPIPC6 = 0;

	// Pull-up/down enable
	s->instance->SPIPC7 = 0x00000800;

	// Pull-up or pull-down
	s->instance->SPIPC8 = 0x00000000;

	// Set SPI pins to SPI mode, but CS pins to GIO mode
	s->instance->SPIPC0 = 0x00000E00;

	// Set SPI control register
	s->instance->SPIGCR1 = 0x00000003;

	// Configure data format 0: Bitrate=VCLK/2, 8-bit words.
	s->instance->SPIFMT[0] = (1 << 24) | 0x00000008 | (1 << 16);

	// Configure timing delays
	s->instance->SPIDELAY = 0x0F0F0000;

	// Interrupt enable, disable all
	s->instance->SPIINT0 = 0;

	// Clear interrupts
	s->instance->SPIFLG = 0x0000FFFF;

	// Default chip selects
	s->instance->SPIDEF = 0x000000FF;

	// Configure transfer
	s->instance->SPIDAT1 = 0x14000000;

	// Enable the SPI
	s->instance->SPIGCR1 |= (1 << 24);

	switch (s->spinum)
	{
	case 0:
		dev_spi_initgio_1 ();
		break;
	case 1:
		dev_spi_initgio_2 ();
		break;
	case 2:
		dev_spi_initgio_3 ();
		break;
	case 3:
		dev_spi_initgio_4 ();
		break;
	default:
		break;
	}
}

///////////////////////////////////////////////////////////////////////////////
// File node functions
///////////////////////////////////////////////////////////////////////////////
int dev_spi_open (struct FileNode *node, const char *filename)
{
	// Get spi number
	unsigned int spinum = filename[3] - '0';

	// Check it
	if (spinum >= N_SPI_MODULES)
		return -ENXIO;

	// Already initialized?
	if (spi_irq_instances[spinum])
		return -EBUSY;

	// Allocate (static for now)
	static struct spi sp[N_SPI_MODULES];
	struct spi *t = &sp[spinum];

	// Clear the structure
	MemSet (t, 0, sizeof (struct spi));

	// Set the num
	t->spinum = spinum;

	// Set the irq-handlers
	spi_irq_instances[spinum] = t;

	// Set the file node
	node->priv = (void*) t;

	// Enable DMA module and stop at suspend
	DMAInitModule ();

	// Initialize SPI
	dev_spi_init (t);

	// Done
	return 0;
}

int dev_spi_close (struct FileNode *node)
{
	// Extract pointer
	struct spi *t = (struct spi*) node->priv;

	// Disable the peripheral
	t->instance->SPIGCR0 = 0;

	// Clear irqhandler (also "clears" the static allocation)
	spi_irq_instances[t->spinum] = 0;

	// Done
	return 0;
}

int dev_spi_read (struct FileNode *node, void *data, unsigned int nbytes)
{
	// Extract pointer
	struct spi *t = (struct spi*) node->priv;

	// Wait for DMA and SPI to be available
	if (t->busystate)
		if (!(DMA->BTCFLAG & (1 << t->dmachrx)))
			return -EBUSY;

	// Stop transfer
	t->instance->SPIINT0 &= (1 << 16);
	t->instance->SPIGCR1 = 3;
	t->instance->SPIFLG = 0x0000FFFF;

	// Clear DMA transfer complete flag and busy state
	DMA->BTCFLAG = 1 << t->dmachtx;
	DMA->BTCFLAG = 1 << t->dmachrx;
	t->busystate = 0;

	// Byte to send when transmitting
	static unsigned char txbyte = 0x7E;

	// DMA Setup
	DMAChannelInit (t->dmachtx, t->reqchtx,
		(unsigned int) &txbyte,
		((unsigned int) &t->instance->SPIDAT1) + 3,
		nbytes,
		0);

	DMAChannelInit (t->dmachrx, t->reqchrx,
		((unsigned int) &t->instance->SPIBUF) + 3,
		((unsigned int) data),
		nbytes,
		2);

	// Set busy state
	t->busystate = 1;

	// Enable transmit
	t->instance->SPIGCR1 |= 1 << 24;
	t->instance->SPIINT0 |= 1 << 16;

	// Done
	return nbytes;
}

int dev_spi_write (struct FileNode *node, void *data, unsigned int nbytes)
{
	// Extract pointer
	struct spi *t = (struct spi*) node->priv;

	// Wait for DMA and SPI to be available
	if (t->busystate)
		if (!(DMA->BTCFLAG & (1 << t->dmachrx)))
			return -EBUSY;

	// Stop transfer
	t->instance->SPIINT0 &= (1 << 16);
	t->instance->SPIGCR1 = 3;
	t->instance->SPIFLG = 0x0000FFFF;

	// Clear DMA transfer complete flag and busy state
	DMA->BTCFLAG = 1 << t->dmachtx;
	DMA->BTCFLAG = 1 << t->dmachrx;
	t->busystate = 0;

	// DMA Setup
	DMAChannelInit (t->dmachtx, t->reqchtx,
		(unsigned int) data,
		((unsigned int) &t->instance->SPIDAT1) + 3,
		nbytes,
		8);

	static unsigned char rxbyte = 0x7E;

	DMAChannelInit (t->dmachrx, t->reqchrx,
		(unsigned int) &rxbyte,
		((unsigned int) &t->instance->SPIBUF) + 3,
		nbytes,
		0);

	// Set busy state
	t->busystate = 2;

	// Enable transmit
	t->instance->SPIGCR1 |= 1 << 24;
	t->instance->SPIINT0 |= 1 << 16;

	// Done
	return nbytes;
}

int dev_spi_ioctl (struct FileNode *node, unsigned int code, void *data)
{
	// Extract pointer
	struct spi *t = (struct spi*) node->priv;

	// Switch on the ioctl control code
	switch (code)
	{
	case SPIBUSY:
		if (t->busystate)
		{
			// The SPI has been set as busy and may have completed.
			// For both read and write we set up the rx dma.
			// It will tell us when the entire transfer is complete,
			// note that the flags in the SPI does not actually tell
			// when the transmit shift register is empty.
			if (!(DMA->BTCFLAG & (1 << t->dmachrx)))
				return 1;
		}

		// Not busy
		return 0;
	case SPICS:
		switch (t->spinum)
		{
		case 0: return dev_spi_handle_cs_1 ((unsigned int) data);
		case 1: return dev_spi_handle_cs_2 ((unsigned int) data);
		case 2: return dev_spi_handle_cs_3 ((unsigned int) data);
		case 3: return dev_spi_handle_cs_4 ((unsigned int) data);
		default: return -EINVAL;
		}
	case SPIFORCEUNLOCK:
		DMA->BTCFLAG = 1 << t->dmachtx;
		DMA->BTCFLAG = 1 << t->dmachrx;
		t->busystate = 0;
		return 0;
	default:
		// Unknown ioctl: Return "Inappropriate ioctl for device"
		return -ENOTTY;
	}
}

MODULE_DESCRIPTOR (spi0,
	dev_spi_open,
	dev_spi_close,
	dev_spi_read,
	dev_spi_write,
	dev_spi_ioctl,
	"spi0");

MODULE_DESCRIPTOR (spi1,
	dev_spi_open,
	dev_spi_close,
	dev_spi_read,
	dev_spi_write,
	dev_spi_ioctl,
	"spi1");

MODULE_DESCRIPTOR (spi2,
	dev_spi_open,
	dev_spi_close,
	dev_spi_read,
	dev_spi_write,
	dev_spi_ioctl,
	"spi2");

MODULE_DESCRIPTOR (spi3,
	dev_spi_open,
	dev_spi_close,
	dev_spi_read,
	dev_spi_write,
	dev_spi_ioctl,
	"spi3");
