/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "mem/string.h"
#include "posix/errno.h"
#include "fs/dev/dev.h"
#include "ioctl/ioctl.h"
#include "drivers/dma/dma-tms570.h"
#include "drivers/tty/tty-tms570.h"

///////////////////////////////////////////////////////////////////////////////
// Functions
///////////////////////////////////////////////////////////////////////////////
int dev_tty_read_direct (tty *t, void *data, unsigned int nbytes)
{
	// Fail if already in the receiving state, ignore error mode
	if (t->busystate == 1)
		return -EBUSY;

	// Reset DMA, stop SCI DMA transfer, disable tx
	DMA->BTCFLAG = 1 << t->dmachrx;
	t->instance->CLEARINT = 1 << 17;
	t->instance->GCR1 &= ~(1 << 24);
	t->busystate = 0;

	// DMA Setup
	DMAChannelInit (t->dmachrx, t->reqchrx,
		((unsigned int) &t->instance->RD) + 3,
		(unsigned int) data,
		nbytes,
		2);

	// Set busy state
	t->busystate = 1;

	// Enable transmit
	t->instance->GCR1 |= 1 << 24;
	t->instance->SETINT = 1 << 17;

	// Done
	return nbytes;
}

int dev_tty_write_direct (tty *t, void *data, unsigned int nbytes)
{
	// Abort on zero write
	if (!nbytes)
		return 0;

	// Make sure we do not accept more than TTY_BUF_SIZE bytes
	unsigned int n = (nbytes < TTY_BUF_SIZE) ? nbytes : TTY_BUF_SIZE;

	// Wait timer
	int timer = 1000000;

	// Wait for DMA to be available
	if (t->busystate)
		while (timer && !(DMA->BTCFLAG & (1 << t->dmachtx)))
			timer--;

	// Wait for serial interface to complete transfers (TXEMPTY)
	while (timer && (t->instance->FLR & (1 << 11)) == 0)
		timer--;

	// Timer error check
	if (!timer)
		return -EBUSY;

	// Reset DMA, stop SCI DMA transfer, disable tx
	DMA->BTCFLAG = 1 << t->dmachtx;
	t->instance->CLEARINT = 1 << 16;
	t->instance->GCR1 &= ~(1 << 25);
	t->busystate = 0;

	// Copy data to transmit buffer
	for (unsigned int i = 0; i < n; i++)
		t->deftxbuf[i] = ((char*)data)[i];

	// DMA Setup
	DMAChannelInit (t->dmachtx, t->reqchtx,
		(unsigned int) t->deftxbuf,
		((unsigned int) &t->instance->TD) + 3,
		n,
		8);

	// Set busy state
	t->busystate = 1;

	// Enable transmit
	t->instance->GCR1 |= 1 << 25;
	t->instance->SETINT = 1 << 16;

	// Return number of bytes
	return n;
}
