/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "mem/string.h"
#include "posix/errno.h"
#include "posix/sys/signal.h"
#include "process/process.h"
#include "fs/dev/dev.h"
#include "ioctl/ioctl.h"
#include "drivers/dma/dma-tms570.h"
#include "drivers/tty/tty-tms570.h"

///////////////////////////////////////////////////////////////////////////////
// Variables
///////////////////////////////////////////////////////////////////////////////
extern tty *irq_instances[2];

///////////////////////////////////////////////////////////////////////////////
// Interrupts
///////////////////////////////////////////////////////////////////////////////
void dev_tty_irq_normal_rdcheck (tty *t)
{
	// Get now position
	unsigned int reg = DMACH->WCP[t->dmachrx].CDADDR;

	if (!reg || reg < (unsigned int) &t->defrxbuf[0] || reg >= (unsigned int) &t->defrxbuf[TTY_BUF_SIZE])
		return;

	unsigned int now = reg - (unsigned int) t->defrxbuf;

	// Calculate number of bytes
	unsigned int oldpos = t->readposirq;
	unsigned int n = now - oldpos;

	// In case of wrap-around
	if (now < oldpos)
		n += TTY_BUF_SIZE;

	// Copy data from the input buffer
	for (int i = 0; i < n; i++)
	{
		// Fetch the character
		char d = t->defrxbuf[(oldpos + i) & TTY_BUF_MASK];

		// Check for SIGINT and SIGTSTP, send them to the process group
		if (d == 0x03)
			ProcessSendSignal (-t->pgid, SIGINT);
		else if (d == 0x1A)
			ProcessSendSignal (-t->pgid, SIGTSTP);
	}

	// Set old position
	t->readposirq = (oldpos + n) & TTY_BUF_MASK;
}

///////////////////////////////////////////////////////////////////////////////
// Functions
///////////////////////////////////////////////////////////////////////////////
int dev_tty_read_normal (tty *t, void *data, unsigned int nbytes)
{
	// Get now position
	unsigned int reg = DMACH->WCP[t->dmachrx].CDADDR;

	if (!reg || reg < (unsigned int) &t->defrxbuf[0] || reg >= (unsigned int) &t->defrxbuf[TTY_BUF_SIZE])
		return 0;

	unsigned int now = reg - (unsigned int) t->defrxbuf;

	// Calculate number of bytes
	unsigned int oldpos = t->readpos;
	unsigned int n = now - oldpos;

	// In case of wrap-around
	if (now < oldpos)
		n += TTY_BUF_SIZE;

	// Make shure we never read more than nbytes
	if (nbytes < n)
		n = nbytes;

	// Number of bytes read
	int nread = 0;

	// Copy data from the input buffer
	for (int i = 0; i < n; i++)
	{
		char d = t->defrxbuf[(oldpos + i) & TTY_BUF_MASK];
		if (d != 0x03 && d != 0x1A)
			((char*)data)[nread++] = d;
	}

	// Set old position
	t->readpos = (oldpos + n) & TTY_BUF_MASK;

	// Done
	return nread;
}

int dev_tty_write_normal (tty *t, void *data, unsigned int nbytes)
{
	// Abort on zero write
	if (!nbytes)
		return 0;

	// Make sure we do not accept more than TTY_BUF_SIZE bytes
	unsigned int n = (nbytes < TTY_BUF_SIZE) ? nbytes : TTY_BUF_SIZE;

	// Wait timer
	int timer = 1000000;

	// Wait for DMA to be available
	if (t->busystate)
		while (timer && !(DMA->BTCFLAG & (1 << t->dmachtx)))
			timer--;

	// Wait for serial interface to complete transfers (TXEMPTY)
	while (timer && (t->instance->FLR & (1 << 11)) == 0)
		timer--;

	// Timer error check
	if (!timer)
		return -EBUSY;

	// Reset DMA, stop SCI DMA transfer, disable tx
	DMA->BTCFLAG = 1 << t->dmachtx;
	t->instance->CLEARINT = 1 << 16;
	t->instance->GCR1 &= ~(1 << 25);
	t->busystate = 0;

	// Copy data to transmit buffer
	for (unsigned int i = 0; i < n; i++)
		t->deftxbuf[i] = ((char*)data)[i];

	// DMA Setup
	DMAChannelInit (t->dmachtx, t->reqchtx,
		(unsigned int) t->deftxbuf,
		((unsigned int) &t->instance->TD) + 3,
		n,
		8);

	// Set busy state
	t->busystate = 1;

	// Enable transmit
	t->instance->GCR1 |= 1 << 25;
	t->instance->SETINT = 1 << 16;

	// Return number of bytes
	return n;
}
