/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "mem/string.h"
#include "posix/errno.h"
#include "fs/dev/dev.h"
#include "ioctl/ioctl.h"
#include "drivers/dma/dma-tms570.h"
#include "drivers/tty/tty-tms570.h"

///////////////////////////////////////////////////////////////////////////////
// Variables
///////////////////////////////////////////////////////////////////////////////

// These are used by the IRQ handlers to get back to the data
tty *irq_instances[2] = {0};

///////////////////////////////////////////////////////////////////////////////
// Forward declarations
///////////////////////////////////////////////////////////////////////////////
void dev_tty_uart_defconfig (tty *t, int ttynum);
void dev_tty_uart_setbaud (tty *t, unsigned int vclk);
void dev_tty_uart_startrecv (tty *t);
void dev_tty_uart_init (tty *t);

// Normal circular mode
int dev_tty_read_normal (tty *t, void *data, unsigned int nbytes);
int dev_tty_write_normal (tty *t, void *data, unsigned int nbytes);
void dev_tty_irq_normal_rdcheck (tty *t);

// Direct buffer DMA mode
int dev_tty_read_direct (tty *t, void *data, unsigned int nbytes);
int dev_tty_write_direct (tty *t, void *data, unsigned int nbytes);

// Direct buffer circular DMA mode
int dev_tty_read_direct_circ (tty *t, void *data, unsigned int nbytes);
void dev_tty_irq_direct_circ (tty *t);

///////////////////////////////////////////////////////////////////////////////
// Interrupt handlers
///////////////////////////////////////////////////////////////////////////////
void dev_tty_irq (int index)
{
	unsigned int vals = DMA->BTCFLAG;
	tty *t = irq_instances[1];

	if (vals & (1 << t->dmachbs))
	{
		DMA->BTCFLAG = (1 << t->dmachbs);
		dev_tty_irq_normal_rdcheck (t);
	}
}

///////////////////////////////////////////////////////////////////////////////
// Uart functions
///////////////////////////////////////////////////////////////////////////////
void dev_tty_uart_defconfig (tty *t, int ttynum)
{
	// Lookup tables
	static const _SCI *instances[] = {SCI1, SCI2};
	static const char reqchrx[] = {30, 28};
	static const char reqchtx[] = {31, 29};
	static const char dmachrx[] = {3, 1};
	static const char dmachtx[] = {4, 0};
	static const char dmachbs[] = {5, 2};

	// Set the instance
	t->instance = (_SCI*) instances[ttynum];

	// Set DMA stuff
	t->reqchrx = reqchrx[ttynum];
	t->reqchtx = reqchtx[ttynum];
	t->dmachrx = dmachrx[ttynum];
	t->dmachtx = dmachtx[ttynum];
	t->dmachbs = dmachbs[ttynum];

	// Set the tty number
	t->ttynum = ttynum;

	// Default parameters
	t->baudrate = 38400;

	// Set function pointers
	t->read = dev_tty_read_normal;
	t->write = dev_tty_write_normal;
	t->irq = dev_tty_irq_normal_rdcheck;
}

void dev_tty_uart_setbaud (tty *t, unsigned int vclk)
{
	// Caclulate and set baud rate register
	t->instance->BRS = (vclk / (16 * t->baudrate)) - 1;
}

void dev_tty_uart_startrecv (tty *t)
{
	// Stop transfer
	t->instance->CLEARINT = 3 << 17;

	// Static arrays used to trick the arbiter
	static const unsigned char bullshitarray[8] = {0, 1, 2, 3, 4, 5, 6, 7};
	static unsigned char bullshitcounter = 0;

	// Set receive DMA
	DMAChannelInit (t->dmachrx, t->reqchrx,
		((unsigned int) &t->instance->RD) + 3,
		(unsigned int) t->defrxbuf,
		TTY_BUF_SIZE,
		3);

	// Set bullshit DMA
	DMAChannelInit (t->dmachbs, t->reqchrx,
		(unsigned int) bullshitarray,
		(unsigned int) &bullshitcounter,
		1,
		9);

	// Enable transmit
	t->instance->SETINT = 3 << 17;

	// Bullshit DMA IRQ enable
	DMA->GCHIENAS = 1 << t->dmachbs;
	DMA->BTCINTENAS = 1 << t->dmachbs;
}

void dev_tty_uart_init (tty *t)
{
	// Fetch the instance pointer
	_SCI *inst = t->instance;

	// Reset the hardware
	inst->GCR0 = 0;
	inst->GCR0 = 1;
	inst->GCR1 = 0;

	// Configuration: RX/TX, CONT, internal clock, 1 stop, no parity, async
	inst->GCR1 = 0x03020022;

	// Disable all interrupts
	inst->CLEARINT = 0xFFFFFFFF;

	// All interrupts are low priority
	inst->CLEARINTLVL = 0xFFFFFFFF;

	// Data format: 8-bit characters
	inst->FORMAT = 7;

	// Set the baud
	dev_tty_uart_setbaud (t, 16000000);

	// Assign pins to SCI
	inst->PIO0 = 6;

	// Output direction for GPIO use
	inst->PIO1 = 4;

	// Output values when used in GPIO mode
	inst->PIO3 = 4;

	// Use push-pull pin mode
	inst->PIO6 = 0;

	// Enable pull-up resistor on both pins
	inst->PIO7 = 0;
	inst->PIO8 = 6;

	// Enable the SCI
	inst->GCR1 |= 0x00000080;

	// Enable DMA reception
	dev_tty_uart_startrecv (t);
}

///////////////////////////////////////////////////////////////////////////////
// Driver functions
///////////////////////////////////////////////////////////////////////////////
int dev_tty_open (struct FileNode *node, const char *filename)
{
	// Get tty number
	unsigned int ttynum = filename[3] - '0';

	// Check it
	if (ttynum > 1)
		return -ENXIO;

	// Already initialized?
	if (irq_instances[ttynum])
	{
		// TODO: Remove this again when we have dup2
		node->priv = (void*) irq_instances[ttynum];
		return 0;
		//return -EBUSY;
	}

	// Allocate (static for now)
	static tty ty[2];
	tty *t = &ty[ttynum];

	// Clear the structure
	MemSet (t, 0, sizeof (tty));

	// Set the irq-handlers
	irq_instances[ttynum] = t;

	// Set the file node
	node->priv = (void*) t;

	// Enable DMA module and stop at suspend
	DMAInitModule ();

	// Set configuration
	dev_tty_uart_defconfig (t, ttynum);

	// Initialize uart
	dev_tty_uart_init (t);

	// Done
	return 0;
}

int dev_tty_close (struct FileNode *node)
{
	// Extract tty pointer
	tty *t = (tty*) node->priv;

	// Make sure DMA channels and UART is disabled
	t->instance->GCR1 = 0;

	// Clear irqhandler (also "clears" the static allocation)
	irq_instances[t->ttynum] = 0;

	// Done
	return 0;
}

int dev_tty_read (struct FileNode *node, void *data, unsigned int nbytes)
{
	// Extract tty pointer and call read
	tty *t = (tty*) node->priv;
	return t->read (t, data, nbytes);
}

int dev_tty_write (struct FileNode *node, void *data, unsigned int nbytes)
{
	// Extract tty pointer
	tty *t = (tty*) node->priv;

	// Call the correct write functions
	return t->write (t, data, nbytes);
}

int dev_tty_ioctl (struct FileNode *node, unsigned int code, void *data)
{
	// Extract tty pointer
	tty *t = (tty*) node->priv;

	// Switch on the ioctl control code
	switch (code)
	{
	case TCGETS:
		{
			struct termios *tios = (struct termios*) data;
			tios->baudrate = t->baudrate;
		}
		return 0;
	case TCSETS:
		{
			struct termios *tios = (struct termios*) data;
			t->baudrate = tios->baudrate;
			dev_tty_uart_setbaud (t, 16000000);
		}
		return 0;
	case TISATTY:
		// Always return 1 since this is a terminal
		return 1;
	case TCGETPGRP:
		return t->pgid;
	case TCSETPGRP:
		t->pgid = (short) ((int)data);
		return 0;
	case TTYSETMODE:
		if ((int)data == 2)
		{
			// Direct mode output, circular receive with no processing
			t->read = dev_tty_read_direct_circ;
			t->write = dev_tty_write_direct;
			t->irq = dev_tty_irq_direct_circ;
			t->busystate = 0;
			t->readpos = 0;
			t->readposirq = 0;

			// Disable DMA irq
			DMA->GCHIENAR = 1 << t->dmachbs;
		}
		else if ((int)data == 1)
		{
			// Direct mode used by save program, DMA direct to user
			// After this, circular receive is still active
			t->read = dev_tty_read_direct;
			t->write = dev_tty_write_direct;
			t->irq = dev_tty_irq_normal_rdcheck;
			t->busystate = 0;

			// Disable DMA irq
			DMA->GCHIENAR = 1 << t->dmachbs;
		}
		else if ((int)data == 0)
		{
			// Normal mode
			t->read = dev_tty_read_normal;
			t->write = dev_tty_write_normal;
			t->irq = dev_tty_irq_normal_rdcheck;
			t->busystate = 0;
			t->readpos = 0;
			t->readposirq = 0;

			// Restart circular receive
			dev_tty_uart_startrecv (t);
		}
		else
			return -ENOTTY;
		break;
	case TTYGETRXBUSY:
		// Check DMA flag for completion
		if (t->busystate && (DMA->BTCFLAG & (1 << t->dmachrx)))
			t->busystate = 0;

		// Return that state
		return t->busystate;
	case TTYGETWRPOS:
		return DMACH->WCP[t->dmachrx].CDADDR;
	default:
		// Unknown ioctl: Return "Inappropriate ioctl for device"
		return -ENOTTY;
	}

	// Default return value
	return 0;
}

///////////////////////////////////////////////////////////////////////////////
// Driver declarations
///////////////////////////////////////////////////////////////////////////////
MODULE_DESCRIPTOR (tty0,
	dev_tty_open,
	dev_tty_close,
	dev_tty_read,
	dev_tty_write,
	dev_tty_ioctl,
	"tty0");

MODULE_DESCRIPTOR (tty1,
	dev_tty_open,
	dev_tty_close,
	dev_tty_read,
	dev_tty_write,
	dev_tty_ioctl,
	"tty1");
