/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "fs/vfs.h"
#include "fs/dev/dev.h"
#include "posix/errno.h"
#include "ioctl/ioctl.h"
#include "kmod/kmod.h"

///////////////////////////////////////////////////////////////////////////////
// Definitions
///////////////////////////////////////////////////////////////////////////////
#define VOLUME_HEADROOM 6

///////////////////////////////////////////////////////////////////////////////
// External functions
///////////////////////////////////////////////////////////////////////////////
int dev_i2c_open (struct FileNode *node, const char *filename);
int dev_i2c_read (struct FileNode *node, void *data, unsigned int nbytes);
int dev_i2c_write (struct FileNode *node, void *data, unsigned int nbytes);

///////////////////////////////////////////////////////////////////////////////
// Variables
///////////////////////////////////////////////////////////////////////////////
int lch_cur;
int rch_cur;

///////////////////////////////////////////////////////////////////////////////
// Helpers
///////////////////////////////////////////////////////////////////////////////
int PT2257_SaveVol (struct FileNode *node, int lch, int rch)
{
	unsigned char data[] = {0x52, 0, lch >> 24, lch >> 16, lch >> 8, lch,
		rch >> 24, rch >> 16, rch >> 8, rch};

	int ret = dev_i2c_write (node, data, sizeof(data));

	if (ret < 0)
		return ret;

	return 0;
}

int PT2257_LoadVol (struct FileNode *node, int *lch, int *rch)
{
	unsigned char data[9] = {0x52, 0};
	int ret;

	if ((ret = dev_i2c_write (node, data, 2)) != 2)
		return ret;

	if ((ret = dev_i2c_read (node, data, 9)) != 9)
		return ret;

	*lch = (data[1] << 24) | (data[2] << 16) | (data[3] << 8) | data[4];
	*rch = (data[5] << 24) | (data[6] << 16) | (data[7] << 8) | data[8];

	return 0;
}

int PT2257_SetVol (struct FileNode *node, int lch, int rch)
{
	lch -= VOLUME_HEADROOM;
	rch -= VOLUME_HEADROOM;

	if (lch > 0) lch = 0;
	if (lch < -79) lch = -79;
	if (rch > 0) rch = 0;
	if (rch < -79) rch = -79;

	lch_cur = lch + VOLUME_HEADROOM;
	rch_cur = rch + VOLUME_HEADROOM;

	lch = -lch;
	rch = -rch;

	unsigned char data[9];

	data[0] = 0x44;
	data[1] = 0xB0 | (lch / 10);
	data[2] = 0xB0 | (lch / 10);
	data[3] = 0xA0 | (lch % 10);
	data[4] = 0xA0 | (lch % 10);
	data[5] = 0x30 | (rch / 10);
	data[6] = 0x30 | (rch / 10);
	data[7] = 0x20 | (rch % 10);
	data[8] = 0x20 | (rch % 10);

	int ret = dev_i2c_write (node, data, 9);

	if (ret < 0)
		return ret;

	return 0;
}

///////////////////////////////////////////////////////////////////////////////
// Driver functions
///////////////////////////////////////////////////////////////////////////////
int dev_pt2257_open (struct FileNode *node, const char *filename)
{
	static struct FileNode pt2257_node;
	static char pt2257_initialized;

	node->priv = &pt2257_node;

	if (!pt2257_initialized)
	{
		int ret;

		if ((ret = dev_i2c_open (&pt2257_node, "/dev/i2c")))
			return ret;

		pt2257_initialized = 1;

		if ((ret = PT2257_LoadVol (&pt2257_node, &lch_cur, &rch_cur)))
			return ret;

		if ((ret = PT2257_SetVol (&pt2257_node, lch_cur, rch_cur)))
			return ret;
	}

	return 0;
}

int dev_pt2257_close (struct FileNode *node)
{
	return 0;
}

int dev_pt2257_read (struct FileNode *node, void *data, unsigned int nbytes)
{
	if (nbytes != 8)
		return -EINVAL;

	((int*)data)[0] = lch_cur;
	((int*)data)[1] = rch_cur;

	return nbytes;
}

int dev_pt2257_write (struct FileNode *node, void *data, unsigned int nbytes)
{
	if (nbytes != 8)
		return -EINVAL;

	int ret;

	if ((ret = PT2257_SetVol ((struct FileNode*)node->priv, ((int*)data)[0], ((int*)data)[1])))
		return ret;

	if ((ret = PT2257_SaveVol (node, lch_cur, rch_cur)))
		return ret;

	return nbytes;
}

int dev_pt2257_ioctl (struct FileNode *node, unsigned int code, void *data)
{
	// Unknown ioctl: Return "Inappropriate ioctl for device"
	return -ENOTTY;
}

void dev_pt2257_init (void)
{
	struct FileNode node;

	dev_pt2257_open (&node, 0);
}

///////////////////////////////////////////////////////////////////////////////
// Driver declarations
///////////////////////////////////////////////////////////////////////////////
MODULE_DESCRIPTOR (pt2257,
	dev_pt2257_open,
	dev_pt2257_close,
	dev_pt2257_read,
	dev_pt2257_write,
	dev_pt2257_ioctl,
	"pt2257");

KMOD_DESCRIPTOR (pt2257, dev_pt2257_init, "pt2257_init");
