/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "mem/string.h"
#include "posix/unistd.h"
#include "elf/elf.h"

///////////////////////////////////////////////////////////////////////////////
// Functions
///////////////////////////////////////////////////////////////////////////////
int CheckHeader (struct Loader *l)
{
	struct ElfHeader eh;

	int r = LoadSectionData (l, 0, sizeof (struct ElfHeader), &eh);

	if (r)
		return r;

	static const char magic[4] = {0x7f, 0x45, 0x4c, 0x46};

	if (MemCmp (eh.magic, magic, 4))
		return EELFLDR_MAGIC;

#ifdef BIGE
	static const char versionpart[5+7+8] =
		{1, 2, 1, 0, 0,
		0, 0, 0, 0, 0, 0, 0,
		0, 1, 0, 0x28, 0, 0, 0, 1};
#else
	static const char versionpart[5+7+8] =
		{1, 1, 1, 0, 0,
		0, 0, 0, 0, 0, 0, 0,
		1, 0, 0x28, 0, 1, 0, 0, 0};
#endif

	if (MemCmp (eh.versionpart, versionpart, sizeof (versionpart)))
		return EELFLDR_VERSION;

	l->entrypoint = eh.entrypoint;
	l->shoff = eh.shoff;
	l->shnum = eh.shnum;
	l->shstrndx = eh.shstrndx;

	return EELFLDR_OK;
}

int LoadSectionHeader (struct Loader *l, struct ElfSection *sh, int index)
{
	unsigned int seekto = l->shoff + ((unsigned int)index * 0x28);
	return LoadSectionData (l, seekto, sizeof (struct ElfSection), sh);
}

int LoadSectionData (struct Loader *l, unsigned int offset, unsigned int size, void *target)
{
	if (lseek (l->fd, offset, SEEK_SET) != offset)
		return EELFLDR_SEEK;

	if (read (l->fd, target, size) != size)
		return EELFLDR_READ;

	return EELFLDR_OK;
}

void FindSizesHelper (struct LoaderSection *dst, struct ElfSection *src, int i)
{
	dst->offset = src->offset;
	dst->size = src->size;
	dst->index = i;
}

int FindSizes (struct Loader *l)
{
	for (int i = 0; i < l->shnum; i++)
	{
		struct ElfSection sh;
		MemSet (&sh, 0, sizeof (sh));

		int r = LoadSectionHeader (l, &sh, i);

		if (r)
			return r;

		if (sh.nameoffset >= SELFLDR_NAMES_SIZE)
			return EELFLDR_WOULDOVERFLOW;

		char *name = &l->names[sh.nameoffset];

		if (!StrCmp (name, ".text"))
			FindSizesHelper (&l->text, &sh, i);
		else if (!StrCmp (name, ".rodata"))
			FindSizesHelper (&l->rodata, &sh, i);
		else if (!StrCmp (name, ".data"))
			FindSizesHelper (&l->data, &sh, i);
		else if (!StrCmp (name, ".bss"))
			FindSizesHelper (&l->bss, &sh, i);
		else if (!StrCmp (name, ".rel.text"))
			FindSizesHelper (&l->rel_text, &sh, i);
		else if (!StrCmp (name, ".rel.rodata"))
			FindSizesHelper (&l->rel_rodata, &sh, i);
		else if (!StrCmp (name, ".rel.data"))
			FindSizesHelper (&l->rel_data, &sh, i);
		else if (!StrCmp (name, ".rel.bss"))
			FindSizesHelper (&l->rel_bss, &sh, i);
		else if (!StrCmp (name, ".symtab"))
			FindSizesHelper (&l->symbols, &sh, i);
		else if (!StrCmp (name, ".strtab"))
			FindSizesHelper (&l->strings, &sh, i);
	}

	return EELFLDR_OK;
}

int LoadData (struct Loader *l, struct Program *p)
{
	int r = 0;

	if (p->text && (r = LoadSectionData (l, l->text.offset, l->text.size, p->text)))
		return r;

	if (p->rodata && (r = LoadSectionData (l, l->rodata.offset, l->rodata.size, p->rodata)))
		return r;

	if (p->data && (r = LoadSectionData (l, l->data.offset, l->data.size, p->data)))
		return r;

	return EELFLDR_OK;
}
