/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "mem/string.h"
#include "posix/errno.h"
#include "posix/sys/statvfs.h"
#include "fs/dev/dev.h"

///////////////////////////////////////////////////////////////////////////////
// Variables
///////////////////////////////////////////////////////////////////////////////
int vfs_dev_curdir;

///////////////////////////////////////////////////////////////////////////////
// External variables
///////////////////////////////////////////////////////////////////////////////
extern struct DeviceDriver __dev_list_start;
extern struct DeviceDriver __dev_list_end;

///////////////////////////////////////////////////////////////////////////////
// Functions
///////////////////////////////////////////////////////////////////////////////
int vfs_dev_open (struct FileNode *node, const char *filename, int flags, int mode)
{
	// We may get a filename with a slash, skip over it
	if (*filename == '/')
		filename++;

	// Loop through list of devices to find it
	for (struct DeviceDriver *i = &__dev_list_start; i < &__dev_list_end; ++i)
	{
		if (!StrCmp (filename, i->name))
		{
			// Found match, store pointer and try opening
			node->dev = i;
			return node->dev->open (node, filename);
		}
	}

	// Device not found
	return -1;
}

int vfs_dev_close (struct ThreadControl *ctc, struct FileNode *node)
{
	return node->dev->close (node);
}

int vfs_dev_read (struct FileNode *node, void *data, int nbytes)
{
	return node->dev->read (node, data, nbytes);
}

int vfs_dev_write (struct FileNode *node, void *data, int nbytes)
{
	return node->dev->write (node, data, nbytes);
}

int vfs_dev_ioctl_fcntl (struct FileNode *node, unsigned int code, void *data)
{
	return node->dev->ioctl (node, code, data);
}

int vfs_dev_stat (const char *filename, struct stat *st)
{
	// We may get a filename with a slash, skip over it
	if (*filename == '/')
		filename++;

	// Loop through list of devices to find it
	for (struct DeviceDriver *i = &__dev_list_start; i < &__dev_list_end; ++i)
	{
		if (!StrCmp (filename, i->name))
		{
			// It exists, set mode and size to char device
			st->st_mode = S_IFCHR;
			st->st_size = 0;
			return 0;
		}
	}

	// Device not found
	return -1;
}

int vfs_dev_opendir (const char *name)
{
	vfs_dev_curdir = 0;
	return 0;
}

int vfs_dev_closedir (void)
{
	return 0;
}

int vfs_dev_readdir (struct dirent *de)
{
	// Create pointer
	struct DeviceDriver *d = &__dev_list_start;
	d += vfs_dev_curdir;

	// Check
	if (d >= &__dev_list_end)
		return -1;

	// Copy up to VFS_MAX_PATH chars
	for (int i = 0; i < VFS_MAX_PATH; i++)
	{
		de->d_name[i] = d->name[i];

		if (!d->name[i])
			break;
	}

	// Done
	vfs_dev_curdir++;
	return 0;
}

int vfs_dev_statvfs (const char *path, struct statvfs *st)
{
	// Dev does not have any interesting attributes
	MemSet (st, 0, sizeof (struct statvfs));
	st->f_bsize = 512;
	return 0;
}

int vfs_dev_mount (struct FileNode *fnode, char *drvpath)
{
	drvpath[0] = 0;
	return 0;
}

int vfs_dev_unsupported (void)
{
	return -1;
}

///////////////////////////////////////////////////////////////////////////////
// Driver struct
///////////////////////////////////////////////////////////////////////////////
const struct VFSDriver vfs_driver_dev =
{
	.open = vfs_dev_open,
	.close = vfs_dev_close,
	.read = vfs_dev_read,
	.write = vfs_dev_write,
	.ioctl_fcntl = vfs_dev_ioctl_fcntl,
	.lseek = (int(*)(struct FileNode*, int, int)) vfs_dev_unsupported,
	.fstat = (int(*)(struct FileNode*, struct stat*)) vfs_dev_unsupported,
	.stat = vfs_dev_stat,
	.link = (int(*)(const char*, const char*)) vfs_dev_unsupported,
	.unlink = (int(*)(const char*)) vfs_dev_unsupported,
	.rename = (int(*)(const char*, const char*)) vfs_dev_unsupported,
	.mkdir = (int(*)(const char*, int)) vfs_dev_unsupported,
	.opendir = vfs_dev_opendir,
	.closedir = vfs_dev_closedir,
	.readdir = vfs_dev_readdir,
	.statvfs = vfs_dev_statvfs,
	.mount = vfs_dev_mount,
	.unmount = vfs_dev_unsupported
};
