/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "fs/pipe/pipe-internal.h"

///////////////////////////////////////////////////////////////////////////////
// Private functions
///////////////////////////////////////////////////////////////////////////////
int vfs_pipe_cancel (struct ThreadControl *ctc, unsigned int args)
{
	// Get the arguments
	int *r = (int*) ARCH_STACK_ACCESSOR(ctc);

	// Return interrupted call
	r[0] = -1;
	*((int*)r[3]) = EINTR;

	// Unused
	(void)args;
	return 0;
}

int vfs_pipe_check_listitem (pipe_data *p, ListItem *it, int iswrite)
{
	// Extract data pointer and nbytes
	struct ThreadControl *trc = it->parent;
	void *data = (void*) trc->param2;
	int nbytes = (int) trc->param3;

	// Run a regular non-blocking call to move the data
	int n;

	if (iswrite > 0)
		n = vfs_pipe_write_nonblocking (p, data, nbytes);
	else
		n = vfs_pipe_read_nonblocking (p, data, nbytes);

	// Wake the thread if it wrote data
	if (n > 0 || (iswrite < 0 && n >= 0))
	{
		// Wake
		trc->callback = 0;
		SchedState (trc, SCHED_STATUS_RUN, 0, SCHED_STATUS_BLOCKED, 0);

		// Get the arguments off the stack and set return value
		int *r = (int*) ARCH_STACK_ACCESSOR(trc);
		r[0] = n;

		// Since the list was modified, return true
		return 1;
	}

	// If not woken, return that list is not modified
	return 0;
}

int vfs_pipe_check_list (pipe_data *p, List *l, int iswrite)
{
	int more = 1;
	int all = 0;

	// Outer while needed because check_reader will alter the list
	while (more && l->nitems)
	{
		ListItem *c = l->last;
		more = 0;

		for (int i = 0; i < l->nitems; i++, c = c->next)
		{
			if (vfs_pipe_check_listitem (p, c, iswrite))
			{
				more = 1;
				all = 1;
				break;
			}
		}
	}

	return all;
}

void vfs_pipe_check_all (pipe_data *p)
{
	int redo = 1;

	while (redo)
	{
		redo = 0;
		redo |= vfs_pipe_check_list (p, &p->blocked_writers, 1);
		redo |= vfs_pipe_check_list (p, &p->blocked_readers, 0);
	}
}

///////////////////////////////////////////////////////////////////////////////
// Functions
///////////////////////////////////////////////////////////////////////////////
int vfs_pipe_read_blocking (pipe_data *p, void *data, int nbytes)
{
	if (!p->nbytesinpipe)
	{
		// There is no data in the pipe, block caller
		SchedState (currentthc, SCHED_STATUS_BLOCKED, &vfs_pipe_cancel,
			(unsigned int)data, (unsigned int)nbytes);
		ListRemove (&currentthc->listentry);
		ListAdd (&p->blocked_readers, &currentthc->listentry);
		return 0;
	}
	else
	{
		// There is data, read some then check for blocked dependencies
		int n = vfs_pipe_read_nonblocking (p, data, nbytes);
		vfs_pipe_check_all (p);
		return n;
	}
}

int vfs_pipe_write_blocking (pipe_data *p, void *data, int nbytes)
{
	// Not enough bytes in pipe, must block. Write what fits first.
	int n = vfs_pipe_write_nonblocking (p, data, nbytes);
	data += n;
	nbytes -= n;

	// See if we must block
	if (nbytes > 0)
	{
		// Now block the rest of the write
		SchedState (currentthc, SCHED_STATUS_BLOCKED, &vfs_pipe_cancel,
			(unsigned int)data, (unsigned int)nbytes);
		ListRemove (&currentthc->listentry);
		ListAdd (&p->blocked_writers, &currentthc->listentry);
	}

	// May have to wake threads
	vfs_pipe_check_all (p);

	// Done
	return n;
}

void vfs_pipe_wake_eof (pipe_data *p)
{
	int redo = 1;

	while (redo)
		redo = vfs_pipe_check_list (p, &p->blocked_readers, -1);
}
