/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "posix/sys/statvfs.h"
#include "posix/fcntl.h"
#include "fs/dev/dev.h"
#include "print/print.h"
#include "fs/pipe/pipe-internal.h"
#include "ioctl/ioctl.h"

///////////////////////////////////////////////////////////////////////////////
// Variables
///////////////////////////////////////////////////////////////////////////////
pipe_data *vfs_pipe_root;
pipe_data *vfs_pipe_dir;
char vfs_interrupted;

///////////////////////////////////////////////////////////////////////////////
// Helpers
///////////////////////////////////////////////////////////////////////////////
pipe_data *vfs_pipe_findpipe (pipe_data *p, const char *named, unsigned int searchflags)
{
	// Search
	while (p)
	{
		// Get the pipe flags
		unsigned int f = p->flags;

		// Ignore flags other than INUSE and NAMED
		f = f & (PIPE_FLAG_INUSE | PIPE_FLAG_NAMED);

		// Check if the current pipe matches our search
		if (f == searchflags)
		{
			// If we are searching for a named pipe, we also have to check the name
			if (searchflags & PIPE_FLAG_NAMED)
			{
				if (!StrCmp (p->name, named))
					return p;
			}
			else
				return p;
		}

		// Go to the next pipe when it wasn't a match
		p = p->next;
	}

	// No more pipes to search
	return 0;
}

///////////////////////////////////////////////////////////////////////////////
// Functions
///////////////////////////////////////////////////////////////////////////////
int vfs_pipe_open (struct FileNode *node, const char *filename, int flags, int mode)
{
	// Skip forward slash
	if (*filename == '/')
		filename++;

	// Check for invalid path
	for (const char *c = filename; c[0]; c++)
		if (c[0] == '/')
			return -EINVAL;

	// Pointer to found pipe
	pipe_data *p = 0;

	// Flags
	node->flags = flags;
	int f = PIPE_FLAG_INUSE;

	// Check for named pipe
	if (StrCmp (filename, "unnamed"))
	{
		// Not unnamed, search for a matching pipe
		p = vfs_pipe_findpipe (vfs_pipe_root, filename, PIPE_FLAG_INUSE | PIPE_FLAG_NAMED);

		// Set the NAMED flag
		f |= PIPE_FLAG_NAMED;
	}

	// At this point, p is either an existing named pipe or null.
	// If none found, create it.
	if (!p)
	{
		// Find slot
		p = vfs_pipe_allocate (&vfs_pipe_root, filename);

		// Set its flags, accounting for MSG and BLOCKING flags specified
		if (p)
			p->flags = f;
	}

	// If we dont have any pipe now, there was an error
	if (!p)
		return -ENOMEM;

	// Set node members
	node->priv = (void*) p;

	// Increase use counter
	if (!(node->flags & 0x0F))
		p->usecounter_read++;
	else
		p->usecounter_write++;

	// Return OK
	return 0;
}

int vfs_pipe_close (struct ThreadControl *ctc, struct FileNode *node)
{
	pipe_data *p = (pipe_data*) node->priv;

	if (!(node->flags & 0x0F))
		p->usecounter_read--;
	else
		p->usecounter_write--;

	if (!p->usecounter_read && !p->usecounter_write && !(p->flags & PIPE_FLAG_NAMED))
		vfs_pipe_free (&vfs_pipe_root, p);

	if (!p->usecounter_write)
		if (p->blocked_readers.nitems)
			vfs_pipe_wake_eof (p);

	return 0;
}

int vfs_pipe_read (struct FileNode *node, void *data, int nbytes)
{
	if (!nbytes)
		return 0;

	pipe_data *p = (pipe_data*) node->priv;

	if (node->flags & O_NONBLOCK)
		return vfs_pipe_read_nonblocking (p, data, nbytes);
	else
	{
		if (!p->nbytesinpipe && !p->usecounter_write)
			return 0;
		else
			return vfs_pipe_read_blocking (p, data, nbytes);
	}
}

int vfs_pipe_write (struct FileNode *node, void *data, int nbytes)
{
	if (!nbytes)
		return 0;

	pipe_data *p = (pipe_data*) node->priv;

	if (node->flags & O_NONBLOCK)
		return vfs_pipe_write_nonblocking (p, data, nbytes);
	else
		return vfs_pipe_write_blocking (p, data, nbytes);
}

int vfs_pipe_ioctl_fcntl (struct FileNode *node, unsigned int code, void *data)
{
	if (code == F_GETFL)
		return node->flags;
	else if (code == F_SETFL)
	{
		node->flags = ((unsigned int) data) | (node->flags & 0x0F);
		return node->flags;
	}

	return -1;
}

int vfs_pipe_stat (const char *filename, struct stat *st)
{
	if (*filename == '/')
		filename++;

	pipe_data *p = vfs_pipe_findpipe (vfs_pipe_root, filename, PIPE_FLAG_INUSE | PIPE_FLAG_NAMED);

	if (p)
	{
		st->st_mode = S_IFIFO;
		st->st_size = PIPE_BUF_SIZE;
		return 0;
	}

	return -ENOENT;
}

int vfs_pipe_unlink (const char *filename)
{
	// Remove starting slash
	if (*filename == '/')
		filename++;

	// Search for a named pipe in use with the filename
	pipe_data *p = vfs_pipe_findpipe (vfs_pipe_root, filename, PIPE_FLAG_INUSE | PIPE_FLAG_NAMED);

	// If found
	if (p)
	{
		if (!p->usecounter_read && !p->usecounter_write)
		{
			vfs_pipe_free (&vfs_pipe_root, p);
			return 0;
		}
		else
			return -EBUSY;
	}

	// Not found
	return -ENOENT;
}

int vfs_pipe_opendir (const char *name)
{
	vfs_pipe_dir = vfs_pipe_root;
	vfs_interrupted = 0;
	return 0;
}

int vfs_pipe_closedir (void)
{
	return 0;
}

int vfs_pipe_readdir (struct dirent *de)
{
	// Abort if the list was changed
	if (vfs_interrupted)
		return -EINTR;

	// Search for named pipe
	unsigned int mask = PIPE_FLAG_INUSE | PIPE_FLAG_NAMED;

	while (vfs_pipe_dir && (vfs_pipe_dir->flags & mask) != mask)
		vfs_pipe_dir = vfs_pipe_dir->next;

	// Check for end of directory
	if (!vfs_pipe_dir)
		return -1;

	// Copy up to PIPE_NAME_MAXSIZE chars
	// TODO: Buffer length checking
	StrCpy (de->d_name, vfs_pipe_dir->name);

	// Done
	vfs_pipe_dir = vfs_pipe_dir->next;
	return 0;
}

int vfs_pipe_statvfs (const char *path, struct statvfs *st)
{
	MemSet (st, 0, sizeof (struct statvfs));
	st->f_bsize = PIPE_BUF_SIZE;
	return 0;
}

int vfs_pipe_mount (struct FileNode *fnode, char *drvpath)
{
	drvpath[0] = 0;
	vfs_pipe_root = 0;
	vfs_pipe_dir = 0;
	vfs_interrupted = 0;
	return 0;
}

int vfs_pipe_unsupported (void)
{
	return -1;
}

///////////////////////////////////////////////////////////////////////////////
// Driver struct
///////////////////////////////////////////////////////////////////////////////
const struct VFSDriver vfs_driver_pipe =
{
	.open = vfs_pipe_open,
	.close = vfs_pipe_close,
	.read = vfs_pipe_read,
	.write = vfs_pipe_write,
	.ioctl_fcntl = vfs_pipe_ioctl_fcntl,
	.lseek = (int(*)(struct FileNode*, int, int)) vfs_pipe_unsupported,
	.fstat = (int(*)(struct FileNode*, struct stat*)) vfs_pipe_unsupported,
	.stat = vfs_pipe_stat,
	.link = (int(*)(const char*, const char*)) vfs_pipe_unsupported,
	.unlink = vfs_pipe_unlink,
	.rename = (int(*)(const char*, const char*)) vfs_pipe_unsupported,
	.mkdir = (int(*)(const char*, int)) vfs_pipe_unsupported,
	.opendir = vfs_pipe_opendir,
	.closedir = vfs_pipe_closedir,
	.readdir = vfs_pipe_readdir,
	.statvfs = vfs_pipe_statvfs,
	.mount = vfs_pipe_mount,
	.unmount = vfs_pipe_unsupported
};
