/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "mem/string.h"
#include "mem/mem.h"
#include "posix/errno.h"
#include "posix/sys/statvfs.h"
#include "fs/socket/socket-usnet.h"
#include "print/print.h"

///////////////////////////////////////////////////////////////////////////////
// Private functions
///////////////////////////////////////////////////////////////////////////////
struct usnet_queue_item *usnet_queue_pull (struct usnet_internal *usnet)
{
	struct usnet_queue_item *queue = usnet->queue;

	while (queue)
	{
		if (queue->fresh)
			return queue;
		queue = queue->next;
	}

	return 0;
}

void usnet_queue_push (struct usnet_internal *usnet, struct usnet_queue_item *it)
{
	it->next = 0;

	if (!usnet->queue)
		usnet->queue = it;
	else
	{
		struct usnet_queue_item *queue = usnet->queue;

		while (queue->next)
			queue = queue->next;

		queue->next = it;
	}
}

struct usnet_queue_item *usnet_queue_find (struct usnet_internal *usnet, int seq)
{
	struct usnet_queue_item *queue = usnet->queue;

	while (queue)
	{
		if (queue->seq == seq)
			return queue;
		queue = queue->next;
	}

	return 0;
}

void usnet_queue_remove (struct usnet_internal *usnet, struct usnet_queue_item *it)
{
	if (usnet->queue == it)
	{
		usnet->queue = it->next;
	}
	else
	{
		struct usnet_queue_item *queue = usnet->queue;

		while (queue)
		{
			if (queue->next == it)
			{
				queue->next = queue->next->next;
				return;
			}

			queue = queue->next;
		}
	}
}

int usnet_cancel (struct ThreadControl *ctc, unsigned int args)
{
	// Get the arguments
	int *r = (int*) ARCH_STACK_ACCESSOR(ctc);

	// Return interrupted call
	r[0] = -1;
	*((int*)r[3]) = EINTR;

	// Unused
	(void)args;
	return 0;
}

int usnet_init_ioctl (struct FileNode *node, unsigned int code, void *data)
{
	if (code != USNET_INIT_MAGIC)
		return -ENOSYS;

	struct ioctl_init_cmd *cdat = (struct ioctl_init_cmd*) data;

	if (!cdat)
		return -EPERM;
	else if (cdat->version != USNET_INIT_VER)
		return -EINVAL;
	else if (vfs_socket_domain_exists (cdat->domain))
		return -EEXIST;

	struct usnet_internal *usnet = (struct usnet_internal*)
		MemAlloc (sizeof (struct usnet_internal), 0);

	if (!usnet)
		return -ENOMEM;

	MemSet (usnet, 0, sizeof (struct usnet_internal));
	usnet->usnet_stack.bind = usnet_user_bind;
	usnet->usnet_stack.recvfrom = usnet_user_recvfrom;
	usnet->usnet_stack.sendto = usnet_user_sendto;
	usnet->usnet_stack.socket = usnet_user_socket;
	usnet->usnet_stack.listen = usnet_user_listen;
	usnet->usnet_stack.accept = usnet_user_accept;
	usnet->usnet_stack.connect = usnet_user_connect;
	usnet->usnet_stack.sockopt = usnet_user_sockopt;
	usnet->usnet_stack.sockmisc = usnet_user_sockmisc;
	usnet->usnet_stack.netdb = usnet_user_netdb;
	usnet->usnet_stack.close = usnet_user_close;
	usnet->usnet_stack.domain = cdat->domain;
	usnet->usnet_stack.next = 0;
	usnet->queue = 0;

	vfs_socket_register_stack (&usnet->usnet_stack);

	node->priv = usnet;

	return 0;
}
