/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "mem/string.h"
#include "mem/mem.h"
#include "posix/errno.h"
#include "posix/sys/statvfs.h"
#include "posix/fcntl.h"
#include "fs/socket/socket-usnet.h"
#include "syscall/svc-basicfileops.h"
#include "print/print.h"

///////////////////////////////////////////////////////////////////////////////
// Private functions
///////////////////////////////////////////////////////////////////////////////
int usnet_copyhelp (void *dst, void *src, int dstlen, int srclen)
{
	if (!dst || !src || dstlen < 0 || srclen < 0)
		return 0;

	int len = dstlen < srclen ? dstlen : srclen;

	MemCpy (dst, src, len);
	return len;
}

///////////////////////////////////////////////////////////////////////////////
// Network stack register entry points
///////////////////////////////////////////////////////////////////////////////
int usnet_bind (struct FileNode *node, const struct sockaddr *addr, socklen_t addrlen)
{
	DEBUG_ENTRY (Print ("usnet_bind\n"));
	return -ENOSYS;
}

int usnet_recvfrom (struct FileNode *node, struct recvfrom_args *args)
{
	DEBUG_ENTRY (Print ("usnet_recvfrom\n"));

	struct usnet_internal *usnet = (struct usnet_internal*) node->priv;

	if (!usnet)
		return -EINVAL;

	struct usnet_queue_item *i = usnet_queue_pull (usnet);

	if (!i)
	{
		DEBUG_USN_OTHER (Print ("usnet_recvfrom: no items in queue\n"));

		if (node->flags & O_NONBLOCK)
			return -EWOULDBLOCK;
		else
		{
			DEBUG_USN_OTHER (Print ("Blocking owner\n"));
			SchedState (currentthc, SCHED_STATUS_BLOCKED, &usnet_cancel, 0, 0);
			ListRemove (&currentthc->listentry);
			ListAdd (&usnet->blocked_owner, &currentthc->listentry);
			return 0;
		}
	}
	else
	{
		// Cast the type used in the addr field
		struct usnet_queue_item_transfer *tr = (struct usnet_queue_item_transfer*)
			args->addr;

		tr->pid = i->pid;
		tr->fd = i->fd;
		tr->flags = i->recvargs.flags;
		tr->seq = i->seq;
		tr->calltype = i->calltype;

		if (args->addrlen)
			*args->addrlen = sizeof (struct usnet_queue_item_transfer);

		// Return value to use
		int len = 4;

		// Check for extra data
		if (i->calltype == USNET_CALLTYP_BIND)
		{
			// Copy addrdat, but no data
			tr->addrlen = usnet_copyhelp (tr->addrdat, (void*)i->sendargs.addr,
				sizeof(tr->addrdat), i->sendargs.addrlen);
		}
		else if (i->calltype == USNET_CALLTYP_RECVFROM)
		{
			// Data is copied in sendto
		}
		else if (i->calltype == USNET_CALLTYP_SENDTO)
		{
			// Copy data from blocked thread, length silently clamped
			len = usnet_copyhelp (args->buf, (void*) i->sendargs.buf,
				args->len, i->sendargs.len);

			// Copy addr if it is specified
			tr->addrlen = usnet_copyhelp (tr->addrdat, (void*)i->sendargs.addr,
				sizeof(tr->addrdat), i->sendargs.addrlen);
		}
		else if (i->calltype == USNET_CALLTYP_SOCKET)
		{
			// Copy socket arguments into main buffer
			len = usnet_copyhelp (args->buf, i->socket, args->len, sizeof (int) * 3);
		}
		else if (i->calltype == USNET_CALLTYP_CLOSE)
		{
			// No arguments
		}
		else if (i->calltype == USNET_CALLTYP_LISTEN)
		{
			// No data, but copy backlog arguments
			tr->addrdat[0] = i->socket[0];
		}
		else if (i->calltype == USNET_CALLTYP_ACCEPT)
		{
			// Data is copied in sendto, but must supply newfd
			((int*)args->buf)[0] = i->recvargs.flags;
		}
		else if (i->calltype == USNET_CALLTYP_CONNECT)
		{
			// Copy address/addrlen
			tr->addrlen = usnet_copyhelp (tr->addrdat, (void*)i->sendargs.addr,
				sizeof(tr->addrdat), i->sendargs.addrlen);
		}
		else if (i->calltype == USNET_CALLTYP_SOCKOPT)
		{
		}
		else if (i->calltype == USNET_CALLTYP_SOCKMISC)
		{
		}
		else if (i->calltype == USNET_CALLTYP_NETDB)
		{
		}

		DEBUG_USN_OTHER (Print ("usnet_recvfrom: item seq %i calltype %i\n",
			tr->seq, tr->calltype));
		return len;
	}
}

int usnet_sendto (struct FileNode *node, struct sendto_args *args)
{
	DEBUG_ENTRY (Print ("usnet_sendto\n"));

	struct usnet_internal *usnet = (struct usnet_internal*) node->priv;

	struct usnet_queue_item_transfer *tr = (struct usnet_queue_item_transfer*)
			args->addr;

	if (!usnet || !tr)
		return -EINVAL;

	struct usnet_queue_item *i = usnet_queue_find (usnet, tr->seq);

	if (!i)
	{
		DEBUG_USN_OTHER (Print ("usnet_sendto: sequence number not found\n"));
		return 0;
	}

	// Remove job item
	usnet_queue_remove (usnet, i);

	// If trc is null the target thread is no more, just ignore the work
	if (i->trc)
	{
		// Get the arguments off the stack and set return value
		int *r = (int*) ARCH_STACK_ACCESSOR(i->trc);

		// Print call type
		DEBUG_USN_OTHER (Print ("usnet_sendto: call type %i\n", i->calltype));

		// Default return value
		r[0] = tr->returnvalue;

		// Process
		switch (i->calltype)
		{
		case USNET_CALLTYP_BIND:
			break;
		case USNET_CALLTYP_RECVFROM:
			r[0] = usnet_copyhelp ((void*) i->recvargs.buf, (void*) args->buf,
				args->len, i->recvargs.len);

			if (i->recvargs.addrlen)
				i->recvargs.addrlen[0] = usnet_copyhelp (i->recvargs.addr,
					tr->addrdat, i->recvargs.addrlen[0], tr->addrlen);
			break;
		case USNET_CALLTYP_SENDTO:
			// Job finished, only return length
			r[0] = args->len;
			break;
		case USNET_CALLTYP_SOCKET:
			// If the returned file descriptor is not the same as supplied,
			// assume error and close it again.
			if (tr->returnvalue != i->fd)
			{
				int err = 0;
				int args[4] = {i->fd, 0, 0, (int)&err};
				svc_close (i->trc, args);
				//usnet_user_close (i->trc, &nodetable[i->trc->pctrl[i->fd]);
				r[0] = -1;
			}
			break;
		case USNET_CALLTYP_CLOSE:
			break;
		case USNET_CALLTYP_LISTEN:
			// Only the default return value
			break;
		case USNET_CALLTYP_ACCEPT:
			// If the returned file descriptor is not the same as supplied,
			// assume error and close it again.
			if (tr->returnvalue != i->recvargs.flags)
			{
				int err = 0;
				int args[4] = {i->recvargs.flags, 0, 0, (int)&err};
				svc_close (i->trc, args);
				//usnet_user_close (i->trc, &nodetable[i->trc->pctrl[i->fd]]);
				r[0] = -1;
			}
			else
			{
				// Copy the addr
				i->recvargs.addrlen[0] = usnet_copyhelp (i->recvargs.addr,
					tr->addrdat, i->recvargs.addrlen[0], tr->addrlen);
			}
			break;
		case USNET_CALLTYP_CONNECT:
			// Only the default return value
			break;
		case USNET_CALLTYP_SOCKOPT:
		case USNET_CALLTYP_SOCKMISC:
		case USNET_CALLTYP_NETDB:
			break;
		default:
			DEBUG_USN_OTHER (Print ("usnet_sendto: unknown call type\n"));
			r[0] = -1;
		}

		// Unblock
		i->trc->callback = 0;
		SchedState (i->trc, SCHED_STATUS_RUN, 0, SCHED_STATUS_BLOCKED, 0);
	}

	// Delete job item
	MemFree ((unsigned int) i);

	// Done
	return args->len;
}

int usnet_socket (struct FileNode *node, int domain, int type, int protocol)
{
	DEBUG_ENTRY (Print ("usnet_socket\n"));
	return -ENOSYS;
}

int usnet_close (struct ThreadControl *ctc, struct FileNode *node)
{
	DEBUG_ENTRY (Print ("usnet_close\n"));

	struct usnet_internal *usnet = (struct usnet_internal*) node->priv;

	if (usnet)
	{
		// This will run in a worker thread if the stack crashed, so do a crit
		int crit = SchedEnterCritical ();

		// Invalidate stack object
		usnet->isdead = 1;

		// Loop through list to cancel any blocking operations
		struct usnet_queue_item *queue = usnet->queue;

		while (queue)
		{
			if (queue->trc && (queue->trc->state & SCHED_STATUS_BLOCKED))
				SchedState (queue->trc, SCHED_STATUS_RUN, 0, SCHED_STATUS_BLOCKED, 0);

			queue = queue->next;
		}

		// Unregister stack
		vfs_socket_unregister_stack (&usnet->usnet_stack);

		// Force close all references to the stack
		vfs_socket_kill_references (&usnet->usnet_stack);

		// Delete queue items
		queue = usnet->queue;

		while (queue)
		{
			struct usnet_queue_item *next = queue->next;
			MemFree ((unsigned int) queue);
			queue = next;
		}

		// Delete the stack object
		MemFree ((unsigned int) usnet);

		// End critical
		SchedExitCritical (crit);
	}

	return 0;
}

///////////////////////////////////////////////////////////////////////////////
// Declare network stack
///////////////////////////////////////////////////////////////////////////////
struct netstack net_usnet
	= {usnet_bind, usnet_recvfrom, usnet_sendto, usnet_socket,
	SOCKET_UNSUPPORTED_LISTEN,
	SOCKET_UNSUPPORTED_ACCEPT,
	SOCKET_UNSUPPORTED_CONNECT,
	SOCKET_UNSUPPORTED_SOCKOPT,
	SOCKET_UNSUPPORTED_SOCKMISC,
	SOCKET_UNSUPPORTED_NETDB,
	usnet_close, 42};
