/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "mem/string.h"
#include "mem/mem.h"
#include "posix/errno.h"
#include "posix/sys/statvfs.h"
#include "fs/socket/socket-usnet.h"
#include "print/print.h"

///////////////////////////////////////////////////////////////////////////////
// Private functions
///////////////////////////////////////////////////////////////////////////////
void usnet_user_setbasics (struct usnet_internal *usnet, struct usnet_queue_item *qi,
	short fd, struct ThreadControl *ctc)
{
	MemSet (qi, 0, sizeof (struct usnet_queue_item));
	qi->pid = ctc->pctrl->pid;
	qi->fd = fd;
	qi->trc = ctc;
	qi->seq = usnet->seq++;
	qi->fresh = 1;

	// If the current thread is not running this may have been called from
	// close as part of a program terminating, and ctc should be
	// considered invalid once we return.
	if (ctc->state != SCHED_STATUS_RUN)
		qi->trc = 0;
}

int usnet_user_finalize (struct usnet_internal *usnet, struct usnet_queue_item *qi)
{
	// Add to list
	usnet_queue_push (usnet, qi);

	// Block this process and return
	if (qi->trc)
	{
		SchedState (qi->trc, SCHED_STATUS_BLOCKED, &usnet_cancel, 0, 0);
		ListRemove (&qi->trc->listentry);
		ListAdd (&usnet->blocked, &qi->trc->listentry);
	}

	// Check if owner of stack is blocked
	if (usnet->blocked_owner.nitems)
	{
		DEBUG_USN_OTHER (Print ("Unblocking owner\n"));
		struct ThreadControl *trc = usnet->blocked_owner.last->parent;
		SchedState (trc, SCHED_STATUS_RUN, 0, SCHED_STATUS_BLOCKED, 0);
	}

	return 0;
}

///////////////////////////////////////////////////////////////////////////////
// Network stack user entry points
///////////////////////////////////////////////////////////////////////////////
int usnet_user_bind (struct FileNode *node, const struct sockaddr *addr, socklen_t addrlen)
{
	DEBUG_ENTRY (Print ("usnet_user_bind\n"));

	struct usnet_internal *usnet = (struct usnet_internal*) node->net;

	struct usnet_queue_item *qi = (struct usnet_queue_item*)
		MemAlloc (sizeof (struct usnet_queue_item), 0);

	if (!qi)
		return -ENOMEM;
	else
	{
		// Set item members
		usnet_user_setbasics (usnet, qi, node->fd, currentthc);
		qi->calltype = USNET_CALLTYP_BIND;

		qi->sendargs.addr = addr;
		qi->sendargs.addrlen = addrlen;
		qi->sendargs.flags = node->flags;

		// Add to list, block process and return
		return usnet_user_finalize (usnet, qi);
	}
}

int usnet_user_recvfrom (struct FileNode *node, struct recvfrom_args *args)
{
	DEBUG_ENTRY (Print ("usnet_user_recvfrom\n"));

	struct usnet_internal *usnet = (struct usnet_internal*) node->net;

	struct usnet_queue_item *qi = (struct usnet_queue_item*)
		MemAlloc (sizeof (struct usnet_queue_item), 0);

	if (!qi)
		return -ENOMEM;
	else
	{
		// Set item members
		usnet_user_setbasics (usnet, qi, node->fd, currentthc);
		qi->calltype = USNET_CALLTYP_RECVFROM;

		qi->recvargs.buf = args->buf;
		qi->recvargs.len = args->len;
		qi->recvargs.flags = args->flags;
		qi->recvargs.addr = args->addr;
		qi->recvargs.addrlen = args->addrlen;

		// Add to list, block process and return
		return usnet_user_finalize (usnet, qi);
	}
}

int usnet_user_sendto (struct FileNode *node, struct sendto_args *args)
{
	DEBUG_ENTRY (Print ("usnet_user_sendto\n"));

	struct usnet_internal *usnet = (struct usnet_internal*) node->net;

	struct usnet_queue_item *qi = (struct usnet_queue_item*)
		MemAlloc (sizeof (struct usnet_queue_item), 0);

	if (!qi)
		return -ENOMEM;
	else
	{
		// Set item members
		usnet_user_setbasics (usnet, qi, node->fd, currentthc);
		qi->calltype = USNET_CALLTYP_SENDTO;

		qi->sendargs.buf = args->buf;
		qi->sendargs.len = args->len;
		qi->sendargs.flags = args->flags;
		qi->sendargs.addr = args->addr;
		qi->sendargs.addrlen = args->addrlen;

		// Add to list, block process and return
		return usnet_user_finalize (usnet, qi);
	}
}

int usnet_user_socket (struct FileNode *node, int domain, int type, int protocol)
{
	DEBUG_ENTRY (Print ("usnet_user_socket\n"));

	struct usnet_internal *usnet = (struct usnet_internal*) node->net;

	struct usnet_queue_item *qi = (struct usnet_queue_item*)
		MemAlloc (sizeof (struct usnet_queue_item), 0);

	if (!qi)
		return -ENOMEM;
	else
	{
		// Set item members
		usnet_user_setbasics (usnet, qi, node->fd, currentthc);
		qi->calltype = USNET_CALLTYP_SOCKET;
		qi->socket[0] = domain;
		qi->socket[1] = type;
		qi->socket[2] = protocol;

		// Add to list, block process and return
		return usnet_user_finalize (usnet, qi);
	}
}

int usnet_user_listen (struct FileNode *node, void *args)
{
	DEBUG_ENTRY (Print ("usnet_user_listen\n"));

	struct usnet_internal *usnet = (struct usnet_internal*) node->net;

	struct usnet_queue_item *qi = (struct usnet_queue_item*)
		MemAlloc (sizeof (struct usnet_queue_item), 0);

	if (!qi)
		return -ENOMEM;
	else
	{
		// Set item members
		usnet_user_setbasics (usnet, qi, node->fd, currentthc);
		qi->calltype = USNET_CALLTYP_LISTEN;
		qi->socket[0] = ((int*)args)[1]; // backlog

		// Add to list, block process and return
		return usnet_user_finalize (usnet, qi);
	}
}

int usnet_user_accept (struct FileNode *node, struct FileNode *newnode, void *args)
{
	DEBUG_ENTRY (Print ("usnet_user_accept\n"));

	struct usnet_internal *usnet = (struct usnet_internal*) node->net;

	struct usnet_queue_item *qi = (struct usnet_queue_item*)
		MemAlloc (sizeof (struct usnet_queue_item), 0);

	if (!qi)
		return -ENOMEM;
	else
	{
		// Set item members
		usnet_user_setbasics (usnet, qi, node->fd, currentthc);
		qi->calltype = USNET_CALLTYP_ACCEPT;
		qi->recvargs.flags = newnode->fd;						// new file descriptor
		qi->recvargs.addr = (struct sockaddr*) ((int*)args)[1];	// address pointer
		qi->recvargs.addrlen = (socklen_t*) ((int*)args)[2];	// addrlen pointer

		// Add to list, block process and return
		return usnet_user_finalize (usnet, qi);
	}
}

int usnet_user_connect (struct FileNode *node, void *args)
{
	DEBUG_ENTRY (Print ("usnet_user_connect\n"));

	struct usnet_internal *usnet = (struct usnet_internal*) node->net;

	struct usnet_queue_item *qi = (struct usnet_queue_item*)
		MemAlloc (sizeof (struct usnet_queue_item), 0);

	if (!qi)
		return -ENOMEM;
	else
	{
		// Set item members
		usnet_user_setbasics (usnet, qi, node->fd, currentthc);
		qi->calltype = USNET_CALLTYP_CONNECT;
		qi->sendargs.addr = (const struct sockaddr*) ((int*)args)[1];	// address
		qi->sendargs.addrlen = (socklen_t) ((int*)args)[2];				// addrlen

		// Add to list, block process and return
		return usnet_user_finalize (usnet, qi);
	}
}

int usnet_user_sockopt (struct FileNode *node, void *args)
{
	DEBUG_ENTRY (Print ("usnet_user_sockopt\n"));

	struct usnet_internal *usnet = (struct usnet_internal*) node->net;

	struct usnet_queue_item *qi = (struct usnet_queue_item*)
		MemAlloc (sizeof (struct usnet_queue_item), 0);

	if (!qi)
		return -ENOMEM;
	else
	{
		// Set item members
		usnet_user_setbasics (usnet, qi, node->fd, currentthc);
		qi->calltype = USNET_CALLTYP_SOCKOPT;
		qi->socket[0] = ((int*)args)[1]; // call type
		qi->socket[1] = ((int*)args)[2]; // arguments pointer

		// Add to list, block process and return
		return usnet_user_finalize (usnet, qi);
	}
}

int usnet_user_sockmisc (struct FileNode *node, void *args)
{
	DEBUG_ENTRY (Print ("usnet_user_sockmisc\n"));

	struct usnet_internal *usnet = (struct usnet_internal*) node->net;

	struct usnet_queue_item *qi = (struct usnet_queue_item*)
		MemAlloc (sizeof (struct usnet_queue_item), 0);

	if (!qi)
		return -ENOMEM;
	else
	{
		// Set item members
		usnet_user_setbasics (usnet, qi, node->fd, currentthc);
		qi->calltype = USNET_CALLTYP_SOCKMISC;
		qi->socket[0] = ((int*)args)[1]; // call type
		qi->socket[1] = ((int*)args)[2]; // arguments pointer

		// Add to list, block process and return
		return usnet_user_finalize (usnet, qi);
	}
}

int usnet_user_netdb (struct FileNode *node, void *args)
{
	DEBUG_ENTRY (Print ("usnet_user_netdb\n"));

	struct usnet_internal *usnet = (struct usnet_internal*) node->net;

	struct usnet_queue_item *qi = (struct usnet_queue_item*)
		MemAlloc (sizeof (struct usnet_queue_item), 0);

	if (!qi)
		return -ENOMEM;
	else
	{
		// Set item members
		usnet_user_setbasics (usnet, qi, node->fd, currentthc);
		qi->calltype = USNET_CALLTYP_NETDB;
		qi->socket[0] = ((int*)args)[1]; // call type
		qi->socket[1] = ((int*)args)[2]; // arguments pointer

		// Add to list, block process and return
		return usnet_user_finalize (usnet, qi);
	}
}

int usnet_user_close (struct ThreadControl *ctc, struct FileNode *node)
{
	DEBUG_ENTRY (Print ("usnet_user_close\n"));

	struct usnet_internal *usnet = (struct usnet_internal*) node->net;

	// If the stack is killed, let usnet_close handle it
	if (usnet->isdead)
		return 0;

	// If terminated, check for earlier wait objects and invalidate
	if (ctc->state != SCHED_STATUS_RUN)
	{
		struct usnet_queue_item *queue = usnet->queue;

		while (queue)
		{
			if (queue->pid == ctc->pctrl->pid &&
				queue->fd == node->fd)
				queue->trc = 0;

			queue = queue->next;
		}
	}

	// Create and set a new queue item
	struct usnet_queue_item *qi = (struct usnet_queue_item*)
		MemAlloc (sizeof (struct usnet_queue_item), 0);

	if (!qi)
		return -ENOMEM;
	else
	{
		// Set item members
		usnet_user_setbasics (usnet, qi, node->fd, ctc);
		qi->calltype = USNET_CALLTYP_CLOSE;

		// Add to list, block process and return
		return usnet_user_finalize (usnet, qi);
	}
}
