/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "mem/string.h"
#include "posix/errno.h"
#include "posix/sys/statvfs.h"
#include "fs/socket/socket.h"
#include "fs/socket/socket-usnet.h"
#include "process/process-allocate.h"
#include "syscall/svc-basicfileops.h"
#include "print/print.h"

///////////////////////////////////////////////////////////////////////////////
// Variables
///////////////////////////////////////////////////////////////////////////////
char vfs_socket_dir;
char vfs_socket_interrupted;
struct netstack *vfs_socket_registered_stacks;

///////////////////////////////////////////////////////////////////////////////
// External variables
///////////////////////////////////////////////////////////////////////////////
extern struct netstack __net_list_start;
extern struct netstack __net_list_end;

///////////////////////////////////////////////////////////////////////////////
// "internal" socket api
///////////////////////////////////////////////////////////////////////////////
int vfs_socket_bind (struct FileNode *node, const struct sockaddr *addr, socklen_t addrlen)
{
	return node->net->bind (node, addr, addrlen);
}

int vfs_socket_recvfrom (struct FileNode *node, struct recvfrom_args *args)
{
	return node->net->recvfrom (node, args);
}

int vfs_socket_sendto (struct FileNode *node, struct sendto_args *args)
{
	return node->net->sendto (node, args);
}

int vfs_socket_socket (struct FileNode *node, int domain, int type, int protocol)
{
	// Set the driver to this vfs
	node->fs = &vfs_driver_socket;

	// Find the stack responsible for the domain
	for (struct netstack *i = &__net_list_start; i < &__net_list_end; ++i)
	{
		if (i->domain == domain)
		{
			// Found match, store pointer and try opening
			node->net = i;
			return i->socket (node, domain, type, protocol);
		}
	}

	// Check dynamically created stacks
	for (struct netstack *i = vfs_socket_registered_stacks; i; i = i->next)
	{
		if (i->domain == domain)
		{
			node->net = i;
			node->priv = i;
			return i->socket (node, domain, type, protocol);
		}
	}

	return -ENOSYS;
}

int vfs_socket_listen (struct FileNode *node, void *args)
{
	return node->net->listen (node, args);
}

int vfs_socket_accept (struct FileNode *node, struct FileNode *newnode, void *args)
{
	// Copy file system and network stack pointers before running accept
	newnode->fs = node->fs;
	newnode->net = node->net;
	newnode->priv = 0;
	return node->net->accept (node, newnode, args);
}

int vfs_socket_connect (struct FileNode *node, void *args)
{
	return node->net->connect (node, args);
}

int vfs_socket_sockopt (struct FileNode *node, void *args)
{
	return node->net->sockopt (node, args);
}

int vfs_socket_sockmisc (struct FileNode *node, void *args)
{
	return node->net->sockmisc (node, args);
}

int vfs_socket_netdb (struct FileNode *node, void *args)
{
	return node->net->netdb (node, args);
}

int vfs_socket_domain_exists (int domain)
{
	for (struct netstack *i = &__net_list_start; i < &__net_list_end; ++i)
		if (i->domain == domain)
			return 1;

	for (struct netstack *i = vfs_socket_registered_stacks; i; i = i->next)
		if (i->domain == domain)
			return 1;

	return 0;
}

int vfs_socket_register_stack (struct netstack *stack)
{
	if (!vfs_socket_registered_stacks)
		vfs_socket_registered_stacks = stack;
	else
	{
		struct netstack *i = vfs_socket_registered_stacks;
		while (i->next) i = i->next;
		i->next = stack;
	}

	return 0;
}

int vfs_socket_unregister_stack (struct netstack *stack)
{
	// Remove from stack list
	if (vfs_socket_registered_stacks == stack)
		vfs_socket_registered_stacks = stack->next;
	else
	{
		struct netstack *i = vfs_socket_registered_stacks;
		while (i->next != stack) i = i->next;
		i->next = stack->next;
	}

	// Done
	return 0;
}

void vfs_socket_kill_references (struct netstack *stack)
{
	for (int i = 0; i < MAX_PROCESSDATA; i++)
	{
		if (processdata[i].inuse)
		{
			for (int j = 0; j < MAX_NODES; j++)
			{
				int kfd = processdata[i].fnodes[j];

				if (kfd != 255 && nodetable[kfd].net == stack)
				{
					int err;
					int args[4] = {j, 0, 0, (int) &err};
					svc_close (&processdata[i].trc, args);
				}
			}
		}
	}
}

///////////////////////////////////////////////////////////////////////////////
// Functions
///////////////////////////////////////////////////////////////////////////////
int vfs_socket_open (struct FileNode *node, const char *filename, int flags, int mode)
{
	// Skip forward slash
	if (*filename == '/')
		filename++;

	// TODO: Use this to create named sockets

	// Process requests to USNET_INIT_PATH
	if (!StrCmp (filename, USNET_INIT_PATH))
	{
		// Set the driver to this vfs
		node->fs = &vfs_driver_socket;
		node->net = &net_usnet;
		node->priv = 0;
		return 0;
	}
	else
		return -ENOSYS;
}

int vfs_socket_close (struct ThreadControl *ctc, struct FileNode *node)
{
	return node->net->close (ctc, node);
}

int vfs_socket_read (struct FileNode *node, void *data, int nbytes)
{
	struct recvfrom_args a;
	a.buf = data;
	a.len = nbytes;
	a.flags = 0;
	a.addr = 0;
	a.addrlen = 0;

	return vfs_socket_recvfrom (node, &a);
}

int vfs_socket_write (struct FileNode *node, void *data, int nbytes)
{
	struct sendto_args a;
	a.buf = data;
	a.len = nbytes;
	a.flags = 0;
	a.addr = 0;
	a.addrlen = 0;

	return vfs_socket_sendto (node, &a);
}

int vfs_socket_ioctl_fcntl (struct FileNode *node, unsigned int code, void *data)
{
	if (node->net == &net_usnet)
		return usnet_init_ioctl (node, code, data);
	else
		return -ENOSYS;
}

int vfs_socket_stat (const char *filename, struct stat *st)
{
	// TODO
	filename++;

	if (!StrCmp (filename, USNET_INIT_PATH))
	{
		MemSet (st, 0, sizeof (struct stat));
		st->st_mode = S_IFSOCK;
		return 0;
	}
	else
		return -ENOENT;
}

int vfs_socket_unlink (const char *filename)
{
	// TODO
	Print ("vfs_socket_unlink\n");
	return -ENOSYS;
}

int vfs_socket_opendir (const char *name)
{
	// TODO: List actual sockets
	if (vfs_socket_dir)
		return -EBUSY;

	vfs_socket_dir = 1;
	return 0;
}

int vfs_socket_closedir (void)
{
	vfs_socket_dir = 0;
	return 0;
}

int vfs_socket_readdir (struct dirent *de)
{
	// TODO: List actual sockets
	if (vfs_socket_interrupted)
		return -EINTR;

	if (vfs_socket_dir == 1)
	{
		StrCpy (de->d_name, USNET_INIT_PATH);
		vfs_socket_dir++;
		return 0;
	}
	else
		return -1;
}

int vfs_socket_statvfs (const char *path, struct statvfs *st)
{
	MemSet (st, 0, sizeof (struct statvfs));
	return 0;
}

int vfs_socket_mount (struct FileNode *fnode, char *drvpath)
{
	drvpath[0] = 0;
	vfs_socket_dir = 0;
	vfs_socket_interrupted = 0;
	return 0;
}

int vfs_socket_unsupported (void)
{
	return -ENOSYS;
}

///////////////////////////////////////////////////////////////////////////////
// Driver struct
///////////////////////////////////////////////////////////////////////////////
const struct VFSDriver vfs_driver_socket =
{
	.open = vfs_socket_open,
	.close = vfs_socket_close,
	.read = vfs_socket_read,
	.write = vfs_socket_write,
	.ioctl_fcntl = vfs_socket_ioctl_fcntl,
	.lseek = (int(*)(struct FileNode*, int, int)) vfs_socket_unsupported,
	.fstat = (int(*)(struct FileNode*, struct stat*)) vfs_socket_unsupported,
	.stat = vfs_socket_stat,
	.link = (int(*)(const char*, const char*)) vfs_socket_unsupported,
	.unlink = vfs_socket_unlink,
	.rename = (int(*)(const char*, const char*)) vfs_socket_unsupported,
	.mkdir = (int(*)(const char*, int)) vfs_socket_unsupported,
	.opendir = vfs_socket_opendir,
	.closedir = vfs_socket_closedir,
	.readdir = vfs_socket_readdir,
	.statvfs = vfs_socket_statvfs,
	.mount = vfs_socket_mount,
	.unmount = vfs_socket_unsupported
};
