/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "process/process.h"
#include "fs/vfs.h"
#include "fs/dev/dev.h"
#include "fs/pipe/pipe.h"
#include "fs/socket/socket.h"
#include "syscall/svc-basicfileops.h"
#include "syscall/svc-procops.h"
#include "print/print.h"
#include "kmod/kmod.h"
#include "posix/spawn.h"
#include "posix/unistd.h"
#include "posix/sys/wait.h"
#include "panic/panic.h"
#include "worker/worker.h"
#include "platform/platform.h"

///////////////////////////////////////////////////////////////////////////////
// Variables
///////////////////////////////////////////////////////////////////////////////
struct ProcessControl *prc;

///////////////////////////////////////////////////////////////////////////////
// Forward declarations
///////////////////////////////////////////////////////////////////////////////
void main (void);

///////////////////////////////////////////////////////////////////////////////
// Functions
///////////////////////////////////////////////////////////////////////////////
void InitProcess (void)
{
	// Re-open standard io
#if INIT_REOPENSTDIO
	close (1);
	close (0);
	open (INIT_REOPEN_FILENAME, 0);
	int fakeerrno = 0;
	int dup2args[4] = {0, 1, 0, (int) &fakeerrno};
	svc_dup2 (&prc->trc, dup2args);
#endif

	// Mount pipe
	vfs_mount ("/pipe", (struct VFSDriver*) &vfs_driver_pipe, 0);

	// Mount socket
	vfs_mount ("/socket", (struct VFSDriver*) &vfs_driver_socket, 0);

	// Initialize kernel modules
	for (struct KernelModule *i = &__kmod_list_start; i < &__kmod_list_end; ++i)
		i->init ();

	// Start kernel worker
	WorkerStart (prc);

	// For minimal configurations, go to a custom main func
#if INIT_CUSTOM_MINIMAL_MAIN == 1
	main ();
#endif

	// Start root shell
#if INIT_CUSTOM_MINIMAL_MAIN == 0
	pid_t pid;
	posix_spawnattr_t spawnattr = {POSIX_SPAWN_SETPGROUP, 0};
	char *argv[] = {"sh", 0};
	char *env[] = {"PATH=/bin", 0};
	int status;

	while (1)
	{
		if (posix_spawn (&pid, "/bin/sh", 0, &spawnattr, argv, env))
			Panic ("No shell\n");

		tcsetpgrp (0, pid);

		waitpid (pid, &status, 0);

		Print ("Root sh terminated, restarting\n");
	}
#endif
}

void InitInitialize (void)
{
	// Start init as the first proper process
	static unsigned int initstack[192];
	prc = ProcessCreateInternal (&InitProcess, initstack, 192);

	// Mount dev
	vfs_mount ("/dev", (struct VFSDriver*) &vfs_driver_dev, 0);

	// Open tty0
	int fakeerrno = 0;
	const char openfile[] = INIT_DEFAULT_TTY;
	int openargs[4] = {(int)openfile, 0, 0, (int) &fakeerrno};
	svc_open (&prc->trc, openargs);

	// dup2 0 -> 1
	int dup2args[4] = {0, 1, 0, (int) &fakeerrno};
	svc_dup2 (&prc->trc, dup2args);

	// Initialize memory
	PlatformInitMemory ();
}
