/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "mem/mem.h"
#include "mem/string.h"
#include "posix/errno.h"
#include "mpu/mpu.h"
#include "../../config/live/platform.h"

///////////////////////////////////////////////////////////////////////////////
// Functions
///////////////////////////////////////////////////////////////////////////////
int MpuSpawnAllocate (struct ProcessControl *slot, struct Loader *loader,
	unsigned int argvsize, unsigned int envsize, unsigned int stacksize,
	struct Program *p)
{
	// Calculate size of allocation regions
	unsigned int rs0 = loader->text.size;
	unsigned int rs1 = stacksize;
	unsigned int rs2 = loader->rodata.size;
	unsigned int rs3 = loader->data.size + loader->bss.size + argvsize + envsize;

	// Data
	unsigned int *mpuregs = slot->trc.mpuregs;
	struct MemMpuReturn mpu;

	// Allocate text from bank 0
	slot->alloc[0] = MemAllocMpu (rs0, 0, &mpu);
	mpuregs[0] = mpu.mpu_rbar;
	mpuregs[1] = mpu.mpu_rasr;
	p->text = (unsigned int*) mpu.dataptr;
	slot->text = mpu.dataptr;

	// Allocate stack from bank 0
	slot->alloc[1] = MemAllocMpu (rs1, 0, &mpu);
	mpuregs[2] = mpu.mpu_rbar;
	mpuregs[3] = mpu.mpu_rasr;
	slot->stack = (unsigned int*) mpu.dataptr;

	// Allocate rodata from bank 0
	slot->alloc[2] = MemAllocMpu (rs2, 0, &mpu);
	mpuregs[4] = mpu.mpu_rbar;
	mpuregs[5] = mpu.mpu_rasr;
	p->rodata = (unsigned int*) mpu.dataptr;

	// Allocate data, bss, argv and environ from bank 0
	slot->alloc[3] = MemAllocMpu (rs3, 0, &mpu);
	mpuregs[6] = mpu.mpu_rbar;
	mpuregs[7] = mpu.mpu_rasr;
	unsigned int allocptr = mpu.dataptr;

	// Check allocations for success now, will be freed later
	if (!slot->alloc[0] || !slot->alloc[1] || !slot->alloc[2] || !slot->alloc[3])
		return -ENOMEM;

	// Clear the data regions
	MemSet (p->text, 0, rs0);
	MemSet (slot->stack, 0, rs1);
	MemSet (p->rodata, 0, rs2);
	MemSet ((int*)allocptr, 0, rs3);

	// Set data, bss, argv and environ
	p->data = (unsigned int*) allocptr;
	allocptr += loader->data.size;

	p->bss = (unsigned int*) allocptr;
	allocptr += loader->bss.size;

	slot->argv = (char**) allocptr;
	allocptr += argvsize;

	slot->environ = (char**) allocptr;

	// Adjust entry point
	loader->entrypoint += (unsigned int) p->text;

	// Enable auto set of regions
	mpuregs[0] |= 0x00000010 | 0;
	mpuregs[2] |= 0x00000010 | 1;
	mpuregs[4] |= 0x00000010 | 2;
	mpuregs[6] |= 0x00000010 | 3;

	// Set region access
	mpuregs[1] |=      AP(2) | TEX(0) |C  |E; // Exec,    RO, Outer write back, Cacheable, Write back, Shareable
	mpuregs[3] |= XN | AP(3) | TEX(0) |C|S|E; // No exec, RW, Outer write back, Cacheable, Write back, Shareable
	mpuregs[5] |= XN | AP(2) | TEX(0) |C|S|E; // No exec, RO, Outer write back, Cacheable, Write back, Shareable
	mpuregs[7] |= XN | AP(3) | TEX(0) |C|S|E; // No exec, RW, Outer write back, Cacheable, Write back, Shareable

	// Cortex-R has a different register layout
	if (MPU_DO_EXTRA_PROCESS == 1)
		PlatformMpuRegisterSetup (mpuregs);

	// Set the two extra regions to the same as the .bss region
	if (PAGER_ENABLE == 1)
		PlatformPagerRegisterSetup (mpuregs);

	// Done
	return 0;
}
