/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "mem/mem.h"
#include "posix/errno.h"
#include "print/print.h"
#include "mpu/pager.h"
#include "mpu/mpu.h"
#include "../config/live/platform.h"

///////////////////////////////////////////////////////////////////////////////
// Definitions
///////////////////////////////////////////////////////////////////////////////
#if 0
#define DODEBUG(cmd) cmd
#else
#define DODEBUG(cmd)
#endif

///////////////////////////////////////////////////////////////////////////////
// Structures
///////////////////////////////////////////////////////////////////////////////
struct PageTableEntry
{
	unsigned int tid;		// Thread ID (same as PID for now), -1 = unused
	unsigned int useptr;	// Pointer where the data is (accounting for subregions)
	unsigned int maxptr;	// Highest address allowable + 1
	unsigned int mpu_rbar;	// Region base address register
	unsigned int mpu_rasr;	// Region attribute and size register
	unsigned int freeptr;	// Pointer to use when freeing memory
};

struct PageTable
{
	struct PageTableEntry entries[PAGER_MAX_ENTRIES];
	struct PageTable *next;
};

///////////////////////////////////////////////////////////////////////////////
// Variables
///////////////////////////////////////////////////////////////////////////////
struct PageTable pagetable;
int pagermpuregion = PAGER_REGION_1;

///////////////////////////////////////////////////////////////////////////////
// Private Functions
///////////////////////////////////////////////////////////////////////////////
inline unsigned int *PagerWalkHandle (struct PageTableEntry *e,
	struct ThreadControl *trc, 	unsigned int *out, unsigned int dfsr)
{
	// Check read/write permission
	if (dfsr & (1 << 11))
	{
		if (!(AP(2) & e->mpu_rasr))
		{
			DODEBUG (Print ("No read access\n"));
			return 0;
		}
	}
	else
	{
		if ((AP(3) & e->mpu_rasr) != AP(3))
		{
			DODEBUG (Print ("No write access\n"));
			return 0;
		}
	}

	// Store output for setting the MPU registers
	out[0] = e->mpu_rbar & 0xFFFFFFE0;
	out[1] = e->mpu_rasr & 0x0000FFFF;
	out[2] = (e->mpu_rasr >> 16);

	// Also store in the ThreadControl structure
	if (pagermpuregion == PAGER_REGION_1)
	{
		trc->mpuregs[PAGER_MPUREGS_1 + 0] = out[0];
		trc->mpuregs[PAGER_MPUREGS_1 + 1] = out[1];
		trc->mpuregs[PAGER_MPUREGS_1 + 2] = out[2];
		pagermpuregion = PAGER_REGION_2;
	}
	else
	{
		trc->mpuregs[PAGER_MPUREGS_2 + 0] = out[0];
		trc->mpuregs[PAGER_MPUREGS_2 + 1] = out[1];
		trc->mpuregs[PAGER_MPUREGS_2 + 2] = out[2];
		pagermpuregion = PAGER_REGION_1;
	}

	// Done
	DODEBUG (Print ("PagerWalk %i %x %x %x\n", trc->pctrl->pid, out[0], out[1], out[2]));
	return out;
}

unsigned int PagerCalculateMaxAddress (unsigned int base, unsigned int rbar, unsigned int rasr)
{
	unsigned int regionsizeexp = (rasr >> 1) & 0x0000001F;
	unsigned int regionsize = 1 << (regionsizeexp + 1);

	if (regionsize < 256)
		return base + regionsize;
	else
	{
		unsigned int subregionsize = regionsize / 8;

		/*
		 * The enabled subregions are always continuous and the base address
		 * is already set to the first enabled subregion. Can therefore simply
		 * loop over the bits and add the subregion size.
		 */
		for (int i = (1 << 8); i < (1 << 16); i = i << 1)
			if (!(rasr & i))
				base += subregionsize;

		return base;
	}
}

///////////////////////////////////////////////////////////////////////////////
// Public Functions
///////////////////////////////////////////////////////////////////////////////
void PagerInitialize (void)
{
	for (int i = 0; i < PAGER_MAX_ENTRIES; i++)
		pagetable.entries[i].tid = -1;

	pagetable.next = 0;
}

void PagerCleanupByPID (unsigned int tid)
{
	struct PageTable *table = &pagetable;

	do
	{
		for (int i = 0; i < PAGER_MAX_ENTRIES; i++)
		{
			if (table->entries[i].tid == tid)
			{
				table->entries[i].tid = -1;
				MemFree (table->entries[i].freeptr);
			}
		}
	} while ((table = table->next));
}

int PagerAddEntry (unsigned int tid, unsigned int rbar, unsigned int rasr,
	unsigned int useptr, unsigned int freeptr)
{
	struct PageTable *table = &pagetable;

	do
	{
		for (int i = 0; i < PAGER_MAX_ENTRIES; i++)
		{
			if (table->entries[i].tid == -1)
			{
				table->entries[i].tid = tid;
				table->entries[i].useptr = useptr;
				table->entries[i].maxptr = PagerCalculateMaxAddress (useptr, rbar, rasr);
				table->entries[i].mpu_rbar = rbar;
				table->entries[i].mpu_rasr = rasr;
				table->entries[i].freeptr = freeptr;
				return 0;
			}
		}
	} while ((table = table->next));

	return -ENOMEM;
}

int PagerFreeEntry (unsigned int tid, unsigned int useptr)
{
	struct PageTable *table = &pagetable;

	do
	{
		for (int i = 0; i < PAGER_MAX_ENTRIES; i++)
		{
			if (table->entries[i].tid == tid && table->entries[i].useptr == useptr)
			{
				table->entries[i].tid = -1;
				MemFree (table->entries[i].freeptr);
				return 0;
			}
		}
	} while ((table = table->next));

	return -1;
}

unsigned int *PagerWalk (unsigned int address, struct ThreadControl *trc,
	unsigned int *out, unsigned int dfsr)
{
	DODEBUG (Print ("PagerWalk 0x%x pid %i\n", address, trc->pctrl->pid));

	struct PageTable *table = &pagetable;
	unsigned int tid = trc->pctrl->pid;

	do
	{
		for (int i = 0; i < PAGER_MAX_ENTRIES; i++)
		{
			if (table->entries[i].tid == tid &&
				address >= table->entries[i].useptr &&
				address < table->entries[i].maxptr)
			{
				return PagerWalkHandle (&table->entries[i], trc, out, dfsr);
			}
		}
	} while ((table = table->next));

	DODEBUG (Print ("PagerWalk no entry\n"));
	return 0;
}
