/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "ioctl/ioctl.h"
#include "posix/fcntl.h"
#include "mem/string.h"
#include "print/print.h"
#include "drivers/tty/tty-stm32.h"
#include "../config/live/network-config.h"
#include "net/enet/layer2/interface.h"
#include "net/enet/layer2/framepool.h"
#include "net/enet/layer2/manager.h"
#include "net/enet/interfaces/ktty.h"

///////////////////////////////////////////////////////////////////////////////
// Debug definitions
///////////////////////////////////////////////////////////////////////////////
#if NETWORK_DEBUG_INTERFACE
	#define DEBUG(str) str
#else
	#define DEBUG(str)
#endif

///////////////////////////////////////////////////////////////////////////////
// Functions
///////////////////////////////////////////////////////////////////////////////
int dev_tty_open (struct FileNode *node, const char *filename);
int dev_tty_close (struct FileNode *node);
int dev_tty_read (struct FileNode *node, void *data, unsigned int nbytes);
int dev_tty_write (struct FileNode *node, void *data, unsigned int nbytes);
int dev_tty_ioctl (struct FileNode *node, unsigned int code, void *data);

///////////////////////////////////////////////////////////////////////////////
// Functions
///////////////////////////////////////////////////////////////////////////////
void KTTYInit (Layer2KTTY *t, char *filename)
{
	// Clear object
	MemSet (t, 0, sizeof (Layer2KTTY));

	// Open tty
	t->fd = dev_tty_open (&t->ttynode, filename);

	if (t->fd < 0)
	{
		IFDEBUG_OTHER (NETWORK_PRINTS ("Could not open port\n"));
		return;
	}
}

void Layer2Init (Layer2 *l)
{
	// Cast
	Layer2KTTY *t = (Layer2KTTY*) l;

	// Initialize link speed
	Layer2LinkspeedConstructor (&t->ls, l);

	// Set raw mode (buffered transmit, direct read and direct irq)
	dev_tty_ioctl (&t->ttynode, TTYSETMODE, (void*)2);

	// Set circular reception mode
	dev_tty_ioctl (&t->ttynode, TTYSETRXCIRC, 0);

	// Start the actual DMA transfer
	dev_tty_read (&t->ttynode, &l->rxbuffer[0], RS485_RXBUFF_SIZE);

	// Set default baud rate
	Layer2SetSpeed (l, 19200);

	// Store interface ID
	l->interfaceid = l;
}

void Layer2Pump (Layer2 *l)
{
	// Cast
	Layer2KTTY *t = (Layer2KTTY*) l;

	// Get the current write position from the tty driver
	l->writepos = dev_tty_ioctl (&t->ttynode, TTYGETWRPOS, 0);

	// Check if we completed any packets
	DataFrame *frame;

	while ((frame = Layer2CheckPos (l)))
	{
		if (t->errcnt)
			t->errcnt--;

		if (frame->destination == LAYER2LINKSPEED_DSTADDRESS)
			Layer2LinkspeedProcess (&t->ls, frame);
		else if (!t->ls.state)
			Layer2StoreFifo (l, frame);
		else
			FramepoolDelete (frame);
	}

	// Check link speed
	Layer2LinkspeedUpdate (&t->ls);
}

int Layer2SendFrameInternal (Layer2 *l, DataFrame *frame)
{
	// Cast
	Layer2KTTY *t = (Layer2KTTY*) l;

	// Check that the file descriptor is valid
	if (t->fd < 0)
	{
		FramepoolDelete(frame);
		return 0;
	}

	// Write all data
	int len = frame->size;
	int n = len;
	int pos = 0;
	char *str = (char*) frame;

	do
	{
		int diff = dev_tty_write (&t->ttynode, str + pos, n);

		if (diff < 0)
			break;

		pos += diff;
		n -= diff;
	} while (pos != len);

	// Free the frame memory
	FramepoolDelete(frame);

	// Increase statistic values
	IFSTATS (l->stats.framessent++);
	IFSTATS (l->stats.bytessent += frame->size);

	// Done
	return 0;
}

void Layer2SetSpeed (Layer2 *l, unsigned int speed)
{
	// Set speed using termios
	struct termios term;
	MemSet (&term, 0, sizeof (struct termios));
	term.baudrate = speed;

	Layer2KTTY *t = (Layer2KTTY*) l;
	dev_tty_ioctl (&t->ttynode, TCSETS, (void*)&term);
}

int Layer2IsUp (Layer2 *l)
{
	// Use getrxbusy ioctl to check for errors
	Layer2KTTY *t = (Layer2KTTY*) l;
	int ret = dev_tty_ioctl (&t->ttynode, TTYGETRXBUSY, 0);

	// Value 2 is an error, bring the link down
	if (ret == 2)
	{
		DEBUG (Print ("KTTY Error detected\n"));
		t->errcnt++;

		if (t->errcnt == 3)
		{
			t->errcnt = 0;
			l->linkstate = 0;
		}
	}

	// Return linkstate, make it 1 from here on
	unsigned char old = l->linkstate;
	l->linkstate = 1;
	return old;
}

int Layer2LinkUp (Layer2 *l)
{
	// Check state
	return Layer2LinkspeedLinkIsReady (&((Layer2KTTY*) l)->ls);
}
