/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "mem/string.h"
#include "net/enet/layer2/manager.h"
#include "net/enet/layer2/framepool.h"
#include "net/enet/layer2/checksum.h"

///////////////////////////////////////////////////////////////////////////////
// Definitions
///////////////////////////////////////////////////////////////////////////////
#define MANAGER_HAVE_ASSERT 0

///////////////////////////////////////////////////////////////////////////////
// Variables
///////////////////////////////////////////////////////////////////////////////

// Global receive fifo for pointers
DataFrame *rxframes[RS485_RXFIFO_SIZE];

// Indexes into the fifo
unsigned char rxframesread;
unsigned char rxframeswrite;

///////////////////////////////////////////////////////////////////////////////
// Helpers
///////////////////////////////////////////////////////////////////////////////
/*
 * Using this for easier debugging
 */
void memcpy_helper (unsigned char *dst, unsigned char *src, unsigned int n)
{
	while (n)
	{
		*dst = *src;
		dst++;
		src++;
		n--;
	}
}

///////////////////////////////////////////////////////////////////////////////
// Functions
///////////////////////////////////////////////////////////////////////////////
void Layer2GlobalInit (void)
{
	// Clear the pointer array and indexes
	MemSet (rxframes, 0, sizeof (rxframes));
	rxframesread = 0;
	rxframeswrite = 0;
}

DataFrame *Layer2GlobalPull (void)
{
	// Fetch the pointer at the read position
	DataFrame *frame = rxframes[rxframesread];

	// If there is anything at the read position
	if (frame)
	{
		// Clear position in the array
		rxframes[rxframesread] = 0;

		// Increment the read position with wrap-around
		if ((++rxframesread) == RS485_RXFIFO_SIZE)
			rxframesread = 0;
	}

	// Done
	return frame;
}

/*
 * Extract data from readfrompos (including) to readtopos (including)
 * readtopos points to the end-of-packet symbol, and readfrompos
 * points to the first byte in the packet.
 */
DataFrame *Layer2Extract (Layer2 *l)
{
	// Some assertions
#if MANAGER_HAVE_ASSERT
	if (l->readfrompos < 0 || l->readfrompos >= RS485_RXBUFF_SIZE)
		Panic ("readfrompos");
	if (l->readtopos < 0 || l->readtopos >= RS485_RXBUFF_SIZE)
		Panic ("readtopos");
	if (l->writepos < 0 || l->writepos >= RS485_RXBUFF_SIZE)
		Panic ("writepos");
#endif

	// Allocate frame
	DataFrame *frame = FramepoolNew ();

	// Did it work?
	if (frame)
	{
		// Number of bytes to memcpy
		unsigned int count;

		// If the packet wraps around the end
		if (l->readtopos < l->readfrompos)
		{
			// Count is the first part, 256 - [1, 255] = [255, 1]
			count = RS485_RXBUFF_SIZE - l->readfrompos;

			// Calculate size of the packet, [255, 1] + [0, 254] + 1 = [256, 256]
			unsigned int size = count + l->readtopos + 1;
			frame->size = (unsigned char) size;

			// If total is too large, discard it. This will only trigger if
			// the receive buffer has more space than 256 bytes (one full frame).
			if (size > 256)
			{
				FramepoolDelete (frame);
				return 0;
			}

			// Copy last part of the packet, located first in the buffer
			memcpy_helper((unsigned char*) frame + count, l->rxbuffer, l->readtopos + 1);
		}
		else
		{
			// Calculate size, [255-0+1, 0-0+1] = [256, 1]
			count = l->readtopos - l->readfrompos + 1;
			frame->size = (unsigned char) count;
		}

		// If too large, discard it. Same rules as above. If wrap-around, the
		// total size has already been checked. Otherwise this is the total.
		// Also check the frame->size
		if (count > 256 || frame->size < 7 || frame->size > 256)
		{
			IFSTATS (l->stats.error_framesize++);
			FramepoolDelete(frame);
			IFDEBUG_OTHER (Print ("Size delete %i if %x\n", count, (unsigned int) l));
			return 0;
		}

		// Copy packet contents
		memcpy_helper((unsigned char*) frame, l->rxbuffer + l->readfrompos, count);

		// Check for invalid destination
		if (frame->destination == 0)
		{
			IFSTATS (l->stats.error_checksum++);
			FramepoolDelete (frame);
			IFDEBUG_OTHER (NETWORK_PRINTI ("Dest delete if %x\n", (unsigned int) l));
			return 0;
		}

#if 0
		if (ChecksumVerify (frame) == 1)
		{
			IFSTATS (l->stats.error_checksum++);
			FramepoolDelete (frame);
			IFDEBUG_OTHER (NETWORK_PRINTI ("Check delete if %x\n", (unsigned int) l));
			return 0;
		}
#endif

		// Set the name of the interface it came from
		frame->originatinginterface = l->interfaceid;

		// Return the frame
		return frame;
	}
	else
	{
		// Allocation failed, log error
		IFDEBUG_OTHER (NETWORK_PRINTS ("Pump alloc err\n"));
		IFSTATS (l->stats.error_pumpallocerr++);

		// Done
		return 0;
	}
}

DataFrame *Layer2CheckPos (Layer2 *l)
{
	// DataFrame pointer
	DataFrame *frame = 0;

	// If read and write positions are inequal we need to process
	while (l->readtopos != l->writepos)
	{
		// Search for packet separator
		if (l->rxbuffer[l->readtopos] == PACKETSEPARATOR)
		{
			// Found separator, try extracting the packet
			// The packet starts at readfrompos and ends at readtopos
			frame = Layer2Extract (l);

			// Update readfrompos to the byte after the separator
			l->readfrompos = l->readtopos + 1;

			// Check wrap-around
			if (l->readfrompos == sizeof(l->rxbuffer))
				l->readfrompos = 0;
		}

		// Increment read position with wrap-around
		if ((++l->readtopos) == sizeof(l->rxbuffer))
			l->readtopos = 0;

		// Return the frame
		if (frame)
			return frame;
	}

	// Send packets in the transmit fifo if we can
	while (l->txframes[l->txframesread] && !Layer2SendFrameInternal (l, l->txframes[l->txframesread]))
	{
		// Clear that pointer
		l->txframes[l->txframesread] = 0;

		// Increment read position with wrap-around
		if ((++l->txframesread) == RS485_TXFIFO_SIZE)
			l->txframesread = 0;
	}

	// Done
	return 0;
}

int Layer2SendFrame (Layer2 *l, DataFrame *frame)
{
	// Set end
	((unsigned char*) frame)[frame->size-1] = PACKETSEPARATOR;

	// Set checksum
	frame->checksum = ChecksumCalculate (frame);

	// Send immediately if the transmitter is ready
	if (!l->txframes[l->txframesread] && !Layer2SendFrameInternal (l, frame))
		return 0;
	else
	{
		// If list has a free spot
		if (!l->txframes[l->txframeswrite])
		{
			// Add to list
			l->txframes[l->txframeswrite] = frame;

			// Increment index with wrap-around
			if ((++l->txframeswrite) == RS485_TXFIFO_SIZE)
				l->txframeswrite = 0;

			// Done
			return 0;
		}
		else
		{
			IFDEBUG_OTHER (NETWORK_PRINTS ("TX FIFO full\n"));
			IFSTATS (l->stats.error_txfifofull++);
			return 1;
		}
	}
}

void Layer2StoreFifo (Layer2 *l, DataFrame *frame)
{
	// Can we add to receive fifo?
	if (!rxframes[rxframeswrite])
	{
		// Yes, store it
		rxframes[rxframeswrite] = frame;

		// Increment the write index with wrap-around
		if ((++rxframeswrite) == RS485_RXFIFO_SIZE)
			rxframeswrite = 0;

		// Log statistics
		IFSTATS (l->stats.framesreceived++);
		IFSTATS (l->stats.bytesreceived += frame->size);
	}
	else
	{
		// No, report error and discard
		IFDEBUG_OTHER (NETWORK_PRINTS ("RX FIFO full\n"));
		IFSTATS (l->stats.error_rxfifofull++);
		FramepoolDelete(frame);
	}
}
