/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by erlendjs on 05/05/16.
//

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "mem/string.h"
#include "net/enet/network.h"
#include "net/enet/layer2/switch.h"
#include "net/enet/layer2/framepool.h"
#include "net/enet/layer2/arp.h"
#include "net/enet/layer2/manager.h"

#if NETWORK_DEBUG_SWITCH
#include NETWORK_DEBUG_INC
#endif

///////////////////////////////////////////////////////////////////////////////
// Debug definitions
///////////////////////////////////////////////////////////////////////////////
#if NETWORK_DEBUG_SWITCH && DEBUG_SWITCH_ADDLIST
	#define SWDEBUG_ADDLIST(str) str
#else
	#define SWDEBUG_ADDLIST(str)
#endif

#if NETWORK_DEBUG_SWITCH && DEBUG_SWITCH_UPDATELIST
	#define SWDEBUG_UPDATELIST(str) str
#else
	#define SWDEBUG_UPDATELIST(str)
#endif

#if NETWORK_DEBUG_SWITCH && DEBUG_SWITCH_DELETELIST
	#define SWDEBUG_DELETELIST(str) str
#else
	#define SWDEBUG_DELETELIST(str)
#endif

#if NETWORK_DEBUG_SWITCH && DEBUG_SWITCH_BROADCASTING
	#define SWDEBUG_BROADCASTING(str) str
#else
	#define SWDEBUG_BROADCASTING(str)
#endif

#if NETWORK_DEBUG_SWITCH && DEBUG_SWITCH_ROUTELOOP
	#define SWDEBUG_ROUTELOOP(str) str
#else
	#define SWDEBUG_ROUTELOOP(str)
#endif

///////////////////////////////////////////////////////////////////////////////
// Structure definitions
///////////////////////////////////////////////////////////////////////////////

// Switch cache table entry
typedef struct SwitchListEntry
{
	Layer2 *interface;
	unsigned short address;
	unsigned char timeout;
}SwitchListEntry;

///////////////////////////////////////////////////////////////////////////////
// Variables
///////////////////////////////////////////////////////////////////////////////

// Static cache table
SwitchListEntry switchlist[SWITCH_LISTSIZE];

///////////////////////////////////////////////////////////////////////////////
// Helpers
///////////////////////////////////////////////////////////////////////////////
void Layer2VirtualSendFrame (DataFrame *frame)
{
	// Give frame to the correct protocol handler, or free it if unknown
	if (frame->protocol == 1)
		IPProcess (frame);
	else if (frame->protocol == 2)
		ARPProcess (frame);
	else
		FramepoolDelete (frame);
}

///////////////////////////////////////////////////////////////////////////////
// Functions
///////////////////////////////////////////////////////////////////////////////
void Layer2SwitchInit (void)
{
	// Clear switch cache table
	MemSet (switchlist, 0, sizeof(switchlist));
}

void Layer2SwitchProcessFrame (DataFrame *frame)
{
	// This is set if the address is in the cache
	int found = 0;

	// Check the switch cache for updates
	for (int i = 0; i < SWITCH_LISTSIZE; i++)
	{
		if (frame->source == switchlist[i].address)
		{
			// If the interfaces are different, this will print additional info.
			// When this is printed, check for possible switching loops!
			SWDEBUG_UPDATELIST (
				if (switchlist[i].interface != frame->originatinginterface)
					Print ("SW Chif i%i %x to %x (hw %i to %i)\n", i,
						(unsigned int) switchlist[i].interface, (unsigned int) frame->originatinginterface,
						frame->source, frame->destination);
				);

			// Update where it came from and reset the timeout
			switchlist[i].interface = frame->originatinginterface;
			switchlist[i].timeout = SWITCH_ENTRY_TTL;
			found = 1;

			// Print debug message
			SWDEBUG_UPDATELIST (Print ("SW Update idx %i: %x (hw %i to %i)\n",
				i, (unsigned int) frame->originatinginterface, frame->source, frame->destination));
			break;
		}
	}

	// If we did not have it in the table we need to add it
	if (!found)
	{
		// Search for a free entry
		for (int i = 0; i < SWITCH_LISTSIZE; i++)
		{
			if (!switchlist[i].address)
			{
				// Set the address, interface and timeout
				switchlist[i].address = frame->source;
				switchlist[i].interface = frame->originatinginterface;
				switchlist[i].timeout = SWITCH_ENTRY_TTL;

				// Print debug message
				SWDEBUG_ADDLIST (NETWORK_PRINTI ("SW Add addr %i\n", (int) frame->source));
				SWDEBUG_ADDLIST (NETWORK_PRINTI ("   at index %i\n", i));
				break;
			}
		}
	}

	// If not explicit broadcast
	if (frame->destination != LAYER2_BROADCAST)
	{
		// Search the table to get the correct outgoing interface
		for (int i = 0; i < SWITCH_LISTSIZE; i++)
		{
			if (frame->destination == switchlist[i].address)
			{
				// Check for possible switching loop. In some cases a device
				// will receive a lot of packets that are broadcasted since
				// the other end does not have the receiver in its table. This
				// device will then have a lot of addresses in its table.
				// These addresses may be both the sender and receivers,
				// pointing to the same interface. Since they are not really
				// broadcast packets they will end up here. Skip to
				// broadcasting on all the other interfaces.
				if (frame->originatinginterface == switchlist[i].interface)
				{
					SWDEBUG_ROUTELOOP (
						Print ("Possible loop detected, if %x packet %i to %i\n",
							(unsigned int) frame->originatinginterface,
							frame->source, frame->destination);
						);
					break;
				}

				// If it resolved to ourself, process it, otherwise send it
				if (!switchlist[i].interface)
					Layer2VirtualSendFrame (frame);
				else if (!Layer2LinkUp (switchlist[i].interface))
					FramepoolDelete(frame);
				else if (Layer2SendFrame (switchlist[i].interface, frame))
					FramepoolDelete (frame);

				return;
			}
		}
	}

	// Could not find the outgoing interface, broadcast instead
	SWDEBUG_BROADCASTING (Print ("SW Broadcasting frame (hw %i to %i)\n",
		frame->source, frame->destination));

	// For every interface
	for (Layer2 *j = eglobals.iflist; j; j = j->next)
	{
		// Attempt sending if it is not the originating one and if the link is up
		if (frame->originatinginterface != j && Layer2LinkUp (j))
		{
			// Allocate for copy
			DataFrame *a = FramepoolNew ();

			// Check allocation
			if (a)
			{
				// Copy contents
				for (int i = 0; i < frame->size; i++)
					((unsigned char *) a)[i] = ((unsigned char *) frame)[i];

				// Set size
				a->size = frame->size;

				// Send frame on the interface
				if (Layer2SendFrame (j, a))
					FramepoolDelete (a);
			}
		}
	}

	// Broadcasts should also be received by self
	Layer2VirtualSendFrame(frame);
}

void Layer2SwitchMaintainList (void)
{
	// Loop through cache to delete timed out entries
	// This loop seems to take between 10 and 20 microseconds when list is empty
	for (int i = 0; i < SWITCH_LISTSIZE; i++)
	{
		// If valid entry
		if (switchlist[i].timeout)
		{
			// Decrement time value
			switchlist[i].timeout--;

			// Now if it timed out
			if (!switchlist[i].timeout)
			{
				// Clear the entry
				switchlist[i].address = 0;
				switchlist[i].interface = 0;

				// Print debug message
				SWDEBUG_DELETELIST (NETWORK_PRINTI ("SW Cleared index %i\n", i));
			}
		}
	}
}

void Layer2SwitchClearInterface (Layer2 *interface)
{
	SWDEBUG_DELETELIST (NETWORK_PRINTI ("SW Clearing list for %x\n", (unsigned int) interface));

	for (int i = 0; i < SWITCH_LISTSIZE; i++)
	{
		if (switchlist[i].interface == interface)
		{
			switchlist[i].address = 0;
			switchlist[i].interface = 0;
			switchlist[i].timeout = 0;
		}
	}
}
