/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "process/processcontrol.h"
#include "process/process.h"
#include "process/process-allocate.h"
#include "process/process-loader.h"
#include "sched/scheduler.h"
#include "sched/scheduler-signal.h"
#include "panic/panic.h"
#include "syscall/svc-basicfileops.h"
#include "posix/sys/signal.h"
#include "mpu/pager.h"

///////////////////////////////////////////////////////////////////////////////
// Variables
///////////////////////////////////////////////////////////////////////////////
int currentmaxpid;

///////////////////////////////////////////////////////////////////////////////
// Functions
///////////////////////////////////////////////////////////////////////////////
struct ProcessControl *ProcessCreateInternal (void(*func)(void),
	unsigned int *stack,
	int stacksize)
{
	// Allocate process data
	struct ProcessControl *slot = ProcessAllocate ();

	if (!slot)
		return 0;

	// Set pid and set current directory to root
	slot->pid = currentmaxpid++;
	slot->cwd[0] = '/';
	slot->cwd[1] = 0;

	// Start internal processes as privileged
	slot->trc.privileged = 0;

	// Add to schedulers run list
	SchedAddThread (&slot->trc, func, stack, stacksize, SCHED_PRIORITY_MEDIUM);
	return slot;
}

void ProcessFileCleanup (struct ProcessControl *p)
{
	// Memmapped resources
	PagerCleanupByPID (p->pid);

	// Files
	for (int i = 0; i < MAX_NODES; i++)
	{
		// Just close all the files. EBADF will be returned if it was not in
		// use, but that does not matter
		int derrno = 0;
		int closeargs[] = {i, 0, 0, (int)&derrno};
		svc_close (&p->trc, (void*) closeargs);
	}
}

void ProcessNotifyOnStatusChange (struct ThreadControl *ctc, int terminated)
{
	// Get structures of child and parent
	struct ProcessControl *child = ctc->pctrl;

	if (!child)
		Panic ("No child pointer");

	struct ProcessControl *parent = ProcessFindByPid (child->ownerpid);

	if (!parent)
		Panic ("No parent pointer");

	// If the process terminated, we have to terminate its subprocesses too
	if (terminated)
	{
		for (int i = 0; i < MAX_PROCESSDATA; i++)
		{
			if (processdata[i].inuse && processdata[i].ownerpid == child->pid)
			{
				SchedRemoveProcess (&processdata[i].trc);
				ProcessFileCleanup (&processdata[i]);
				ProcessSpawnFree (&processdata[i]);
				ProcessFree (&processdata[i]);
			}
		}
	}

	// Check the state of the parent. Is it blocked because of waitpid?
	if ((parent->trc.state & SCHED_STATUS_BLOCKED) && parent->trc.param3 == 32)
	{
		// Get the arguments
		int *r = (int*) ARCH_STACK_ACCESSOR(&parent->trc);

		// If it is blocking on any pid or this child
		if (r[0] < 0 || r[0] == child->pid)
		{
			// Unblock it with the appropriate pid and exit code
			r[0] = child->pid;
			*((int*)r[1]) = child->returncode;
			SchedState (&parent->trc, SCHED_STATUS_RUN, 0, SCHED_STATUS_BLOCKED, 0);

			// Because this ends the waitpid call we have to do more cleanup also
			if (terminated)
			{
				ProcessSpawnFree (child);
				ProcessFree (child);
			}

			// Debug message
			//PrintValue ("Exit wake", prc->pid);

			// Done
			return;
		}
	}

	// Just notify owner
	SchedPostSignal (&parent->trc, SIGCHLD);
}

int ProcessSendSignal (int pid, int sig)
{
	if (pid > 0)
	{
		// We are singaling a single process, find it
		struct ProcessControl *p = ProcessFindByPid (pid);

		if (p)
		{
			// Attempt sending the signal to the process
			int ret = SchedPostSignal (&p->trc, sig);

			// If it was ignored, handle it accordingly
			if (ret == 1)
				return ProcessHandleIgnoredSignal (p, sig);

			return ret;
		}

		// No process with that pid
		return 2;
	}
	else if (pid < 0)
	{
		// We are signaling a group, negate the argument to get the group id
		pid = -pid;

		// Search in all processes, and signal the first process found
		for (int i = 0; i < MAX_PROCESSDATA; i++)
		{
			if (processdata[i].inuse && processdata[i].gid == pid)
			{
				int ret = SchedPostSignal (&processdata[i].trc, sig);

				if (ret == 1)
					return ProcessHandleIgnoredSignal (&processdata[i], sig);

				return ret;
			}
		}

		// No process with that group id
		return 0;
	}
	else
	{
		// Cannot signal process 0
		return 2;
	}
}

int ProcessHandleIgnoredSignal (struct ProcessControl *p, int sig)
{
	if (sig == SIGKILL || sig == SIGINT)
	{
		// Terminate the process
		SchedRemoveProcess (&p->trc);
		ProcessFileCleanup (p);
		p->returncode = 2;
		ProcessNotifyOnStatusChange (&p->trc, 1);
		return 0;
	}
	else if (sig == SIGTSTP)
	{
		// Stop/suspend the process
		SchedState (&p->trc, SCHED_STATUS_STOPPED, 0, 0, 0);
		p->returncode = 1;
		ProcessNotifyOnStatusChange (&p->trc, 0);
		return 0;
	}
	else if (sig == SIGCONT)
	{
		// Resume the process
		if (currentthc == &p->trc)
			Panic ("Cant resume current process");

		SchedState (&p->trc, SCHED_STATUS_RUN, 0, SCHED_STATUS_STOPPED, 0);
		return 0;
	}

	// No default action for this signal
	return 1;
}

int ProcessCrashRecover (struct ThreadControl *ctc, unsigned int args)
{
	// Store its return value with WIFSIGNALED
	ctc->pctrl->returncode = 42 << 8 | 2;

	// Delete leftover resources
	ProcessFileCleanup (ctc->pctrl);

	// Print crash report
	PanicPrintReport (ctc);

	// Signal the parent
	ProcessNotifyOnStatusChange (ctc, 1);

	// Return true so that the dead thread is not woken
	return 1;
}
