/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "sched/list.h"

///////////////////////////////////////////////////////////////////////////////
// Functions
///////////////////////////////////////////////////////////////////////////////
void ListInit (List *l)
{
	l->nitems = 0;
	l->current = 0;
	l->last = 0;
	l->stepback = 0;

	ListVerify (l);
}

void ListAdd (List *l, ListItem *it)
{
	if (l->nitems)
	{
		it->next = l->last->next;
		it->previous = l->last;
		l->last->next->previous = it;
		l->last->next = it;
	}
	else
	{
		it->next = it;
		it->previous = it;
		l->current = it;
		l->last = it;
	}

	it->container = l;
	l->nitems++;

	ListVerify (l);
}

void ListAddSorted (List *l, ListItem *it, unsigned int sortvalue)
{
	if (l->nitems)
	{
		ListItem *cur = l->current;

		while (cur->sortvalue < sortvalue)
		{
			cur = cur->next;
			if (cur == l->current)
				break;
		}

		// At this point, cur is either:
		//  * the only item with lower sort value, add after it, do not set current
		//  * the only item with higher sort value, add before it and set current
		//  * the first item because all items has lower values, add before it and do not set current
		//  * some item with higher or equal sort value, add before it and do not set current

		it->next = cur;
		it->previous = cur->previous;
		cur->previous->next = it;
		cur->previous = it;

		// Because the above code adds the item before the current item,
		// l->current has to be set when their sort values are equal also.
		// Otherwise the added item will appear at the end of the list
		// and the process will appear to "lag".
		if (sortvalue <= l->current->sortvalue)
			l->current = it;
	}
	else
	{
		it->next = it;
		it->previous = it;
		l->current = it;
		l->last = it;
	}

	it->sortvalue = sortvalue;
	it->container = l;
	l->nitems++;

	ListVerify (l);
}

void ListRemove (ListItem *it)
{
	List *l = it->container;

	if (l->nitems == 1)
	{
		l->nitems = 0;
		l->current = 0;
		l->last = 0;
	}
	else
	{
		it->previous->next = it->next;
		it->next->previous = it->previous;

		if (l->last == it)
			l->last = it->previous;

		if (l->current == it)
		{
			// Bugfix: Prevent double-jump by going back instead of forwards
			// Problem: If SchedState removes the current process from the run
			// list it also triggers a reschedule. If SchedState advanced the
			// 'current' by one spot the scheduler will do it again,
			// effectively skipping over a thread. In bad cases this may cause
			// a deadlock. The run lists should have stepback set true so that
			// removing a thread sets the current one step back, and the
			// scheduler steps one step forward before context switching.
			if (l->stepback)
				l->current = it->previous;
			else
				l->current = it->next;
		}

		l->nitems--;
	}

	ListVerify (l);
}
