/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "sched/list.h"
#include "panic/panic.h"

///////////////////////////////////////////////////////////////////////////////
// Functions
///////////////////////////////////////////////////////////////////////////////

// Verify the "next" link of an item
void ListVerifyForwards (ListItem *it)
{
	// Never call this func on null
	if (!it)
		Panic ("ListVerifyForwards null");

	// The next pointer should never be null for an item that is part of
	// a list.
	if (!it->next)
		Panic ("ListVerifyForwards null next");

	// Verify that the next element has its previous pointer set to us.
	// This also verifies our next pointer at the same time.
	if (it->next->previous != it)
		Panic ("ListVerifyForwards failed it->previous");
}

// Verify the "previous" link of an item
void ListVerifyBackwards (ListItem *it)
{
	// Never call this func on null
	if (!it)
		Panic ("ListVerifyBackwards null");

	// The previous pointer should never be null for an item that is part of
	// a list.
	if (!it->previous)
		Panic ("ListVerifyBackwards null previous");

	// Verify that the previous element has its next pointer set to us.
	// This also verifies our previous pointer at the same time.
	if (it->previous->next != it)
		Panic ("ListVerifyBackwards failed it->previous");
}

// Verify a single item
void ListVerifyItem (List *l, ListItem *it)
{
	// Check forwards and backwards links
	ListVerifyForwards (it);
	ListVerifyBackwards (it);

	// Check that the container is valid
	if (it->container != l)
		Panic ("ListVerifyItem invalid container");

	// Check the ThreadControl pointer
	if (!it->parent)
		Panic ("ListVerifyItem invalid parent");
}

void ListVerify (List *l)
{
	// Verify contents if list is not empty
	if (l->nitems)
	{
		// Verify each item, and count. Start at "last" pointer
		int cnt = 0;
		ListItem *c = l->last;

		// Force the loop to run at least once, then until we encounter last
		// item again.
		while (!cnt || l->last != c)
		{
			ListVerifyItem (l, c);
			c = c->next;
			cnt++;

			// If there is a looping error, the while loop will run forever.
			// Do a sanity check on the counter.
			if (cnt > 10000)
				Panic ("ListVerify looping error");
		}

		// All items has been checked. Now check that the count is correct.
		if (l->nitems != cnt)
			Panic ("ListVerify incorrect count");

		// Verify that "current" can be found withing nitems
		int found = 0;
		c = l->last;

		for (int i = 0; i < l->nitems; i++, c = c->next)
			if (c == l->current)
				found = 1;

		if (!found)
			Panic ("ListVerify current not found\n");
	}
}
