/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "sched/scheduler.h"
#include "sched/scheduler-signal.h"
#include "panic/panic.h"

///////////////////////////////////////////////////////////////////////////////
// External variables
///////////////////////////////////////////////////////////////////////////////
extern List listrun[SCHED_NPRIORITIES];

///////////////////////////////////////////////////////////////////////////////
// Functions
///////////////////////////////////////////////////////////////////////////////

/*
 * Used by the context switch interrupt to select the next process to execute
 */
void SchedSelectNext (void)
{
	// Always do the selection in a critical section
	int crit = SchedEnterCritical ();

	// If the thread is in privileged (kernel) mode, check the bottom of the stack
	if (!currentthc->privileged)
	{
		if (currentthc->stackbottom[0] != 0xDEAD0123)
			Panic ("Possible stack overflow detected!");
	}

	// Special handling if signal state is set
	if (currentthc->sigstate)
	{
		// Process must be in the run list
		// TODO: Handle problems here
		if (currentthc->state != SCHED_STATUS_RUN)
			Panic ("Not in run state");

		// Leave signal processing if 1, enter if 2
		if (currentthc->sigstate == 1)
		{
			// Restore stack
			currentthc->stacktop = currentthc->stackstored;
			currentthc->stackstored = 0;
			currentthc->sigstate = 0;

			// If there are more signals pending, we need to re-enter
			if (currentthc->sigpending)
			{
				for (int i = 0; i < 32; i++)
				{
					if (currentthc->sigpending & (1 << i))
					{
						currentthc->param1 = i + 1;
						SchedSignalStackSetup (currentthc);
						break;
					}
				}
			}
		}
		else
			SchedSignalStackSetup (currentthc);
	}
	else
	{
		// Regular scheduling, select next of highest priority
		unsigned int toppri = 0;

		// TODO: Fix bad stuff happening if medium pri is empty but high is not
		while (listrun[toppri].nitems == 0)
			toppri++;

		listrun[toppri].current = listrun[toppri].current->next;
		currentthc = listrun[toppri].current->parent;
	}

	// Exit critical section
	SchedExitCritical (crit);
}
