/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "sched/scheduler.h"
#include "sched/scheduler-signal.h"
#include "panic/panic.h"

///////////////////////////////////////////////////////////////////////////////
// Functions
///////////////////////////////////////////////////////////////////////////////
int SchedPostSignal (struct ThreadControl *thc, int sig)
{
	// Calculate signal bit
	int signal = 1 << (sig - 1);

	// Check mask and function pointers, return 1 if signaling is not possible
	if (!(thc->sigmask & signal) || !thc->sigh || !thc->sigt)
		return 1;

	// If the thread is in an "abortable" blocking call, cancel that
	if ((thc->state & SCHED_STATUS_BLOCKED) && thc->callback)
	{
		thc->callback (thc, thc->param2);
		thc->callback = 0;
		SchedState (thc, SCHED_STATUS_RUN, 0, SCHED_STATUS_BLOCKED, 0);
	}

	// Disable interrupts
	int crit = SchedEnterCritical ();

	// Set signal to pending
	thc->sigpending |= signal;

	// Enter signal
	if (thc->stackstored)
	{
		// The thread is already in the signal processing state, so set the
		// pending bit. sigstate can be in either 0 state or 1 state. If 0,
		// it will eventually be set 1 by the trampoline and the scheduler
		// will check for pending signals after restoring the stack. If 1,
		// the same will happen. It cannot be 2 because if it is, the scheduler
		// has not been involved yet and has not gotten to setting the stack.
	}
	else if (thc == currentthc)
	{
		// The stack has not been stored (not in signal processing state), and
		// the target is also the current process. Set sigstate to 2 so that
		// the scheduler will set the stack up for us when "clean", and yield.
		// sigstate may already be 2, in that case the last one is delivered
		// first and the rest is up to the scheduler when it checks for pending
		// signals. sigstate may be 1 if the signal was posted after the
		// trampoline but before the scheduler was invoked. In that case, do
		// not set sigstate and let the scheduler set us up after restoring the
		// stack. sigstate will be 0 if in normal processing state, when in
		// signal processing state the previous "if" would have been invoked.
		if (thc->sigstate != 1)
			thc->sigstate = 2;

		thc->param1 = (unsigned int) sig;
		SchedYield ();
	}
	else
	{
		// The stack has not been stored and the process sending the signal
		// (current) is not the target. We may the set the stack up immediately
		// and the process will continue normally, but execute the signal once
		// it is scheduled. In the meantime, any further calls to PostSignals
		// will end up in the first "if" check.
		thc->param1 = (unsigned int) sig;
		SchedSignalStackSetup (thc);
	}

	// Wake the thread if sleeping
	if (thc->state & SCHED_STATUS_SLEEPING)
		SchedState (thc, SCHED_STATUS_RUN, 0, SCHED_STATUS_SLEEPING, 0);

	SchedExitCritical (crit);
	return 0;
}
