/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "syscall/svc-basicfileops.h"
#include "fs/vfs.h"
#include "fs/dev/dev.h"
#include "posix/errno.h"

///////////////////////////////////////////////////////////////////////////////
// Definitions
///////////////////////////////////////////////////////////////////////////////

// For manipulating stack arguments
#define RET ((int*)args)[0]
#define ERRN ((int**)args)[3]

///////////////////////////////////////////////////////////////////////////////
// Functions
///////////////////////////////////////////////////////////////////////////////
int svc_open (struct ThreadControl *ctc, void *args)
{
	// Find a free file descriptor in the process
	int pfd = 255;

	for (int i = 0; i < MAX_NODES; i++)
		if (ctc->pctrl->fnodes[i] == 255)
			{pfd = i; break;}

	// Check if it is valid
	if (pfd == 255)
	{
		*ERRN = ENOMEM;
		return -1;
	}

	// It is valid, find a kernel file descriptor
	int kfd = -1;

	for (int i = 0; i < NODETABLE_SIZE; i++)
		if (!nodetable[i].usecounter)
			{kfd = i; break;}

	// Check if it is valid
	if (kfd < 0)
	{
		*ERRN = ENOMEM;
		return -1;
	}

	// Cast variables
	const char *path = (const char*)((int*)args)[0];
	int flags = ((int*)args)[1];
	int mode = ((int*)args)[2];

	// Set dev and flags
	nodetable[kfd].dev = 0;
	nodetable[kfd].flags = flags;

	// Attempt open
	int ret = vfs_open (&nodetable[kfd], ctc->pctrl->cwd, path, flags, mode);

	// Check for error
	if (ret < 0)
	{
		*ERRN = -ret;
		return -1;
	}

	// Success
	nodetable[kfd].usecounter++;
	ctc->pctrl->fnodes[pfd] = kfd;
	return pfd;
}

int svc_close (struct ThreadControl *ctc, void *args)
{
	// File descriptor index from the caller
	unsigned int pfd = ((unsigned int*)args)[0];

	// Index to file descriptor in the kernel
	int kfd;

	// Check for validity and get kfd
	if (pfd < MAX_NODES && (kfd = ctc->pctrl->fnodes[pfd]) != 255)
	{
		// If this is the last use of the descriptor
		if (nodetable[kfd].usecounter == 1)
		{
			// Close it properly
			int ret = vfs_close (ctc, &nodetable[kfd]);

			if (ret < 0)
			{
				*ERRN = -ret;
				return -1;
			}
		}

		// It is still used elsewhere, decrement use counter
		nodetable[kfd].usecounter--;
		ctc->pctrl->fnodes[pfd] = 255;
		return 0;
	}

	// Invalid descriptor
	*ERRN = EBADF;
	return -1;
}

int svc_read (struct ThreadControl *ctc, void *args)
{
	// File descriptor index from the caller
	unsigned int pfd = ((unsigned int*)args)[0];

	// Index to file descriptor in the kernel
	int kfd;

	// Check for validity and get kfd
	if (pfd < MAX_NODES && (kfd = ctc->pctrl->fnodes[pfd]) != 255)
	{
		int ret = vfs_read (&nodetable[kfd], (char*)((int*)args)[1], ((int*)args)[2]);

		if (ret < 0)
		{
			*ERRN = -ret;
			return -1;
		}

		return ret;
	}

	// Invalid descriptor
	*ERRN = EBADF;
	return -1;
}

int svc_write (struct ThreadControl *ctc, void *args)
{
	// File descriptor index from the caller
	unsigned int pfd = ((unsigned int*)args)[0];

	// Index to file descriptor in the kernel
	int kfd;

	// Check for validity and get kfd
	if (pfd < MAX_NODES && (kfd = ctc->pctrl->fnodes[pfd]) != 255)
	{
		int ret = vfs_write (&nodetable[kfd], (char*)((int*)args)[1], ((int*)args)[2]);

		if (ret < 0)
		{
			*ERRN = -ret;
			return -1;
		}

		return ret;
	}

	// Invalid descriptor
	*ERRN = EBADF;
    return -1;
}

int svc_ioctl_fcntl (struct ThreadControl *ctc, void *args)
{
	// File descriptor index from the caller
	unsigned int pfd = ((unsigned int*)args)[0];

	// Index to file descriptor in the kernel
	int kfd;

	// Check for validity and get kfd
	if (pfd < MAX_NODES && (kfd = ctc->pctrl->fnodes[pfd]) != 255)
	{
		// Check if we have a file system pointer
		if (!nodetable[kfd].fs)
		{
			*ERRN = EINVAL;
			return -1;
		}

		// Call the file system ioctl
		int ret = nodetable[kfd].fs->ioctl_fcntl (&nodetable[kfd],
			(unsigned int)((int*)args)[1], (void*)((int*)args)[2]);

		if (ret < 0)
		{
			*ERRN = -ret;
			return -1;
		}

		return ret;
	}

	// Invalid descriptor
	*ERRN = EBADF;
    return -1;
}
