/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "process/processcontrol.h"
#include "fs/vfs.h"
#include "posix/errno.h"
#include "posix/sys/stat.h"

///////////////////////////////////////////////////////////////////////////////
// Definitions
///////////////////////////////////////////////////////////////////////////////

// For manipulating stack arguments
#define RET ((int*)args)[0]
#define ERRN ((int**)args)[3]

///////////////////////////////////////////////////////////////////////////////
// Functions
///////////////////////////////////////////////////////////////////////////////
int svc_stat (struct ThreadControl *ctc, void *args)
{
	// Get the name argument
	const char *name = (const char*)((int*)args)[0];

	// Get the stat argument
	struct stat *st = (struct stat *)((int*)args)[1];

	// Do the stat operation
	int ret = vfs_stat (ctc->pctrl->cwd, name, st);

	// Check for errors and return
	if (ret < 0)
	{
		*ERRN = -ret;
		return -1;
	}

	return ret;
}

int svc_link (struct ThreadControl *ctc, void *args)
{
	// The link syscall is currently not supported, use rename instead
	(void)ctc;
	(void)args;
	return -1;
}

int svc_unlink (struct ThreadControl *ctc, void *args)
{
	// Get the name argument
	const char *name = (const char*)((int*)args)[0];

	// Do the unlink operation
	int ret = vfs_unlink (ctc->pctrl->cwd, name);

	// Check for errors and return
	if (ret < 0)
	{
		*ERRN = -ret;
		return -1;
	}

	return ret;
}

int svc_rename (struct ThreadControl *ctc, void *args)
{
	// Get the name arguments
	const char *namea = (const char*)((int*)args)[0];
	const char *nameb = (const char*)((int*)args)[1];

	// Do the rename operation
	int ret = vfs_rename (ctc->pctrl->cwd, namea, nameb);

	// Check for errors and return
	if (ret < 0)
	{
		*ERRN = -ret;
		return -1;
	}

	return ret;
}

int svc_mkdir (struct ThreadControl *ctc, void *args)
{
	// Get the name and mode flags
	const char *name = (const char*)((int*)args)[0];
	int mode = ((int*)args)[1];

	// Do the mkdir operation
	int ret = vfs_mkdir (ctc->pctrl->cwd, name, mode);

	// Check for errors and return
	if (ret < 0)
	{
		*ERRN = -ret;
		return -1;
	}

	return ret;
}

int svc_opendir (struct ThreadControl *ctc, void *args)
{
	// Get the name
	const char *name = (const char*)((int*)args)[0];

	// Open the directory
	int ret = vfs_opendir (ctc->pctrl->cwd, name);

	// Check for errors and return
	if (ret < 0)
	{
		*ERRN = -ret;
		return -1;
	}

	return ret;
}

int svc_closedir (struct ThreadControl *ctc, void *args)
{
	// CTC Is not used!
	(void)ctc;

	// Close the directory
	int ret = vfs_closedir ();

	// Check for errors and return
	if (ret < 0)
	{
		*ERRN = -ret;
		return -1;
	}

	return ret;
}

int svc_readdir (struct ThreadControl *ctc, void *args)
{
	// CTC Is not used!
	(void)ctc;

	// Read the directory
	int ret = vfs_readdir ((struct dirent*)((int*)args)[1]);

	// Check for errors and return
	if (ret < 0)
	{
		*ERRN = -ret;
		return -1;
	}

	return ret;
}

int svc_statvfs (struct ThreadControl *ctc, void *args)
{
	// Get the path and statvfs structure
	const char *path = (const char*)((int*)args)[0];
	struct statvfs *buf = (struct statvfs*)((int*)args)[1];

	// Do the statvfs operation
	int ret = vfs_statvfs (ctc->pctrl->cwd, path, buf);

	// Check for errors and return
	if (ret < 0)
	{
		*ERRN = -ret;
		return -1;
	}

	return ret;
}
