/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "process/processcontrol.h"
#include "posix/errno.h"
#include "posix/sys/mman.h"
#include "mem/mem.h"
#include "print/print.h"
#include "sched/scheduler.h"
#include "mpu/pager.h"
#include "mpu/mpu.h"

///////////////////////////////////////////////////////////////////////////////
// Definitions
///////////////////////////////////////////////////////////////////////////////

// Used for checking against valid flags
#define PROT_ALLFLAGS (PROT_READ | PROT_WRITE | PROT_EXEC | PROT_NONE)

// Set to 1 in order to enable some debug messages
#if 0
#define DODEBUG(cmd) cmd
#else
#define DODEBUG(cmd)
#endif

///////////////////////////////////////////////////////////////////////////////
// Functions
///////////////////////////////////////////////////////////////////////////////
int SvcMemmanHandleMmap (struct ThreadControl *ctc, int *args, int *errptr)
{
	// Debug
	DODEBUG (Print ("SvcMemmanHandleMmap %x %x %x %x\n", args[0], args[1], args[2], args[3]));

	// Check for specified (non-null) address, not supported
	if (args[0])
	{
		errptr[0] = ENOTSUP;
		return -1;
	}

	// Length of zero should result in EINVAL
	if (!args[1])
	{
		errptr[0] = EINVAL;
		return -1;
	}

	// Check that the protection flags make sense
	if (args[2] & ~PROT_ALLFLAGS)
	{
		errptr[0] = EINVAL;
		return -1;
	}

	// Also disallow executable memory allocation for now
	// TODO: Should have some sort of permissions for this?
	if (args[2] & PROT_EXEC)
	{
		errptr[0] = ENOTSUP;
		return -1;
	}

	// Only MAP_PRIVATE and MAP_ANONYMOUS are supported at the moment
	if (args[3] != (MAP_PRIVATE | MAP_ANONYMOUS))
	{
		errptr[0] = EINVAL;
		return -1;
	}

	// File descriptor/offset, not supported, ignored with anonymous alloc

	// Allocate
	struct MemMpuReturn mpu;
	unsigned int mem = MemAllocMpu (args[1], 0, &mpu);

	// Check the allocation
	if (!mem)
	{
		DODEBUG (Print ("MemAllocMpu failed\n"));
		errptr[0] = ENOMEM;
		return -1;
	}

	// Set region attributes
	mpu.mpu_rasr |= TEX(0)|C|S|E;

	if (!(args[2] & PROT_EXEC))
		mpu.mpu_rasr |= XN;
	if (args[2] & PROT_READ)
		mpu.mpu_rasr |= AP(2);
	if (args[2] & PROT_WRITE)
		mpu.mpu_rasr |= AP(3);

	// Set the page table
	if (PagerAddEntry (ctc->pctrl->pid, mpu.mpu_rbar, mpu.mpu_rasr, mpu.dataptr, mem))
	{
		// Page table entry allocation failed
		DODEBUG (Print ("PagerAddEntry failed\n"));
		MemFree (mem);
		errptr[0] = ENOMEM;
		return -1;
	}

	// Success
	DODEBUG (Print ("SvcMemmanHandleMmap success 0x%x\n", mpu.dataptr));
	return mpu.dataptr;
}

int SvcMemmanHandleMunmap (struct ThreadControl *ctc, int *args, int *errptr)
{
	DODEBUG (Print ("SvcMemmanHandleMunmap addr %x len %x\n", args[0], args[1]));

	if (PagerFreeEntry (ctc->pctrl->pid, args[0]))
	{
		DODEBUG (Print ("SvcMemmanHandleMunmap Could not free memory\n"));
		errptr[0] = EINVAL;
		return -1;
	}
	else
	{
		PlatformPagerRegisterSetup (ctc->mpuregs);
		SchedYield ();
		return 0;
	}
}

int svc_memman (struct ThreadControl *ctc, void *args)
{
	// Get the arguments
	int calltype = ((int*)args)[0];
	int *arguments = ((int**)args)[1];
	int *errptr = ((int**)args)[3];

	// Only perform work if the pager is enabled (configuration option)
	if (PAGER_ENABLE == 1)
	{
		// Branch on the call type
		if (calltype == 0)
		{
			// This is a mmap call
			// TODO: Verify access validity of arguments
			return SvcMemmanHandleMmap (ctc, arguments, errptr);
		}
		else if (calltype == 1)
		{
			// This is a munmap call
			// TODO: Verify access validity of arguments
			return SvcMemmanHandleMunmap (ctc, arguments, errptr);
		}
	}

	// Unknown call
	DODEBUG (Print ("svc_memman unknown call\n"));
	errptr[0] = EINVAL;
	return -1;
}
