/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "fs/vfs.h"
#include "fs/dev/dev.h"
#include "posix/errno.h"
#include "process/process-allocate.h"
#include "process/process-loader.h"
#include "process/process.h"
#include "sched/scheduler.h"
#include "worker/worker.h"

///////////////////////////////////////////////////////////////////////////////
// Definitions
///////////////////////////////////////////////////////////////////////////////

// For manipulating stack arguments
#define RET ((int*)args)[0]
#define ERRN ((int**)args)[3]

///////////////////////////////////////////////////////////////////////////////
// Local functions
///////////////////////////////////////////////////////////////////////////////
int svc_spawn_rethelper (struct ThreadControl *ctc, unsigned int args)
{
	int ret = ProcessSpawn (ctc, (void*)args);

	if (ret < 0)
	{
		*ERRN = -ret;
		ret = -1;
	}

	RET = ret;
	return 0;
}

///////////////////////////////////////////////////////////////////////////////
// Functions
///////////////////////////////////////////////////////////////////////////////
int svc_spawn (struct ThreadControl *ctc, void *args)
{
	// Postpone processing of this syscall
	SchedState (ctc, SCHED_STATUS_DEFERRED,
		&svc_spawn_rethelper, (unsigned int) args, ((int*)args)[0]);

	// Wake the worker
	WorkerUnblock ();

	return 0;
}

int svc_waitpid (struct ThreadControl *ctc, void *args)
{
	// Get the arguments
	int pid = ((int*)args)[0];
	int *status = (int*)((int*)args)[1];
	int options = ((int*)args)[2];

	// Vars
	int found = 0;

	// Search list for the child we're waiting for
	for (int i = 0; i < MAX_PROCESSDATA; i++)
	{
		// Is it a child of ours?
		if (processdata[i].inuse && processdata[i].ownerpid == ctc->pctrl->pid)
		{
			// Mark presence of any child
			found = 1;

			// Skip if positive and it is not the pid we're looking for.
			// Otherwise, it is a specific child or any child
			if (pid > 0 && pid != processdata[i].pid)
				continue;

			// Check state change
			if (!processdata[i].trc.statechanged)
				continue;

			processdata[i].trc.statechanged = 0;

			// Check if regular stopped
			if (processdata[i].trc.state == SCHED_STATUS_STOPPED)
			{
				// Set status to stopped and return pid
				*status = 1;
				return processdata[i].pid;
			}

			// Skip if it's status isnt zombie
			if (processdata[i].trc.state == SCHED_STATUS_ZOMBIE ||
				processdata[i].trc.state == SCHED_STATUS_CRASHED)
			{
				// Set the status code according to the return code
				*status = processdata[i].returncode;

				// Free the childs memory now and return its pid
				int retpid = processdata[i].pid;
				ProcessSpawnFree (&processdata[i]);
				ProcessFree (&processdata[i]);
				return retpid;
			}
		}
	}

	// At this point, the process we're looking for is either still active,
	// or does not exist. Check if there are no children.
	if (!found)
	{
		*ERRN = ECHILD;
		return -1;
	}

	// Now, we know children exist but all are still active. When WNOHANG is
	// set we should just return 0.
	if (options & 1)
		return 0;

	// It is okay to block
	SchedState (ctc, SCHED_STATUS_BLOCKED, 0, (unsigned int) args, 32);

	// This return should always be "manipulated" on the stack later on
	// It does not actually matter what it is set to here
	return pid;

}

int svc_kill (struct ThreadControl *ctc, void *args)
{
	int pid = ((int*)args)[0];
	int sig = ((int*)args)[1];

	int ret = ProcessSendSignal (pid, sig);

	if (!ret)
		return 0;
	else if (ret == 1)
		*ERRN = EINVAL;
	else if (ret == 2)
		*ERRN = ESRCH;

	return -1;
}
