/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "syscall/svc-basicfileops.h"
#include "fs/vfs.h"
#include "fs/socket/socket.h"
#include "posix/errno.h"
#include "process/process-allocate-file.h"
#include "print/print.h"

///////////////////////////////////////////////////////////////////////////////
// Definitions
///////////////////////////////////////////////////////////////////////////////

// For manipulating stack arguments
#define RET ((int*)args)[0]
#define ERRN ((int**)args)[3]

///////////////////////////////////////////////////////////////////////////////
// Functions
///////////////////////////////////////////////////////////////////////////////
int svc_bind (struct ThreadControl *ctc, void *args)
{
	// File descriptor index from the caller
	unsigned int pfd = ((unsigned int*)args)[0];

	// Index to file descriptor in the kernel
	int kfd;

	// Check for validity and get kfd
	if (pfd < MAX_NODES && (kfd = ctc->pctrl->fnodes[pfd]) != 255)
	{
		int ret = vfs_socket_bind (&nodetable[kfd], (const struct sockaddr *)((int*)args)[1], (socklen_t)((int*)args)[2]);

		if (ret < 0)
		{
			*ERRN = -ret;
			return -1;
		}

		return ret;
	}

	// Invalid descriptor
	*ERRN = EBADF;
	return -1;
}

int svc_recvfrom (struct ThreadControl *ctc, void *args)
{
	// File descriptor index from the caller
	unsigned int pfd = ((unsigned int*)args)[0];

	// Index to file descriptor in the kernel
	int kfd;

	// Check for validity and get kfd
	if (pfd < MAX_NODES && (kfd = ctc->pctrl->fnodes[pfd]) != 255)
	{
		int ret = vfs_socket_recvfrom (&nodetable[kfd], (struct recvfrom_args*)((int*)args)[1]);

		if (ret < 0)
		{
			*ERRN = -ret;
			return -1;
		}

		return ret;
	}

	// Invalid descriptor
	*ERRN = EBADF;
	return -1;
}

int svc_sendto (struct ThreadControl *ctc, void *args)
{
	// File descriptor index from the caller
	unsigned int pfd = ((unsigned int*)args)[0];

	// Index to file descriptor in the kernel
	int kfd;

	// Check for validity and get kfd
	if (pfd < MAX_NODES && (kfd = ctc->pctrl->fnodes[pfd]) != 255)
	{
		int ret = vfs_socket_sendto (&nodetable[kfd], (struct sendto_args*)((int*)args)[1]);

		if (ret < 0)
		{
			*ERRN = -ret;
			return -1;
		}

		return ret;
	}

	// Invalid descriptor
	*ERRN = EBADF;
    return -1;
}

int svc_socket (struct ThreadControl *ctc, void *args)
{
	// Allocate a file descriptor in the process
	struct FileNode *fn;
	int pfd = ProcessAllocateFileDesc (ctc, &fn);

	// Check if it is valid
	if (pfd < 0)
	{
		*ERRN = -pfd;
		return -1;
	}

	// Attempt open
	int ret = vfs_socket_socket (fn, ((int*)args)[0], ((int*)args)[1], ((int*)args)[2]);

	// Check for error, deallocate and return error
	if (ret < 0)
	{
		ctc->pctrl->fnodes[pfd] = 255;
		fn->usecounter = 0;
		*ERRN = -ret;
		return -1;
	}

	// Success
	return pfd;
}

int svc_listen (struct ThreadControl *ctc, void *args)
{
	// File descriptor index from the caller
	unsigned int pfd = ((unsigned int*)args)[0];

	// Index to file descriptor in the kernel
	int kfd;

	// Check for validity and get kfd
	if (pfd < MAX_NODES && (kfd = ctc->pctrl->fnodes[pfd]) != 255)
	{
		int ret = vfs_socket_listen (&nodetable[kfd], args);

		if (ret < 0)
		{
			*ERRN = -ret;
			return -1;
		}

		return ret;
	}

	// Invalid descriptor
	*ERRN = EBADF;
    return -1;
}

int svc_accept (struct ThreadControl *ctc, void *args)
{
	// File descriptor index from the caller
	unsigned int pfd = ((unsigned int*)args)[0];

	// Index to file descriptor in the kernel
	int kfd;

	// Check for validity and get kfd
	if (pfd < MAX_NODES && (kfd = ctc->pctrl->fnodes[pfd]) != 255)
	{
		// Allocate a new file descriptor
		struct FileNode *newfn;
		int newpfd = ProcessAllocateFileDesc (ctc, &newfn);

		// Check if it is valid
		if (newpfd < 0)
		{
			*ERRN = -newpfd;
			return -1;
		}

		// Try the accept function
		int ret = vfs_socket_accept (&nodetable[kfd], newfn, args);

		if (ret < 0)
		{
			// Delete the new file desc
			ctc->pctrl->fnodes[newpfd] = 255;
			newfn->usecounter = 0;

			// Return error code
			*ERRN = -ret;
			return -1;
		}

		return ret;
	}

	// Invalid descriptor
	*ERRN = EBADF;
    return -1;
}

int svc_connect (struct ThreadControl *ctc, void *args)
{
	// File descriptor index from the caller
	unsigned int pfd = ((unsigned int*)args)[0];

	// Index to file descriptor in the kernel
	int kfd;

	// Check for validity and get kfd
	if (pfd < MAX_NODES && (kfd = ctc->pctrl->fnodes[pfd]) != 255)
	{
		int ret = vfs_socket_connect (&nodetable[kfd], args);

		if (ret < 0)
		{
			*ERRN = -ret;
			return -1;
		}

		return ret;
	}

	// Invalid descriptor
	*ERRN = EBADF;
    return -1;
}

int svc_sockopt (struct ThreadControl *ctc, void *args)
{
	// File descriptor index from the caller
	unsigned int pfd = ((unsigned int*)args)[0];

	// Index to file descriptor in the kernel
	int kfd;

	// Check for validity and get kfd
	if (pfd < MAX_NODES && (kfd = ctc->pctrl->fnodes[pfd]) != 255)
	{
		int ret = vfs_socket_sockopt (&nodetable[kfd], args);

		if (ret < 0)
		{
			*ERRN = -ret;
			return -1;
		}

		return ret;
	}

	// Invalid descriptor
	*ERRN = EBADF;
    return -1;
}

int svc_sockmisc (struct ThreadControl *ctc, void *args)
{
	// File descriptor index from the caller
	unsigned int pfd = ((unsigned int*)args)[0];

	// Index to file descriptor in the kernel
	int kfd;

	// Check for validity and get kfd
	if (pfd < MAX_NODES && (kfd = ctc->pctrl->fnodes[pfd]) != 255)
	{
		int ret = vfs_socket_sockmisc (&nodetable[kfd], args);

		if (ret < 0)
		{
			*ERRN = -ret;
			return -1;
		}

		return ret;
	}

	// Invalid descriptor
	*ERRN = EBADF;
    return -1;
}

int svc_netdb (struct ThreadControl *ctc, void *args)
{
	// File descriptor index from the caller
	unsigned int pfd = ((unsigned int*)args)[0];

	// Index to file descriptor in the kernel
	int kfd;

	// Check for validity and get kfd
	if (pfd < MAX_NODES && (kfd = ctc->pctrl->fnodes[pfd]) != 255)
	{
		int ret = vfs_socket_netdb (&nodetable[kfd], args);

		if (ret < 0)
		{
			*ERRN = -ret;
			return -1;
		}

		return ret;
	}

	// Invalid descriptor
	*ERRN = EBADF;
    return -1;
}
