/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "process/processcontrol.h"
#include "fs/vfs.h"
#include "posix/errno.h"
#include "posix/time.h"
#include "posix/sys/time.h"
#include "posix/sys/stat.h"
#include "sched/scheduler.h"
#include "platform/platform.h"
#include "print/print.h"

///////////////////////////////////////////////////////////////////////////////
// Definitions
///////////////////////////////////////////////////////////////////////////////

// For manipulating stack arguments
#define ERRN ((int**)args)[3]

///////////////////////////////////////////////////////////////////////////////
// Helpers
///////////////////////////////////////////////////////////////////////////////
int svc_nanosleep_callback (struct ThreadControl *ctc, unsigned int args)
{
	// The sleep was either interrupted or elapsed. Since we do not care about
	// the remainder time, just return success for now.
	return 0;
}

///////////////////////////////////////////////////////////////////////////////
// Functions
///////////////////////////////////////////////////////////////////////////////
int svc_times (struct ThreadControl *ctc, void *args)
{
	Print ("TODO: times\n");
	return -ENOSYS;
}

int svc_nanosleep (struct ThreadControl *ctc, void *args)
{
	// Get arguments
	struct timespec *req = (struct timespec*)((int*)args)[0];
	//struct timespec *rem = (struct timespec*)((int*)args)[1];

	// Do not continue if invalid arguments
	if (!req)
		return -EINVAL;

	// Calculate the time in milliseconds
	unsigned int time = (req->tv_sec * 1000) + (req->tv_nsec / 1000000);

	// Do not accept a time of zero
	if (time <= 0)
		return 0;

	// In a critical section, check that we may indeed sleep (no signals pending)
	int crit = SchedEnterCritical ();

	if (ctc->sigpending)
	{
		// There is a signal pending, abort sleep
		SchedExitCritical (crit);
		return 0;
	}

	// Set state to sleeping while in critical, then exit critical
	SchedState (ctc, SCHED_STATUS_SLEEPING, &svc_nanosleep_callback, ((int*)args)[1], time);
	SchedExitCritical (crit);

	// If a signal is asserted here, the thread MUST BE WOKEN or it will
	// trigger a kernel panic in scheduler-algo.c
	return 0;
}

int svc_gettimeofday (struct ThreadControl *ctc, void *args)
{
	struct timeval *tv = (struct timeval*)((int*)args)[0];
	//struct timezone *tz = (struct timezone*)((int*)args)[1];

	if (!tv)
		return -EINVAL;

	unsigned int t = PlatformGetTick ();
	tv->tv_sec = t / 1000;
	tv->tv_usec = (t % 1000) * 1000;
	return 0;
}
