/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "sched/threadcontrol.h"
#include "sched/scheduler.h"
#include "sched/list.h"
#include "platform/platform.h"
#include "timer/timer.h"

///////////////////////////////////////////////////////////////////////////////
// Variables
///////////////////////////////////////////////////////////////////////////////
List timerlist;
unsigned int timer_now;

///////////////////////////////////////////////////////////////////////////////
// Private functions
///////////////////////////////////////////////////////////////////////////////
void TimerUpdate (void)
{
	// Get current time
	unsigned int now = PlatformGetTick ();

	// Loop until we catch up
	while (timer_now != now)
	{
		// Increment our timer "position"
		timer_now++;

		// Briefly disable interrupt when we check
		// TODO: Check if we can spend less time in critical
		int crit = SchedEnterCritical ();

		// Check for timers, there may be multiple elapsing now
		while (timerlist.nitems)
		{
			ListItem *it = timerlist.current;

			// This holds because we increment it syncronously
			if (timer_now == it->sortvalue)
			{
				TimerHandle *handle = (TimerHandle*) it;

				ListRemove (it);
				it->container = 0;

				if (handle->flags & TIMER_FLAG_REPEATING)
					ListAddSorted (&timerlist, it, timer_now + handle->interval);

				// TODO: Do not call while in critical
				handle->Callback (handle);
			}
			else
				break;
		}

		// Re-enable interrupts
		SchedExitCritical (crit);
	}
}

void TimerThread (void)
{
	while (1)
	{
		TimerUpdate ();
		SchedState (currentthc, SCHED_STATUS_SLEEPING, 0, 0, 1);
	}
}

///////////////////////////////////////////////////////////////////////////////
// Functions
///////////////////////////////////////////////////////////////////////////////
void TimerInitialize (void)
{
	static struct ThreadControl timerthread;
	static unsigned int timerstack[64];

	timer_now = PlatformGetTick ();
	ListInit (&timerlist);

	timerthread.privileged = 0;
	SchedAddThread (&timerthread, &TimerThread, timerstack, 64, SCHED_PRIORITY_MEDIUM);
}

void TimerRegister (TimerHandle *handle)
{
	ListItem *it = (ListItem*) handle;

	if (it->container)
		ListRemove (it);

	ListAddSorted (&timerlist, it, timer_now + handle->interval);
}

void TimerUnRegister (TimerHandle *handle)
{
	ListItem *it = (ListItem*) handle;

	if (it->container)
	{
		ListRemove (it);
		it->container = 0;
	}
}

int TimerIsRegistered (TimerHandle *handle)
{
	ListItem *it = (ListItem*) handle;

	return it->container ? 1 : 0;
}
