/*
 * Copyright (c) 2018, Erlend Sveen
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>

///////////////////////////////////////////////////////////////////////////////
// Internal functions
///////////////////////////////////////////////////////////////////////////////
int puts_internal (const char *s)
{
	// Get length
	int len = strlen (s);

	// Working vars
	int n = len;
	int pos = 0;

	// Write while there is more data
	do
	{
		int diff = write (1, s + pos, n);

		if (diff < 0)
			return diff;

		pos += diff;
		n -= diff;
	} while (pos != len);

	return len;
}

__attribute__((weak))
void Reverse (char *s)
{
	for (int i = 0, j = strlen(s)-1; i < j; i++, j--)
	{
		char c = s[i];
		s[i] = s[j];
		s[j] = c;
	}
}

__attribute__((weak))
void _itoa (int n, char *str)
{
	int sign = n;

	if (n < 0)
		n = -n;

	int i = 0;

	do {
		str[i++] = '0' + n % 10;
	} while ((n /= 10));

	if (sign < 0)
		str[i++] = '-';

	str[i] = 0;
	Reverse(str);
}

__attribute__((weak))
void itoa_uint32_hex (unsigned int in, char *out)
{
	// http://stackoverflow.com/questions/311165/how-do-you-convert-byte-array-to-hexadecimal-string-and-vice-versa
	unsigned char b;

	b = (in >> 28) & 0x0F;
	out[0] = (55 + b + (((b-10)>>31)&-7));
	b = (in >> 24) & 0x0F;
	out[1] = (55 + b + (((b-10)>>31)&-7));

	b = (in >> 20) & 0x0F;
	out[2] = (55 + b + (((b-10)>>31)&-7));
	b = (in >> 16) & 0x0F;
	out[3] = (55 + b + (((b-10)>>31)&-7));

	b = (in >> 12) & 0x0F;
	out[4] = (55 + b + (((b-10)>>31)&-7));
	b = (in >> 8) & 0x0F;
	out[5] = (55 + b + (((b-10)>>31)&-7));

	b = (in >> 4) & 0x0F;
	out[6] = (55 + b + (((b-10)>>31)&-7));
	b = (in >> 0) & 0x0F;
	out[7] = (55 + b + (((b-10)>>31)&-7));
	out[8] = 0;
}

int printf_internal (char *output, size_t size, const char *format, va_list list)
{
	// Initialize output data
	if (size != -1)
		memset (output, 0, size);

	// Process the format string
	for (char *out = output; *format; format++)
	{
		if (*format == '%')
		{
			format++;
			int modifier = -1;

			if (*format == '.')
			{
				format++;
				modifier = *format - '0';

				if (!*format)
					break;

				format++;
			}

			if (*format == 'i' || *format == 'u')
			{
				int arg = va_arg (list, int);
				_itoa (arg, out);
			}
			else if (*format == 'x' || *format == 'X' || *format == 'p')
			{
				unsigned int arg = va_arg (list, unsigned int);
				char str[16];
				char *st = str;
				itoa_uint32_hex (arg, str);

				if (modifier > 0)
				{
					modifier = 8 - modifier;

					if (modifier > 0 && modifier < 8)
						st += modifier;
				}

				while (*st)
					*out++ = *st++;
			}
			else if (*format == 'c')
			{
				char c = va_arg (list, int);
				*out = c;
			}
			else if (*format == 's')
			{
				char *s = va_arg (list, char*);
				while (*s)
					*out++ = *s++;
			}
			else
			{
				int discard = va_arg (list, int);
				(void) discard;
				*out = *format;
			}
		}
		else
			*out = *format;

		while (*out) out++;
	}

	return 0;
}

///////////////////////////////////////////////////////////////////////////////
// Functions
///////////////////////////////////////////////////////////////////////////////
int fprintf (FILE *file, const char *format, ...)
{
	va_list list;
	va_start (list, format);
	int ret = printf (format, list);
	va_end (list);
	return ret;
}

int printf (const char *format, ...)
{
	// Initialize argument list
	va_list list;
	va_start (list, format);

	// Declare output data
	char output[128];

	// Process the format string
	printf_internal (output, sizeof (output), format, list);

	// Print the string and exit
	va_end (list);
	return puts_internal (output);
}

int puts (const char *s)
{
	int ret = puts_internal (s);

	if (ret < 0)
		return ret;

	return ret + puts_internal ("\n");
}

int snprintf (char *str, size_t size, const char *fmt, ...)
{
	// Initialize argument list
	va_list list;
	va_start (list, fmt);

	// Process the format string
	int ret = printf_internal (str, size, fmt, list);

	// End of list
	va_end (list);
	return ret;
}

int sprintf (char *str, const char *fmt, ...)
{
	// Initialize argument list
	va_list list;
	va_start (list, fmt);

	// Process the format string
	int ret = printf_internal (str, -1, fmt, list);

	// End of list
	va_end (list);
	return ret;
}

int vprintf (const char *fmt, va_list list)
{
	// Declare output data
	char output[128];

	// Process the format string
	printf_internal (output, sizeof (output), fmt, list);

	// Done
	return puts_internal (output);
}
