/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include "lcdtest.h"

///////////////////////////////////////////////////////////////////////////////
// Misc
///////////////////////////////////////////////////////////////////////////////
void Delay (unsigned int time)
{
	struct timespec req;

	req.tv_sec = time / 1000;
	req.tv_nsec = (time % 1000) * 1000000;

	nanosleep (&req, 0);
}

///////////////////////////////////////////////////////////////////////////////
// Functions
///////////////////////////////////////////////////////////////////////////////
void LCDRotate (lcd *l, int direction)
{
	if (direction)
	{
		// Vertical
		writecommandarg1 (l, 0x36, 0x48);
	}
	else
	{
		// Horizontal
		writecommandarg1 (l, 0x36, /*0x80 |*/ 0x40 | 0x20 | 0x08);
	}
}

// ILI9341 Display Driver
void initili (lcd *l)
{
	writecommandarg0 (l, 0x11);
	Delay (120);

	writecommandarg3 (l, 0xCF, 0x00, 0xC3, 0x30); // Power control B
	writecommandarg4 (l, 0xED, 0x64, 0x03, 0x12, 0x81); // Power on sequence control
	writecommandarg3 (l, 0xE8, 0x85, 0x10, 0x79); // Driver timing control A
	writecommandarg5 (l, 0xCB, 0x39, 0x2C, 0x00, 0x34, 0x02); // Power control A
	writecommandarg1 (l, 0xF7, 0x20); // Pump ratio control
	writecommandarg2 (l, 0xEA, 0x00, 0x00); // Driver timing control B
	writecommandarg1 (l, 0xC0, 0x22); // Power control 1
	writecommandarg1 (l, 0xC1, 0x11); // Power control 2
	writecommandarg2 (l, 0xC5, 0x3D, 0x20); // VCOM control 1
	writecommandarg1 (l, 0xC7, 0xAA); // VCOM control 2
	writecommandarg1 (l, 0x36, 0x08); // Memory Access Control
	writecommandarg1 (l, 0x3A, 0x55); // Pixel format set
	writecommandarg2 (l, 0xB1, 0x00, 0x13); // Frame rate control in normal mode / full colors
	writecommandarg2 (l, 0xB6, 0x0A, 0xA2); // Display Function Control
	writecommandarg2 (l, 0xF6, 0x01, 0x30); // Interface control
	writecommandarg1 (l, 0xF2, 0x00); // Enable 3 gamma
	writecommandarg1 (l, 0x26, 0x01); // Gamma set

	// Positive and negative gamma correction
	unsigned char gammap[] = {0x0F, 0x3F, 0x2F, 0x0C, 0x10, 0x0A, 0x53, 0xD5, 0x40, 0x0A, 0x13, 0x03, 0x08, 0x03, 0x00};
	unsigned char gamman[] = {0x00, 0x00, 0x10, 0x03, 0x0F, 0x05, 0x2C, 0xA2, 0x3F, 0x05, 0x0E, 0x0C, 0x37, 0x3C, 0x0F};

	cslow (l);
	writecommand (l, 0xE0);
	writedata (l, gammap, sizeof(gammap));
	cshigh (l);

	cslow (l);
	writecommand (l, 0xE1);
	writedata (l, gamman, sizeof(gamman));
	cshigh (l);

	writecommandarg0 (l, 0x11);
	Delay (120);
	LCDRotate (l, 0);
	writecommandarg0 (l, 0x29);
}

int initlcd (lcd *l)
{
	printf ("Init start\n");

	initili (l);

	printf ("Init complete\n");
	return 0;
}

int tst (lcd *l, char *filename)
{
	char c;

	while (read (0, &c, 1) != 1) writecommandarg0 (l, 0x85);

	while (read (0, &c, 1) != 1) writecommandarg1 (l, 0x85, 0);

	while (read (0, &c, 1) != 1) writecommandarg2 (l, 0x85, 0, 1);

	while (read (0, &c, 1) != 1) writecommandarg3 (l, 0x85, 0, 1, 2);

	while (read (0, &c, 1) != 1) writecommandarg4 (l, 0x85, 0, 1, 2, 3);

	while (read (0, &c, 1) != 1) writecommandarg5 (l, 0x85, 0, 1, 2, 3, 4);

	return 0;
}

unsigned int LCDColor (unsigned int r, unsigned int g, unsigned int b)
{
	// rrrrrggggggbbbbb
	unsigned int temp = 0;
	temp = ((r & 0xF8) << 8) | ((g & 0xFC) << 3) | ((b & 0xF8) >> 3);
	return temp;
}

int img (lcd *l, char *filename)
{
	int f = open (filename, O_RDONLY);

	if (f >= 0)
	{
		unsigned short dat[256];
		setaddress (l, 0, 319, 0, 239);
		cslow (l);

		while (1)
		{
			int rd = read (f, dat, sizeof (dat));

			for (int i = 0; i < 256; i++)
			{
				unsigned short a = dat[i];
				dat[i] = (a >> 8) | (a << 8);
			}

			writedata (l, (unsigned char*) dat, sizeof(dat));
			if (rd <= 0)
				break;
		}

		cshigh (l);
		close (f);
	}
	else
		printf ("Failed to open: %s\n", strerror (errno));

	return 0;
}

int pattern (lcd *l, char *filename)
{
	unsigned short dat[256];

	while (1)
	{
		setaddress (l, 20, 100, 10, 120);

		cslow (l);

		for (int i = 0; i < 256; i++)
			dat[i] = (unsigned short) LCDColor (i, 0, 0);

		writedata (l, (unsigned char*) dat, sizeof(dat));
		writedata (l, (unsigned char*) dat, sizeof(dat));
		writedata (l, (unsigned char*) dat, sizeof(dat));
		writedata (l, (unsigned char*) dat, sizeof(dat));

		for (int i = 0; i < 256; i++)
			dat[i] = (unsigned short) LCDColor (0, i, 0);

		writedata (l, (unsigned char*) dat, sizeof(dat));
		writedata (l, (unsigned char*) dat, sizeof(dat));
		writedata (l, (unsigned char*) dat, sizeof(dat));
		writedata (l, (unsigned char*) dat, sizeof(dat));

		for (int i = 0; i < 256; i++)
			dat[i] = (unsigned short) LCDColor (0, 0, i);

		writedata (l, (unsigned char*) dat, sizeof(dat));
		writedata (l, (unsigned char*) dat, sizeof(dat));
		writedata (l, (unsigned char*) dat, sizeof(dat));
		writedata (l, (unsigned char*) dat, sizeof(dat));

		for (int i = 0; i < 256; i++)
			dat[i] = (unsigned short) LCDColor (0, i, i);

		writedata (l, (unsigned char*) dat, sizeof(dat));
		writedata (l, (unsigned char*) dat, sizeof(dat));
		writedata (l, (unsigned char*) dat, sizeof(dat));
		writedata (l, (unsigned char*) dat, sizeof(dat));

		cshigh (l);
	}
}

int main (int argc, char **argv)
{
	lcd l;

	if (argc < 2)
	{
		printf ("Need at least two arguments\n");
		return 1;
	}

	if (initbus (&l))
		return 1;

	if (initlcd (&l))
		return 2;

	if (!strcmp (argv[1], "img"))
		return img (&l, argv[2]);
	else if (!strcmp (argv[1], "pattern"))
		return pattern (&l, argv[2]);
	else if (!strcmp (argv[1], "tst"))
		return tst (&l, argv[2]);
	else
		printf ("Don't know that command\n");

	return 0;
}
