/*
 * Copyright (C) 2018 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <time.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <simpleprint/simpleprint.h>

#define LINEWIDTH 8

void Delay (unsigned int time)
{
	struct timespec req;

	req.tv_sec = time / 1000;
	req.tv_nsec = (time % 1000) * 1000000;

	nanosleep (&req, 0);
}

void PrintLinePortion (unsigned int offs, unsigned char *data, unsigned int n)
{
	char str[16];
	char out[8+2+(LINEWIDTH*2)+1+LINEWIDTH+2];
	int i = 8;

	itoa_uint32_hex (offs, out);
	out[i++] = ':';
	out[i++] = ' ';

	for (int t = 0; t < n; t++)
	{
		itoa_uint32_hex (data[t], str);
		out[i++] = str[6];
		out[i++] = str[7];
	}

	out[i++] = ' ';

	for (int t = 0; t < n; t++)
	{
		char d = data[t];

		if (d < ' ' || d > '~')
			d = '.';

		out[i++] = d;
	}

	out[i++] = '\n';
	out[i++] = 0;
	SimplePrint (out);
}

void cs (int fd, int cs)
{
	int cs0 = ioctl (fd, SPICS, cs + 1);

	if (!cs0)
		printf ("CS%i Appears OK\n", cs);
	else
		printf ("CS%i Set failed with %i (%s)\n", cs, cs0, strerror (errno));

	int cs0dis = ioctl (fd, SPICS, 0);

	if (cs0dis < 0)
		printf ("CS%i Disable wailed with: %s\n", cs, strerror (errno));
}

void findallcs (int fd)
{
	cs (fd, 0);
	cs (fd, 1);
	cs (fd, 2);
	cs (fd, 3);
}

void csloop (int fd, int cs)
{
	printf ("Running chip select toggle loop on cs%i\n", cs);

	while (1)
	{
		int cs0 = ioctl (fd, SPICS, cs + 1);

		if (cs0)
		{
			printf ("CS%i Set failed with %i (%s)\n", cs, cs0, strerror (errno));
			break;
		}

		Delay (5);

		int cs0dis = ioctl (fd, SPICS, 0);

		if (cs0dis < 0)
		{
			printf ("CS%i Disable wailed with: %s\n", cs, strerror (errno));
			break;
		}

		Delay (5);
	}

	printf ("Aborting toggle loop\n");
}

void csread (int fd, int cs, int argc, char **argv)
{
	int nbytes = 1;

	if (argc >= 4)
		nbytes = atoi (argv[3]);

	if (argc < 4 || nbytes < 1 || nbytes > 128)
	{
		printf ("Specify number of bytes [1, 128]\n");
		return;
	}

	int cs0 = ioctl (fd, SPICS, cs + 1);

	if (cs0)
	{
		printf ("CS%i Set failed with %i (%s)\n", cs, cs0, strerror (errno));
		return;
	}

	unsigned char readbuf[128];

	for (int i = 0; i < sizeof (readbuf); i++)
		readbuf[i] = 0xAB;

	int nread = read (fd, readbuf, nbytes);

	if (nread < 0)
		printf ("Read failed with: %s\n", strerror (errno));
	else
	{
		// Wait for completion
		int timer = 1000000;

		while (timer && ioctl (fd, SPIBUSY, 0))
			timer--;

		// Check
		if (!timer)
			printf ("Timeout waiting for data\n");
		else
		{
			// Now print data
			for (int i = 0; i < sizeof (readbuf) && i < nread; i += 8)
				PrintLinePortion (i, &readbuf[i], 8);
		}
	}

	int cs0dis = ioctl (fd, SPICS, 0);

	if (cs0dis < 0)
		printf ("CS%i Disable wailed with: %s\n", cs, strerror (errno));
}

void cswrite (int fd, int cs, int argc, char **argv)
{
	int nbytes = 1;

	if (argc >= 4)
		nbytes = strlen (argv[3]);

	if (argc < 4 || nbytes < 1 || nbytes > 128)
	{
		printf ("Specify string with number of bytes [1, 128]\n");
		return;
	}

	int cs0 = ioctl (fd, SPICS, cs + 1);

	if (cs0)
	{
		printf ("CS%i Set failed with %i (%s)\n", cs, cs0, strerror (errno));
		return;
	}

	int nwrite = write (fd, argv[3], nbytes);

	if (nwrite < 0)
		printf ("Write failed with: %s\n", strerror (errno));
	else
	{
		// Wait for completion
		int timer = 1000000;

		while (timer && ioctl (fd, SPIBUSY, 0))
			timer--;

		// Check
		if (!timer)
			printf ("Timeout waiting for data\n");
		else
			printf ("Write completed after %i cycles\n", 1000000 - timer);
	}

	int cs0dis = ioctl (fd, SPICS, 0);

	if (cs0dis < 0)
		printf ("CS%i Disable wailed with: %s\n", cs, strerror (errno));
}

int main (int argc, char **argv)
{
	if (argc < 3)
	{
		printf ("Please specify device and test type\n");
		return 1;
	}

	int fd = open (argv[1], O_RDWR);

	if (fd < 0)
	{
		printf ("Error opening device: %s\n", strerror (errno));
		return 2;
	}

	if (!strcmp (argv[2], "findallcs"))
		findallcs (fd);
	else if (!strcmp (argv[2], "cs0"))
		cs (fd, 0);
	else if (!strcmp (argv[2], "cs1"))
		cs (fd, 1);
	else if (!strcmp (argv[2], "cs2"))
		cs (fd, 2);
	else if (!strcmp (argv[2], "cs3"))
		cs (fd, 3);
	else if (!strcmp (argv[2], "cs0loop"))
		csloop (fd, 0);
	else if (!strcmp (argv[2], "cs1loop"))
		csloop (fd, 1);
	else if (!strcmp (argv[2], "cs2loop"))
		csloop (fd, 2);
	else if (!strcmp (argv[2], "cs3loop"))
		csloop (fd, 3);
	else if (!strcmp (argv[2], "cs0read"))
		csread (fd, 0, argc, argv);
	else if (!strcmp (argv[2], "cs1read"))
		csread (fd, 1, argc, argv);
	else if (!strcmp (argv[2], "cs2read"))
		csread (fd, 2, argc, argv);
	else if (!strcmp (argv[2], "cs3read"))
		csread (fd, 3, argc, argv);
	else if (!strcmp (argv[2], "cs0write"))
		cswrite (fd, 0, argc, argv);
	else if (!strcmp (argv[2], "cs1write"))
		cswrite (fd, 1, argc, argv);
	else if (!strcmp (argv[2], "cs2write"))
		cswrite (fd, 2, argc, argv);
	else if (!strcmp (argv[2], "cs3write"))
		cswrite (fd, 3, argc, argv);
	else
		printf ("Don't know that command\n");

	close (fd);
	return 0;
}
