/*
 * Copyright (c) 2018, Erlend Sveen
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/mman.h>

///////////////////////////////////////////////////////////////////////////////
// Functions
///////////////////////////////////////////////////////////////////////////////
int nullptr (char **argv)
{
	printf ("Testing access violation\n");

	int a = 42;
	volatile int *a1 = 0;
	volatile int *a2 = &a;

	if (*a1 == *a2)
		printf ("Failed: Got equal\n");
	else
		printf ("Failed: Got unequal\n");

	return 0;
}

int ptr (int argc, char **argv)
{
	if (argc != 3 || strlen(argv[2]) != 8)
	{
		printf ("Invalid arguments, specify address in hex\n");
		return 1;
	}

	int a = 42;
	int b = 0;
	volatile int *a1 = 0;
	volatile int *a2 = &a;

	for (int i = 0, j = 7; i < 8; i++, j--)
	{
		if (argv[2][i] >= '0' && argv[2][i] <= '9')
			b |= (argv[2][i] - '0') << (j * 4);
		if (argv[2][i] >= 'a' && argv[2][i] <= 'f')
			b |= (argv[2][i] - 'a' + 10) << (j * 4);
		if (argv[2][i] >= 'A' && argv[2][i] <= 'F')
			b |= (argv[2][i] - 'A' + 10) << (j * 4);
	}

	a1 = (volatile int*) b;

	printf ("Testing pointers\n");

	if (*a1 == *a2)
		printf ("Failed: Got equal\n");
	else
		printf ("Failed: Got unequal\n");

	return 0;
}

int singlemmap (int argc, char **argv)
{
	printf ("Allocating space: 1024 bytes\n");

	int size = 1024;
	int *addr = mmap (0, size, PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_ANONYMOUS, 0, 0);

	if (addr == MAP_FAILED)
	{
		printf ("Allocation failed\n");
		return 1;
	}

	printf ("Interacting with memory...\n");

	volatile int *a = addr;
	a[0] = 10;
	a[2] = 10;

	if (a[0] != a[2])
		printf ("Data was not stored\n");
	if (a[1] != 0)
		printf ("Data is not zeroed\n");

	printf ("Unmapping\n");
	munmap (addr, size);
	return 0;
}

int doublemmap (int argc, char **argv)
{
	printf ("Allocating space: 1024 bytes x2\n");

	int size = 1024;
	int *addr1 = mmap (0, size, PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_ANONYMOUS, 0, 0);
	int *addr2 = mmap (0, size, PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_ANONYMOUS, 0, 0);

	if (addr1 == MAP_FAILED || addr2 == MAP_FAILED)
	{
		printf ("Allocation failed\n");
		return 1;
	}

	printf ("Interacting with memory...\n");

	volatile int *a = addr1;
	volatile int *b = addr2;

	a[0] = 10;
	a[2] = 10;

	if (a[0] != a[2])
		printf ("Data was not stored\n");
	if (a[1] != 0)
		printf ("Data is not zeroed\n");

	memcpy ((void*)b, (void*)a, size);

	if (b[0] != 10 || b[1] || b[2] != 10)
		printf ("Data copy failed\n");

	printf ("Unmapping\n");
	munmap (addr1, size);
	munmap (addr2, size);
	return 0;
}

int mmapleak (int argc, char **argv)
{
	printf ("Allocating space and leaking it: 1024 bytes\n");

	int size = 1024;
	mmap (0, size, PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_ANONYMOUS, 0, 0);
	return 0;
}

///////////////////////////////////////////////////////////////////////////////
// Usage
///////////////////////////////////////////////////////////////////////////////
int PrintUsage (void)
{
	printf (
		"Usage: crashtest [options]\n"
		"  nullptr\n"
		"    Test access to 0x0\n"
		"  ptr <addr>\n"
		"    Test access to addr\n"
		"  singlemmap\n"
		"    Test a simple mmap allocation\n"
		"  doublemmap\n"
		"    Test a double mmap allocation\n"
		"  mmapleak\n"
		"    Test leaking memory\n"
		"  --help\n"
		"    Print this message\n");
	return 0;
}

int PrintUnknownCommand (void)
{
	printf (
		"Invalid argument, use\n"
		"   crashtest --help\n"
		"for a usage list\n");
	return 1;
}

///////////////////////////////////////////////////////////////////////////////
// Main function
///////////////////////////////////////////////////////////////////////////////
int main (int argc, char **argv)
{
	// If no arguments, print default
	if (argc == 1)
		return PrintUnknownCommand ();

	// Check arguments
	if (!strcmp (argv[1], "nullptr"))
		return nullptr (argv);
	else if (!strcmp (argv[1], "ptr"))
		return ptr (argc, argv);
	else if (!strcmp (argv[1], "singlemmap"))
		return singlemmap (argc, argv);
	else if (!strcmp (argv[1], "doublemmap"))
		return doublemmap (argc, argv);
	else if (!strcmp (argv[1], "mmapleak"))
		return mmapleak (argc, argv);
	else if (!strcmp(argv[1], "--help"))
		return PrintUsage ();
	else
		return PrintUnknownCommand ();
}
