/*
 * Copyright (c) 2018, Erlend Sveen
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "structures.h"

///////////////////////////////////////////////////////////////////////////////
// Functions
///////////////////////////////////////////////////////////////////////////////
void resolve_blx (struct function *head, struct function *f, struct instruction *h)
{
	char *targetregtok = strtok (h->rest, " \t");

	if (!targetregtok || !strcmp (targetregtok, "lr"))
		return;

	char targetreg[32];
	strcpy (targetreg, targetregtok);

	//fprintf (stderr, "Found register branch and link in func %s, reg %s\n",
	//	f->name, targetreg);

	struct instruction *i = h;

	while (i)
	{
		// In some cases the address goes through a move instruction
		if (!strcmp (i->operation, "mov"))
		{
			// Make a copy of the "rest"
			int l = strlen (i->rest);
			char *s = malloc (l + 1);
			strcpy (s, i->rest);

			// Tokenize into target reg, source reg
			char *arg1 = strtok (s, ", []");
			char *arg2 = strtok (0, ", []");

			// Debug message
			//fprintf (stderr, "Got %s %s\n", arg1, arg2);

			// Check if all are valid
			if (arg1 && arg2 && !strcmp (arg1, targetreg))
				strcpy (targetreg, arg2);

			// Done
			free (s);
		}
		else if (!strcmp (i->operation, "ldr.w") ||
			!strcmp (i->operation, "ldr"))
		{
			// Make a copy of the "rest"
			int l = strlen (i->rest);
			char *s = malloc (l + 1);
			strcpy (s, i->rest);

			// Tokenize into target reg, base reg and offset
			char *arg1 = strtok (s, ", []");
			char *arg2 = strtok (0, ", []");
			char *arg3 = strtok (0, ", []");
			char *arg4 = strtok (0, ", [];(\t");

			// Set if done
			int done = 0;

			// If all are valid
			if (arg1 && arg2 && arg3 &&
				!strcmp (arg1, targetreg) && !strcmp (arg2, "pc"))
			{
				// Get the address of the data that ldr loads
				if (arg3[0] == '#')
					arg3++;

				unsigned int adder = strtol (arg3, 0, 10);
				unsigned int searchfor = i->address + adder;

				if (!strcmp (i->operation, "ldr.w"))
					searchfor += 4;
				else
					searchfor += 4;

				if (arg4)
					searchfor = strtol (arg4, 0, 16);

				// Search for
				//fprintf (stderr, "%s %s %s %.8X\n", arg1, arg2, arg3, searchfor);

				// Search for it in the current function only
				struct instruction *ii = f->inst;

				while (ii)
				{
					if (ii->address == searchfor)
					{
						unsigned int data = ii->data & ~1;

						struct function *func = findfuncbyaddr (head, data);

						if (func)
						{
							//fprintf (stderr, "Found blx from %s to %.8X (%s)\n",
							//	f->name, data, func->name);

							func->isreferenced++;
							h->branch = func;
							done = 1;
							break;
						}
					}

					ii = ii->next;
				}
			}

			// Done
			free (s);

			if (done)
				return;
		}

		i = i->prev;
	}
}

void analyzefunc (struct function *head, struct function *f)
{
	//fprintf (stderr, "Analyzing function %s\n", f->name);

	struct instruction *h = f->inst;

	while (h)
	{
		if (!strcmp (h->operation, "b.w"))
		{
			unsigned int target = strtol (h->rest, 0, 16);
			struct function *func = findfuncbyaddr (head, target);

			if (func)
			{
				//fprintf (stderr, "Direct branch found from %s to %.8X (%s)\n",
				//	f->name, target, func->name);

				func->isreferenced++;
				h->branch = func;
			}
		}
		else if (!strcmp (h->operation, "bl"))
		{
			unsigned int target = strtol (h->rest, 0, 16);
			struct function *func = findfuncbyaddr (head, target);

			if (func)
			{
				//fprintf (stderr, "Branch and link found from %s to %.8X (%s)\n",
				//	f->name, target, func->name);

				func->isreferenced++;
				h->branch = func;
			}
		}
		else if (!strcmp (h->operation, "blx"))
		{
			resolve_blx (head, f, h);
		}
		else if (!strcmp (h->operation, "bx"))
		{
			resolve_blx (head, f, h);
		}

		h = h->next;
	}
}

void analyzefuncs (struct function *head)
{
	struct function *h = head;

	while (h)
	{
		analyzefunc (head, h);
		h = h->next;
	}
}

void readinput (struct function **head)
{
	char buf[1024];
	struct function *current = 0;

	while (gets (buf))
	{
		char *a = strtok (buf, " \t");

		if (a)
		{
			int len = strlen (a);

			if (a[len-1] != ':')
			{
				char *name = strtok (0, "<>:");

				if (name)
				{
					current = newfunc (name, a);
					addlist (head, current);
				}
			}
			else if (a[len-1] == ':')
			{
				struct instruction *inst = newinst (a, a + len + 1);

				if (inst)
				{
					addinst (&current->inst, inst);
					inst->parent = current;
				}
			}
		}
	}
}

int main (int argc, char **argv)
{
	struct function *head = 0;

	fprintf (stderr, "Reading input data\n");
	readinput (&head);

	fprintf (stderr, "Done reading input, analyzing functions\n");
	analyzefuncs (head);

	fprintf (stderr, "Coloring the graph\n");
	colorgraph (head);

	fprintf (stderr, "Checking for unreferenced functions\n");
	printnobranchfuncs (head);

	fprintf (stderr, "Sanitizing names\n");
	sanitizegraphnames (head);

	fprintf (stderr, "Trunking edges\n");
	dographtrunc (head);

	fprintf (stderr, "Printing graph\n");
	printgraph (head);

	return 0;
}
