/*
 * Copyright (c) 2018, Erlend Sveen
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "structures.h"

///////////////////////////////////////////////////////////////////////////////
// Data
///////////////////////////////////////////////////////////////////////////////
const char *colors[] = {"black", "red", "green", "blue", "darkorange"};
const char *stdlibfuncs[] = {"malloc", "free", "_read", "read", "_read_r", "write",
	"_write_r", "_write", "ioctl", "_ioctl", "sendto", "sendto_asm", "recvfrom",
	"recvfrom_asm", "puts", "puts_internal", "strlen", "strncat", "strncmp", "printf",
	"vprintf", "printf_internal", "_itoa", "itoa_uint32_hex", "Reverse", "_exit",
	"memcpy", "loopforever", "_init", "Concat", "__libc_init_array", "EntryPoint",
	"memset", "rand", "srand", "open", "_open", "_open_r", "strerror", 0};
const char *nobranch_whitelist[] = {
	"_open", "_close", "_read", "_write", "_ioctl", 0};

///////////////////////////////////////////////////////////////////////////////
// Private Functions
///////////////////////////////////////////////////////////////////////////////
void sanitizename (char *name)
{
	int len = strlen (name);

	for (int i = 0; i < len; i++)
		if (name[i] == '.')
			name[i] = '_';
}

void dographtruncfunc (struct function *head)
{
	struct instruction *i = head->inst;

	while (i)
	{
		if (i->branch && !i->isduplicate)
		{
			i->isfirst = 1;

			struct instruction *ii = i->next;

			while (ii)
			{
				if (ii->branch == i->branch)
				{
					ii->isduplicate = 1;
					i->hastrunced = 1;
				}

				ii = ii->next;
			}
		}

		i = i->next;
	}
}

///////////////////////////////////////////////////////////////////////////////
// Functions
///////////////////////////////////////////////////////////////////////////////
void colorgraph (struct function *head)
{
	int cindex = 0;

	struct function *h = head;

	while (h)
	{
		h->color = colors[cindex++];

		if (cindex >= 5)
			cindex = 0;

		for (int i = 0; stdlibfuncs[i]; i++)
		{
			if (!strcmp (h->name, stdlibfuncs[i]))
			{
				h->isclib = 1;
				break;
			}
		}

		h = h->next;
	}
}

void sanitizegraphnames (struct function *head)
{
	while (head)
	{
		sanitizename (head->name);
		head = head->next;
	}
}

void printnobranchfuncs (struct function *head)
{
	struct function *h = head;

	while (h)
	{
		int hasbranch = 0;

		struct instruction *i = h->inst;

		while (i)
		{
			if (i->branch)
				hasbranch = 1;

			i = i->next;
		}

		if (!hasbranch)
		{
			int whitelisted = 0;

			for (int i = 0; nobranch_whitelist[i]; i++)
			{
				if (!strcmp (h->name, nobranch_whitelist[i]))
				{
					whitelisted = 1;
					break;
				}
			}

			if (!whitelisted)
				fprintf (stderr, "Makes no function calls: %s\n", h->name);
		}

		h = h->next;
	}

	h = head;

	while (h)
	{
		if (!h->isreferenced)
			fprintf (stderr, "Never called (only by func ptr?): %s\n", h->name);

		h = h->next;
	}
}

void dographtrunc (struct function *head)
{
	struct function *h = head;

	while (h)
	{
		dographtruncfunc (h);
		h = h->next;
	}
}

void printgraph (struct function *head)
{
	// Header
	printf ("digraph graphname {\n");
	printf ("rankdir=LR;\n");
	printf ("node [shape=ellipse];\n");

	// Function names
	struct function *h = head;

	while (h)
	{
		if (h->isclib)
			printf ("%s [label=\"0x%.8X: %s\", color=darkgreen, shape=\"box\", penwidth=2];\n",
				h->name, h->address, h->name);
		else
			printf ("%s [label=\"0x%.8X: %s\"];\n",
				h->name, h->address, h->name);

		h = h->next;
	}

	// Edges
	h = head;

	while (h)
	{
		struct instruction *i = h->inst;

		while (i)
		{
			if (i->branch && i->isfirst)
			{
				if (i->hastrunced)
					printf ("%s -> %s [label=\"%.8X\",color=%s, penwidth=3]\n",
						h->name, i->branch->name, h->address, h->color);
				else
					printf ("%s -> %s [label=\"%.8X\",color=%s]\n",
						h->name, i->branch->name, h->address, h->color);
			}
			i = i->next;
		}

		h = h->next;
	}

	// End
	printf ("}\n");
}
